/*
    This file is part of CrabEmu.

    Copyright (C) 2007, 2008, 2009, 2011, 2012, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <stdio.h>

#include <dc/pvr.h>
#include <dc/maple.h>
#include <dc/maple/controller.h>
#include <kos/sem.h>
#include <dc/video.h>
#include <dc/flashrom.h>
#include <arch/timer.h>
#include <arch/irq.h>
#include <dc/vblank.h>

#include <dc/sd.h>
#include <kos/blockdev.h>
#include <ext2/fs_ext2.h>

#include "CrabEmu.h"
#include "sms.h"
#include "smsvdp.h"
#include "smsz80.h"
#include "sn76489.h"
#include "ym2413.h"
#include "smsmem.h"
#include "colecovision.h"
#include "colecomem.h"
#include "sound.h"
#include "cheats.h"
#include "font.h"
#include "pvrutils.h"
#include "nes.h"
#include "nesmem.h"
#include "nesppu.h"

static pvr_ptr_t tex[2];
static pvr_ptr_t txr;
static pvr_poly_hdr_t hdr[2];
static int active_tex;
static semaphore_t dma_sem = SEM_INITIALIZER(0);
static int ugly_stretching_on = 0;
static int (*do_frame)(int, int);
static pvr_ptr_t font;
static int have_sd = 0;
static kos_blockdev_t sd_dev;

/* This is needed in menu.c and sound.c... */
int region;

extern uint16 sms_pad;
extern uint16 coleco_cont_bits[2];
extern int sms_console;
extern int sms_region;
extern sn76489_t psg;
extern int sms_console;
extern int sms_psg_enabled;
extern uint32 psg_samples[313];
extern int sms_cycles_run;
extern int sms_cycles_to_run;
extern int sms_cheats_enabled;

extern const char *menu(const char *dir, int *stretch, int *console,
                        pvr_ptr_t font);
extern int pal_menu(void);

static const float NTSC_Z80_CLOCK = 3579545.0f;
static const int PSG_DIVISOR = 16;
static const int NTSC_FPS = 60;
static const int NTSC_LINES_PER_FRAME = 262;
static const float NTSC_CLOCKS_PER_SAMPLE = 5.073051303855f;
#define NTSC_SAMPLES_PER_FRAME 735

static const float PAL_Z80_CLOCK = 3546893.0f;
static const int PAL_FPS = 50;
static const int PAL_LINES_PER_FRAME = 313;
static const float PAL_CLOCKS_PER_SAMPLE = 5.02677579365f;
#define PAL_SAMPLES_PER_FRAME 882

#ifdef FPS_COUNTER
static int tdiff = 0;
static int frames = 0;
#endif

static void dma_done(ptr_t data __UNUSED__) {
    sem_signal(&dma_sem);
}

static void render_fb(void) {
    pvr_vertex_t vert;

    pvr_prim(hdr + active_tex, sizeof(pvr_poly_hdr_t));

    vert.flags = PVR_CMD_VERTEX;
    vert.argb = 0xFFFFFFFF;
    vert.oargb = 0;

    if(!ugly_stretching_on) {
        if(sms_console != CONSOLE_GG && sms_console <= CONSOLE_COLECOVISION) {
            float y = (480.0f - (smsvdp.lines << 1)) / 2.0f;

            vert.x = 64.0f;
            vert.y = y + (smsvdp.lines << 1);
            vert.z = 1.0f;
            vert.u = 0.0f;
            vert.v = smsvdp.lines / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.y = y;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.x = 576.0f;
            vert.y = y + (smsvdp.lines << 1);
            vert.u = 1.0f;
            vert.v = smsvdp.lines / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.flags = PVR_CMD_VERTEX_EOL;
            vert.y = y;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
        }
        else if(sms_console == CONSOLE_GG) {
            vert.x = 80.0f;
            vert.y = 456.0f;
            vert.z = 1.0f;
            vert.u = 48.0f / 256.0f;
            vert.v = 168.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.y = 24.0f;
            vert.v = 24.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.x = 560.0f;
            vert.y = 456.0f;
            vert.u = 208.0f / 256.0f;
            vert.v = 168.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.flags = PVR_CMD_VERTEX_EOL;
            vert.y = 24.0f;
            vert.v = 24.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
        }
        else if(sms_console == CONSOLE_NES) {
            float y = (480.0f - (240 << 1)) / 2.0f;
            
            vert.x = 64.0f;
            vert.y = y + (240 << 1);
            vert.z = 1.0f;
            vert.u = 0.0f;
            vert.v = 240.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
            
            vert.y = y;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
            
            vert.x = 576.0f;
            vert.y = y + (240 << 1);
            vert.u = 1.0f;
            vert.v = 240.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
            
            vert.flags = PVR_CMD_VERTEX_EOL;
            vert.y = y;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
        }
    }
    else {
        if(sms_console != CONSOLE_GG && sms_console <= CONSOLE_COLECOVISION) {
            vert.x = 0.0f;
            vert.y = 480.0f;
            vert.z = 1.0f;
            vert.u = 0.0f;
            vert.v = smsvdp.lines / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.y = 0.0f;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.x = 640.0f;
            vert.y = 480.0f;
            vert.u = 1.0f;
            vert.v = smsvdp.lines / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.flags = PVR_CMD_VERTEX_EOL;
            vert.y = 0.0f;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
        }
        else if(sms_console == CONSOLE_GG) {
            vert.x = 0.0f;
            vert.y = 480.0f;
            vert.z = 1.0f;
            vert.u = 48.0f / 256.0f;
            vert.v = 168.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.y = 0.0f;
            vert.v = 24.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.x = 640.0f;
            vert.y = 480.0f;
            vert.u = 208.0f / 256.0f;
            vert.v = 168.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));

            vert.flags = PVR_CMD_VERTEX_EOL;
            vert.y = 0.0f;
            vert.v = 24.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
        }
        else if(sms_console == CONSOLE_NES) {
            vert.x = 0.0f;
            vert.y = 480.0f;
            vert.z = 1.0f;
            vert.u = 0.0f;
            vert.v = 240.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
            
            vert.y = 0.0f;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
            
            vert.x = 640.0f;
            vert.y = 480.0f;
            vert.u = 1.0f;
            vert.v = 240.0f / 256.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
            
            vert.flags = PVR_CMD_VERTEX_EOL;
            vert.y = 0.0f;
            vert.v = 0.0f;
            pvr_prim(&vert, sizeof(pvr_vertex_t));
        }
    }
}

static void render_scene(void) {
#ifdef FPS_COUNTER
    static char str[256];
#endif

    sem_wait(&dma_sem);

    pvr_wait_ready();
    pvr_scene_begin();

    pvr_list_begin(PVR_LIST_OP_POLY);
    render_fb();

#ifdef FPS_COUNTER
    ++frames;

    /* Hooray GCC's being stupid again! Apparently if double and float are the
       same type, passing a float to a printflike function produces a warning.
       Hence, the useless cast to double here. */
    sprintf(str, "%4.1f", (double)(1000.0f * frames / tdiff));
    font_draw_str(str, font, 0, 0, 1);

    if(frames == 60)
        tdiff = frames = 0;
#endif

    pvr_list_finish();
    pvr_scene_finish();
}

static uint32 modal_dialog(int dofb, uint32 btn_mask, const char *str) {
    maple_device_t *cont1;
    cont_state_t *state;

    pvr_wait_ready();
    pvr_scene_begin();
    
    pvr_list_begin(PVR_LIST_OP_POLY);

    if(dofb) {
        render_fb();
    }

    pvr_list_finish();

    pvr_list_begin(PVR_LIST_TR_POLY);
    draw_dialog_centered(str, 0.0f, 640.0f, 0.0f, 480.0f, 10.0f, 0xC00000B0,
                         font);
    pvr_list_finish();
    pvr_scene_finish();

    /* Wait for input... */
    for(;;) {
        if((cont1 = maple_enum_type(0, MAPLE_FUNC_CONTROLLER))) {
            if((state = (cont_state_t *)maple_dev_status(cont1))) {
                if(state->buttons & btn_mask) {
                    return state->buttons & btn_mask;
                }
            }
        }

        thd_sleep(100);
    }

    /* XXXX: Shouldn't ever get here */
    return 0;
}

static int sms_frame_ntsc(int run, int skip) {
    static int16 bufl[NTSC_SAMPLES_PER_FRAME << 1],
        bufr[NTSC_SAMPLES_PER_FRAME << 1], fmbuf[PAL_SAMPLES_PER_FRAME << 1];
    static int samples = 0;
    static int frame = 0;
    int line;

    sms_cycles_run = run;
    sms_cycles_to_run = 0;

    for(line = 0; line < NTSC_LINES_PER_FRAME; ++line) {
        sms_cycles_to_run += SMS_CYCLES_PER_LINE;
        sms_cycles_run += sms_vdp_execute(line, skip);
        sms_cycles_run += sms_z80_run(sms_cycles_to_run - sms_cycles_run);

        sn76489_execute_samples_dc(&psg, bufl + samples, bufr + samples,
                                   psg_samples[line]);
#ifdef CRABEMU_DC_ENABLE_YM2413
        YM2413UpdateOneMono(0, fmbuf + samples, psg_samples[line]);
#endif
        samples += psg_samples[line];
    }

    frame ^= 1;

    if(!skip)
        pvr_txr_load_dma(smsvdp.framebuffer, txr, 256 * 256 * 2, 0, dma_done,
                         0);

    if(frame == 0) {
        sound_update_buffer_noint(bufl, bufr, fmbuf,
                                  NTSC_SAMPLES_PER_FRAME << 2);
        samples = 0;
    }

    return sms_cycles_run - sms_cycles_to_run;
}

static int sms_frame_pal(int run, int skip) {
    static int16 bufl[PAL_SAMPLES_PER_FRAME * 2],
        bufr[PAL_SAMPLES_PER_FRAME * 2], fmbuf[PAL_SAMPLES_PER_FRAME << 1];
    static int samples = 0;
    static int frame = 0;
    int line;

    sms_cycles_run = run;
    sms_cycles_to_run = 0;

    for(line = 0; line < PAL_LINES_PER_FRAME; ++line) {
        sms_cycles_to_run += SMS_CYCLES_PER_LINE;
        sms_cycles_run += sms_vdp_execute(line, skip);
        sms_cycles_run += sms_z80_run(sms_cycles_to_run - sms_cycles_run);

        sn76489_execute_samples_dc(&psg, bufl + samples, bufr + samples,
                                   psg_samples[line]);
#ifdef CRABEMU_DC_ENABLE_YM2413
        YM2413UpdateOneMono(0, fmbuf + samples, psg_samples[line]);
#endif
        samples += psg_samples[line];
    }

    frame ^= 1;

    if(!skip)
        pvr_txr_load_dma(smsvdp.framebuffer, txr, 256 * 256 * 2, 0, dma_done,
                         0);

    if(frame == 0) {
        sound_update_buffer_noint(bufl, bufr, fmbuf,
                                  PAL_SAMPLES_PER_FRAME << 2);
        samples = 0;
    }

    return sms_cycles_run - sms_cycles_to_run;
}

static int coleco_frame_ntsc(int run, int skip) {
    static int16 bufl[NTSC_SAMPLES_PER_FRAME << 1],
        bufr[NTSC_SAMPLES_PER_FRAME << 1];
    static int samples = 0;
    static int frame = 0;
    int line;

    sms_cycles_run = run;
    sms_cycles_to_run = 0;

    for(line = 0; line < NTSC_LINES_PER_FRAME; ++line) {
        sms_cycles_to_run += SMS_CYCLES_PER_LINE;
        sms_cycles_run += tms9918a_vdp_execute(line, &sms_z80_nmi, skip);
        sms_cycles_run += sms_z80_run(sms_cycles_to_run - sms_cycles_run);

        sn76489_execute_samples_dc(&psg, bufl + samples, bufr + samples,
                                   psg_samples[line]);
        samples += psg_samples[line];
    }

    frame ^= 1;

    if(!skip)
        pvr_txr_load_dma(smsvdp.framebuffer, txr, 256 * 256 * 2, 0, dma_done,
                         0);

    if(frame == 0) {
        sound_update_buffer_noint(bufl, bufr, NULL,
                                  NTSC_SAMPLES_PER_FRAME << 2);
        samples = 0;
    }

    return sms_cycles_run - sms_cycles_to_run;
}

static int coleco_frame_pal(int run, int skip) {
    static int16 bufl[PAL_SAMPLES_PER_FRAME << 1],
        bufr[PAL_SAMPLES_PER_FRAME << 1];
    static int samples = 0;
    static int frame = 0;
    int line;

    sms_cycles_run = run;
    sms_cycles_to_run = 0;

    for(line = 0; line < PAL_LINES_PER_FRAME; ++line) {
        sms_cycles_to_run += SMS_CYCLES_PER_LINE;
        sms_cycles_run += tms9918a_vdp_execute(line, &sms_z80_nmi, skip);
        sms_cycles_run += sms_z80_run(sms_cycles_to_run - sms_cycles_run);

        sn76489_execute_samples_dc(&psg, bufl + samples, bufr + samples,
                                   psg_samples[line]);
        samples += psg_samples[line];
    }

    frame ^= 1;

    if(!skip)
        pvr_txr_load_dma(smsvdp.framebuffer, txr, 256 * 256 * 2, 0, dma_done,
                         0);

    if(frame == 0) {
        sound_update_buffer_noint(bufl, bufr, NULL, PAL_SAMPLES_PER_FRAME << 2);
        samples = 0;
    }

    return sms_cycles_run - sms_cycles_to_run;
}

static int nes_frame_ntsc(int run, int skip) {
    int rv = nes_frame(run, skip);

    if(!skip)
        pvr_txr_load_dma(nes_ppu_framebuffer(), txr, 256 * 256 * 2, 0, dma_done,
                         0);
    return rv;
}

static int do_error_dialog(int code) {
    uint32 buttons;
    char str[512];
    int amt = 0;

    if(code > 0) {
        amt = code;
        code = -101;
    }

    thd_sleep(500);

    switch(code) {
        case -100:
            buttons = modal_dialog(1, CONT_A | CONT_B,
                                   "Cannot save file!\n"
                                   "Please insert a memory card into the\n"
                                   "first port on the first controller.\n"
                                   "Press A to retry, B to cancel save.");
            break;

        case -101:
            sprintf(str, "Cannot save file!\n"
                         "The memory card in slot A1 does not\n"
                         "have enough free space. Please insert\n"
                         "a memory card with at least %d blocks\n"
                         "free.\n"
                         "Press A to retry, B to cancel save.", amt);
            buttons = modal_dialog(1, CONT_A | CONT_B, str);
            break;

        case -1:
            buttons = modal_dialog(1, CONT_A | CONT_B,
                                   "An unknown error occurred!\n"
                                   "Press A to retry, B to cancel.");
            break;

        default:
            /* ? */
            return 0;
    }

    thd_sleep(500);

    return buttons & CONT_A;
}

static void do_save_dialog(int dofb) {
    pvr_wait_ready();
    pvr_scene_begin();

    pvr_list_begin(PVR_LIST_OP_POLY);

    if(dofb) {
        render_fb();
    }

    pvr_list_finish();

    pvr_list_begin(PVR_LIST_TR_POLY);
    draw_dialog_centered("Saving...\nDo not remove any controllers or\n"
                         "memory cards!", 0.0f, 640.0f, 0.0f, 480.0f, 10.0f,
                         0xC00000B0, font);
    pvr_list_finish();
    pvr_scene_finish();
}

static void emu_loop(const char *fn) {
    maple_device_t *cont1, *cont2;
    cont_state_t *state;
    uint16 mask;
    int oldrun = 0;
    int start_pressed = 0;
    int cheats_pressed = 0;
    const char *romfn = fn;
    int romconsole, rv;
#ifdef FPS_COUNTER
    uint64 start, end;
#endif

    cont1 = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);
    cont2 = maple_enum_type(1, MAPLE_FUNC_CONTROLLER);

    for(;;) {
#ifdef FPS_COUNTER
        start = timer_ms_gettime64();
#endif
        oldrun = do_frame(oldrun, 0);
#ifdef FPS_COUNTER
        end = timer_ms_gettime64();
        tdiff += end - start;
#endif

        render_scene();

        mask = 0x0000;

        if(cont1) {
            state = (cont_state_t *)maple_dev_status(cont1);
            if(state) {
                if(sms_console < CONSOLE_COLECOVISION) {
                    if((state->buttons & CONT_START) && !start_pressed) {
                        sms_button_pressed(1, GAMEGEAR_START);
                        start_pressed = 1;
                    }
                    else if(!(state->buttons & CONT_START) && start_pressed) {
                        sms_button_released(1, GAMEGEAR_START);
                        start_pressed = 0;
                    }

                    if(state->rtrig > 20) {
                        if(state->buttons & CONT_Y) {
                            sound_pause();

                            do {
                                do_save_dialog(1);
                                rv = sms_write_cartram_to_file(romfn);

                                if(rv)
                                    rv = do_error_dialog(rv);
                            } while(rv);

                            sound_unpause();
                        }

                        if(state->buttons & CONT_A) {
                            sound_pause();

                            do {
                                do_save_dialog(1);
                                rv = sms_save_state(romfn);

                                if(rv)
                                    rv = do_error_dialog(rv);
                            } while(rv);

                            sound_unpause();
                            state->buttons &= ~CONT_A;
                        }
                        else if(state->buttons & CONT_B) {
                            sound_pause();
                            sms_load_state(NULL);
                            sound_unpause();
                            state->buttons &= ~CONT_B;
                        }

                        if(state->buttons & CONT_X && !cheats_pressed) {
                            if(sms_cheats_enabled) {
                                sms_cheat_disable();
                            }
                            else {
                                sms_cheat_enable();
                            }

                            cheats_pressed = 1;
                        }
                    }

                    mask = ((state->buttons & 0xF0) >> 4) |
                        ((state->buttons & 0x02) << 4) |
                        ((state->buttons & 0x04) << 2);
                }
                else if(sms_console == CONSOLE_COLECOVISION) {
                    coleco_cont_bits[0] = 0;

                    if(state->joyy < -20)
                        coleco_cont_bits[0] |= 1 << 8;
                    else if(state->joyy > 20)
                        coleco_cont_bits[0] |= 1 << 10;

                    if(state->joyx < -20)
                        coleco_cont_bits[0] |= 1 << 11;
                    else if(state->joyx > 20)
                        coleco_cont_bits[0] |= 1 << 9;

                    if(state->ltrig <= 20 && state->rtrig <= 20) {
                        if(state->buttons & CONT_A)
                            coleco_cont_bits[0] |= 1 << 14;
                        if(state->buttons & CONT_B)
                            coleco_cont_bits[0] |= 1 << 6;
                        if(state->buttons & CONT_DPAD_UP)
                            coleco_cont_bits[0] |= 0x02;
                        else if(state->buttons & CONT_DPAD_DOWN)
                            coleco_cont_bits[0] |= 0x08;
                        else if(state->buttons & CONT_DPAD_LEFT)
                            coleco_cont_bits[0] |= 0x03;
                        else if(state->buttons & CONT_DPAD_RIGHT)
                            coleco_cont_bits[0] |= 0x0D;
                        else if(state->buttons & CONT_X)
                            coleco_cont_bits[0] |= 0x0C;
                        else if(state->buttons & CONT_Y)
                            coleco_cont_bits[0] |= 0x01;
                    }
                    else if(state->ltrig > 20 && state->rtrig <= 20) {
                        if(state->buttons & CONT_A)
                            coleco_cont_bits[0] |= 0x0A;
                        else if(state->buttons & CONT_B)
                            coleco_cont_bits[0] |= 0x0E;
                        else if(state->buttons & CONT_DPAD_UP)
                            coleco_cont_bits[0] |= 0x04;
                        else if(state->buttons & CONT_DPAD_DOWN)
                            coleco_cont_bits[0] |= 0x06;
                        else if(state->buttons & CONT_DPAD_LEFT)
                            coleco_cont_bits[0] |= 0x05;
                        else if(state->buttons & CONT_DPAD_RIGHT)
                            coleco_cont_bits[0] |= 0x09;
                    }
                    else if(state->rtrig > 20) {
                        if(state->buttons & CONT_A) {
                            sound_pause();

                            do {
                                do_save_dialog(1);
                                rv = coleco_save_state(romfn);

                                if(rv)
                                    rv = do_error_dialog(rv);
                            } while(rv);

                            sound_unpause();
                        }
                        else if(state->buttons & CONT_B) {
                            sound_pause();
                            coleco_load_state(NULL);
                            sound_unpause();
                        }
                    }
                }
                else if(sms_console == CONSOLE_NES) {
                    if(state->rtrig > 20) {
                        if(state->buttons & CONT_Y) {
                            sound_pause();

                            do {
                                do_save_dialog(1);
                                rv = nes_mem_write_sram(romfn);

                                if(rv)
                                    rv = do_error_dialog(rv);
                            } while(rv);

                            sound_unpause();
                        }
                        else if(state->buttons & CONT_A) {
                            sound_pause();

                            do {
                                do_save_dialog(1);
                                rv = nes_save_state(romfn);

                                if(rv)
                                    rv = do_error_dialog(rv);
                            } while(rv);

                            sound_unpause();
                        }
                        else if(state->buttons & CONT_B) {
                            sound_pause();
                            nes_load_state(NULL);
                            sound_unpause();
                        }
                    }

                    nes_pad = (state->buttons & 0x00F8) |
                        ((state->buttons & 0x0006) >> 1) |
                        ((state->buttons & 0x0200) >> 7);
                }

                if(state->ltrig > 20 && state->buttons & CONT_Y) {
                    start_pressed = 0;

                    sound_pause();

                    if(sms_console < CONSOLE_COLECOVISION) {
                        do {
                            do_save_dialog(1);
                            rv = sms_write_cartram_to_file(romfn);

                            if(rv)
                                rv = do_error_dialog(rv);
                        } while(rv);

                        sms_reset();
                    }
                    else if(sms_console == CONSOLE_COLECOVISION) {
                        coleco_reset();
                    }
                    else if(sms_console == CONSOLE_NES) {
                        do {
                            do_save_dialog(1);
                            rv = nes_mem_write_sram(romfn);

                            if(rv)
                                rv = do_error_dialog(rv);
                        } while(rv);

                        nes_reset();
                    }

                    rv = 9001;
                    while(rv != ROM_LOAD_SUCCESS) {
                        romfn = menu(NULL, &ugly_stretching_on, &romconsole,
                                     font);

                        if(romconsole < CONSOLE_COLECOVISION) {
                            rv = sms_mem_load_rom(romfn, romconsole);

                            if(rv == ROM_LOAD_SUCCESS) {
                                sms_read_cartram_from_file(NULL);

                                if(region == SMS_VIDEO_NTSC)
                                    do_frame = &sms_frame_ntsc;
                                else
                                    do_frame = &sms_frame_pal;
                            }
                        }
                        else if(romconsole == CONSOLE_COLECOVISION) {
                            rv = coleco_mem_load_rom(romfn);

                            if(rv == ROM_LOAD_SUCCESS) {
                                if(region == SMS_VIDEO_NTSC)
                                    do_frame = &coleco_frame_ntsc;
                                else
                                    do_frame = &coleco_frame_pal;
                            }
                        }
                        else if(romconsole == CONSOLE_NES) {
                            rv = nes_mem_load_rom(romfn);

                            if(rv == ROM_LOAD_SUCCESS) {
                                nes_mem_read_sram(NULL);
                                do_frame = &nes_frame_ntsc;
                            }
                        }

                        if(rv != ROM_LOAD_SUCCESS) {
                            /* I probably should show a more descriptive error
                               at some point, but for now, this will do. */
                            modal_dialog(0, CONT_A, "Cannot load rom!\n"
                                         "The specified file could not\n"
                                         "be loaded. Please select a\n"
                                         "different file.\n"
                                         "Press A to continue.");
                        }
                    }

                    sound_unpause();
                }

                if(state->rtrig <= 20 || !(state->buttons & CONT_X)) {
                    cheats_pressed = 0;
                }
            }
        }

        if(cont2) {
            state = (cont_state_t *)maple_dev_status(cont2);
            if(state) {
                if(sms_console < CONSOLE_COLECOVISION) {
                    mask |= ((state->buttons & 0xF0) << 2) |
                        ((state->buttons & 0x02) << 10) |
                        ((state->buttons & 0x04) << 8);
                }
                else if(sms_console == CONSOLE_COLECOVISION) {
                    coleco_cont_bits[1] = 0;

                    if(state->joyy < -20)
                        coleco_cont_bits[1] |= 1 << 8;
                    else if(state->joyy > 20)
                        coleco_cont_bits[1] |= 1 << 10;

                    if(state->joyx < -20)
                        coleco_cont_bits[1] |= 1 << 11;
                    else if(state->joyx > 20)
                        coleco_cont_bits[1] |= 1 << 9;

                    if(state->ltrig < 20) {
                        if(state->buttons & CONT_A)
                            coleco_cont_bits[1] |= 1 << 14;
                        if(state->buttons & CONT_B)
                            coleco_cont_bits[1] |= 1 << 6;
                        if(state->buttons & CONT_DPAD_UP)
                            coleco_cont_bits[1] |= 0x02;
                        else if(state->buttons & CONT_DPAD_DOWN)
                            coleco_cont_bits[1] |= 0x08;
                        else if(state->buttons & CONT_DPAD_LEFT)
                            coleco_cont_bits[1] |= 0x03;
                        else if(state->buttons & CONT_DPAD_RIGHT)
                            coleco_cont_bits[1] |= 0x0D;
                        else if(state->buttons & CONT_X)
                            coleco_cont_bits[1] |= 0x0C;
                        else if(state->buttons & CONT_Y)
                            coleco_cont_bits[1] |= 0x01;
                    }
                    else {
                        if(state->buttons & CONT_A)
                            coleco_cont_bits[1] |= 0x0A;
                        else if(state->buttons & CONT_B)
                            coleco_cont_bits[1] |= 0x0E;
                        else if(state->buttons & CONT_DPAD_UP)
                            coleco_cont_bits[1] |= 0x04;
                        else if(state->buttons & CONT_DPAD_DOWN)
                            coleco_cont_bits[1] |= 0x06;
                        else if(state->buttons & CONT_DPAD_LEFT)
                            coleco_cont_bits[1] |= 0x05;
                        else if(state->buttons & CONT_DPAD_RIGHT)
                            coleco_cont_bits[1] |= 0x09;
                    }
                }
                else if(sms_console == CONSOLE_NES) {
                    nes_pad |= ((state->buttons & 0x00F8) |
                        ((state->buttons & 0x0006) >> 1) |
                        ((state->buttons & 0x0200) >> 7)) << 8;
                }
            }
        }

        if(sms_console < CONSOLE_COLECOVISION)
            sms_pad = mask ^ 0xFFFF;
    }
}

void gui_set_aspect(float x __UNUSED__, float y __UNUSED__) {
    /* XXXX */
}

void gui_set_title(const char *str __UNUSED__) {
    /* XXXX */
}

static int init_sd_card(void) {
    uint8 partition_type;

    if(sd_init()) {
        printf("Could not init sd support. There probably isn't a card "
               "connected.\n");
        return -1;
    }

    /* Grab the block device for the first partition on the SD card. Note that
       you must have the SD card formatted with an MBR partitioning scheme. */
    if(sd_blockdev_for_partition(0, &sd_dev, &partition_type)) {
        printf("Could not find the first partition on the SD card!\n");
        return -2;
    }

    /* Check to see if the MBR says that we have a Linux partition. */
    if(partition_type != 0x83) {
        printf("MBR indicates a non-ext2 filesystem. Will try to mount "
               "anyway\n");
    }

    /* Initialize fs_ext2 and attempt to mount the device. */
    if(fs_ext2_init()) {
        printf("Could not initialize fs_ext2!\n");
        return -3;
    }

    if(fs_ext2_mount("/sd", &sd_dev, 0)) {
        printf("Could not mount SD card as ext2fs. Please make sure the card "
               "has been properly formatted.\n");
        return -4;
    }

    have_sd = 1;

    return 0;
}

static void shutdown_sd_card(void) {
    if(!have_sd)
        return;

    /* Clean up the filesystem and everything else */
    fs_ext2_unmount("/sd");
    fs_ext2_shutdown();
    sd_shutdown();
}

static void cont_callback(uint8 addr __UNUSED__, uint32 buttons __UNUSED__) {
    shutdown_sd_card();
    exit(0);
}

static const char *crabemu_dir(void) {
    FILE *fp;

    /* If we don't have an SD card, this is easy. */
    if(!have_sd) {
        return "/cd/crabemu";
    }
    else {
        /* Ok, the user has a card, look to see if there's a crabemu directory 
           on the card and if it looks like we have the UI files there (for now,
           I only look for the first file needed). */
        if(!(fp = fopen("/sd/crabemu/title.png", "rb"))) {
            return "/cd/crabemu";
        }

        /* Ok, the title screen is there. Assume we'll start there. */
        fclose(fp);
        return "/sd/crabemu";
    }
}

int main(int argc __UNUSED__, char *argv[] __UNUSED__) {
    pvr_poly_cxt_t cxt;
    int dc_region, ct;
    const char *romfn;
    int romconsole;
    int err = 9001;

    printf("...\n");

    dc_region = flashrom_get_region();
    ct = vid_check_cable();

    /* Prompt the user for whether to run in PAL50 or PAL60 if the flashrom says
       the Dreamcast is European and a VGA Box is not hooked up. */ 
    if(dc_region == FLASHROM_REGION_EUROPE && ct != CT_VGA) {
        if(pal_menu()) {
            vid_set_mode(DM_640x480_NTSC_IL, PM_RGB565);
            region = SMS_VIDEO_NTSC;
        }
        else {
            vid_set_mode(DM_640x480_PAL_IL, PM_RGB565);
            region = SMS_VIDEO_PAL;
        }
    }
    else {
        region = SMS_VIDEO_NTSC;
    }

    pvr_init_defaults();

    cont_btn_callback(0, CONT_A | CONT_B | CONT_X | CONT_Y, cont_callback);

    tex[0] = pvr_mem_malloc(256 * 256 * 2);
    tex[1] = pvr_mem_malloc(256 * 256 * 2);

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY,
                     PVR_TXRFMT_ARGB1555 | PVR_TXRFMT_NONTWIDDLED, 256, 256,
                     tex[0], PVR_FILTER_NONE);
    pvr_poly_compile(&hdr[0], &cxt);

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY,
                     PVR_TXRFMT_ARGB1555 | PVR_TXRFMT_NONTWIDDLED, 256, 256,
                     tex[1], PVR_FILTER_NONE);
    pvr_poly_compile(&hdr[1], &cxt);

    active_tex = 0;
    txr = tex[0];

    font = font_load_txr();

    init_sd_card();

    /* Are we going to start on the SD card or the CD? */
    fs_chdir(crabemu_dir());

    while(err != ROM_LOAD_SUCCESS) {
#ifndef NO_GUI
        romfn = menu("roms", &ugly_stretching_on, &romconsole, font);
#else
        romfn = menu("/", &ugly_stretching_on, &romconsole, font);
#endif

        if(romconsole < CONSOLE_COLECOVISION) {
            err = sms_mem_load_rom(romfn, romconsole);

            if(err == ROM_LOAD_SUCCESS) {
                sms_read_cartram_from_file(NULL);

                if(region == SMS_VIDEO_NTSC)
                    do_frame = &sms_frame_ntsc;
                else
                    do_frame = &sms_frame_pal;
            }
        }
        else if(romconsole == CONSOLE_COLECOVISION) {
            err = coleco_mem_load_rom(romfn);

            if(err == ROM_LOAD_SUCCESS) {
                if(region == SMS_VIDEO_NTSC)
                    do_frame = &coleco_frame_ntsc;
                else
                    do_frame = &coleco_frame_pal;
            }
        }
        else if(romconsole == CONSOLE_NES) {
            err = nes_mem_load_rom(romfn);

            if(err == ROM_LOAD_SUCCESS) {
                nes_mem_read_sram(NULL);

                do_frame = &nes_frame_ntsc;
            }
        }

        if(err != ROM_LOAD_SUCCESS) {
            /* I probably should show a more descriptive error at some point,
               but for now, this will do. */
            modal_dialog(0, CONT_A, "Cannot load rom!\n"
                                    "The specified file could not\n"
                                    "be loaded. Please select a\n"
                                    "different file.\n"
                                    "Press A to continue.");
        }
    }

    emu_loop(romfn);

    shutdown_sd_card();

    return 0;
}
