/*
    This file is part of CrabEmu.

    Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/queue.h>

#include <kos/fs.h>
#include <png/png.h>
#include <dc/pvr.h>
#include <dc/maple.h>
#include <dc/maple/controller.h>
#include <arch/timer.h>

#include "sms.h"
#include "colecovision.h"
#include "nes.h"
#include "smsmem.h"
#include "smsmem-gg.h"
#include "colecomem.h"
#include "font.h"
#include "sound.h"
#include "rom.h"
#include "pvrutils.h"

static char filename[MAX_FN_LEN];
static char curdir[4096];

#ifndef NO_GUI
static char uidir[4096];
static int firstTime = 1;
#endif

struct RomData {
    dirent_t dirent;
    int type;

    STAILQ_ENTRY(RomData) romlist;
};

STAILQ_HEAD(RomDataList, RomData);
static struct RomDataList romList;
static char **filenames;
static int files = 0;

extern int sms_initialized;
extern int coleco_initialized;
extern int nes_initialized;
extern int region;

static struct RomData *get_rom_entry(int num) {
    struct RomData *i;

    STAILQ_FOREACH(i, &romList, romlist) {
        if(!num--) {
            return i;
        }
    }

    return NULL;
}

static int have_coleco_bios(void) {
    FILE *fp;

#ifndef DEBUG
    fp = fopen("coleco.rom", "rb");
#else
    fp = fopen("/pc/crabemu/coleco.rom", "rb");
#endif

    if(fp) {
        fclose(fp);
        return 1;
    }

    return 0;
}

#ifndef NO_GUI
static pvr_ptr_t load_texture(const char *fn, uint32 mask, pvr_list_t list,
                              pvr_poly_hdr_t *hdr) {
    pvr_poly_cxt_t cxt;
    pvr_ptr_t rv;
    uint32 w, h;
    uint32 fmt;

    if(mask == PNG_NO_ALPHA) {
        fmt = PVR_TXRFMT_RGB565;
    }
    else if(mask == PNG_MASK_ALPHA) {
        fmt = PVR_TXRFMT_ARGB1555;
    }
    else {
        fmt = PVR_TXRFMT_ARGB4444;
    }

    png_load_texture(fn, &rv, mask, &w, &h);
    pvr_poly_cxt_txr(&cxt, list, fmt, w, h, rv, PVR_FILTER_BILINEAR);
    pvr_poly_compile(hdr, &cxt);

    return rv;
}
#endif

static void create_list(const char *dir) {
    file_t d;
    dirent_t *ent;
    int type, i = 0;
    struct RomData *data;
    int have_coleco = have_coleco_bios();

    /* Open up the directory in question */
    d = fs_open(dir, O_DIR);

    if(d == FILEHND_INVALID) {
        return;
    }

    STAILQ_INIT(&romList);

    while((ent = fs_readdir(d))) {
        type = -1;

        /* Don't include any . or .. here */
        if(!strcmp(ent->name, ".") || !strcmp(ent->name, "..")) {
            continue;
        }

        /* Mark subdirectories */
        if(ent->attr & O_DIR) {
            type = -2;
        }
        /* Determine if this rom is of a supported console */
        else if((type = rom_detect_console(ent->name)) <= 0) {
            type = -1;
        }

        if(type == CONSOLE_COLECOVISION && !have_coleco) {
            type = -1;
        }

        /* Add the rom to the list if it is supported */
        if(type != -1) {
            data = (struct RomData *)malloc(sizeof(struct RomData));
            data->type = type;
            data->dirent = *ent;

            STAILQ_INSERT_TAIL(&romList, data, romlist);
            ++files;
        }
    }

    fs_close(d);

    /* Add an entry for ../, as long as we're not at / */
    if(strcmp(dir, "/")) {
        data = (struct RomData *)malloc(sizeof(struct RomData));
        data->type = -2;
        strcpy(data->dirent.name, "..");

        STAILQ_INSERT_HEAD(&romList, data, romlist);
        ++files;
    }

    if(files) {
        filenames = (char **)malloc(sizeof(char *) * files);

        STAILQ_FOREACH(data, &romList, romlist) {
            if(data->type == -2) {
                filenames[i] = (char *)malloc(strlen(data->dirent.name) + 2);
                sprintf(filenames[i], "%s/", data->dirent.name);
            }
            else {
                filenames[i] = (char *)malloc(strlen(data->dirent.name) + 1);
                strcpy(filenames[i], data->dirent.name);
            }

            ++i;
        }
    }
}

static void destroy_list() {
    struct RomData *ent, *ent2;
    int i;

    ent = STAILQ_FIRST(&romList);
    while(ent) {
        ent2 = STAILQ_NEXT(ent, romlist);
        STAILQ_REMOVE_HEAD(&romList, romlist);
        free(ent);
        ent = ent2;
    }

    for(i = 0; i < files; ++i) {
        free(filenames[i]);
    }

    free(filenames);
    files = 0;
}

#ifndef NO_GUI
static void draw_textured_quad(pvr_poly_hdr_t *hdr, float x, float y, float z,
                               float w, float h, float umax, float vmax) {
    pvr_vertex_t vert;

    pvr_prim(hdr, sizeof(pvr_poly_hdr_t));

    vert.flags = PVR_CMD_VERTEX;
    vert.argb = 0xFFFFFFFF;
    vert.oargb = 0x00000000;

    vert.x = x;
    vert.y = y + h;
    vert.z = z;
    vert.u = 0.0f;
    vert.v = vmax;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.y = y;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.x = x + w;
    vert.y = y + h;
    vert.u = umax;
    vert.v = vmax;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.y = y;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));
}
#endif

static void load_bios_sms(void) {
    FILE *fp;

#ifndef DEBUG
    fp = fopen("bios.sms", "rb");
    if(fp) {
        fclose(fp);
        sms_mem_load_bios("bios.sms");
        sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                               SMS_MEMCTL_RAM) ^ 0xFF);
    }
#else
    fp = fopen("/pc/crabemu/bios.sms", "rb");
    if(fp) {
        fclose(fp);
        sms_mem_load_bios("/pc/crabemu/bios.sms");
        sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                              SMS_MEMCTL_RAM) ^ 0xFF);
    }
#endif
}

static void load_bios_gg(void) {
    FILE *fp;

#ifndef DEBUG
    fp = fopen("bios.gg", "rb");
    if(fp) {
        fclose(fp);
        sms_mem_load_bios("bios.gg");
        sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                 SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
    }
#else
    fp = fopen("/pc/crabemu/bios.gg", "rb");
    if(fp) {
        fclose(fp);
        sms_mem_load_bios("/pc/crabemu/bios.gg");
        sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                 SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
    }
#endif
}

static void load_bios_coleco(void) {
    FILE *fp;

#ifndef DEBUG
    fp = fopen("coleco.rom", "rb");
    if(fp) {
        fclose(fp);
        coleco_mem_load_bios("coleco.rom");
    }
#else
    fp = fopen("/pc/crabemu/coleco.rom", "rb");
    if(fp) {
        fclose(fp);
        coleco_mem_load_bios("/pc/crabemu/coleco.rom");
    }
#endif
}

const char *menu(const char *dir, int *stretch, int *console, pvr_ptr_t font) {
#ifndef NO_GUI
    pvr_ptr_t crab;
    pvr_ptr_t title;
    pvr_poly_hdr_t hdr_title, hdr_crab;
#endif
    int topItem = 0;
    int selected = 0;
    int selectedType = -1;
    int i;
    maple_device_t *cont1;
    cont_state_t *state;
    uint64 timer = timer_ms_gettime64();
    file_t tmpf;

    *stretch = 0;

#ifndef NO_GUI
    if(firstTime) {
        strcpy(uidir, fs_getwd());
        title = load_texture("title.png", PNG_NO_ALPHA, PVR_LIST_OP_POLY,
                             &hdr_title);

        timer = timer_ms_gettime64();

        while(timer + 5000 > timer_ms_gettime64()) {
            pvr_wait_ready();
            pvr_scene_begin();
            
            pvr_list_begin(PVR_LIST_OP_POLY);
            draw_textured_quad(&hdr_title, 0.0f, 0.0f, 1.0f, 640.0f, 480.0f,
                               640.0f / 1024.0f, 480.0f / 512.0f);
            
            pvr_list_finish();
            pvr_scene_finish();
        }

        firstTime = 0;
        pvr_mem_free(title);
    }
#endif

    cont1 = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);

#ifndef NO_GUI
    fs_chdir(uidir);
    crab = load_texture("crab.png", PNG_FULL_ALPHA, PVR_LIST_TR_POLY,
                        &hdr_crab);
#endif

    /* If we've specified a directory, use it, as long as we can get to it,
       otherwise try /cd... If that fails, go with /. */
    if(dir) {
        if((tmpf = fs_open(dir, O_DIR)) != FILEHND_INVALID) {
            fs_close(tmpf);
            fs_chdir(dir);
        }
        else if((tmpf = fs_open("/cd", O_DIR)) != FILEHND_INVALID) {
            fs_close(tmpf);
            fs_chdir("/cd");
        }
        else {
            fs_chdir("/");
        }
    }
    else {
        fs_chdir(curdir);
    }

    strcpy(curdir, fs_getwd());
    
    /* Put a / at the end if we're not at / */
    if(curdir[1] != '\0') {
        strcat(curdir, "/");
    }

    create_list(curdir);

    for(;;) {
        if(timer + 200 < timer_ms_gettime64()) {
            state = (cont_state_t *)maple_dev_status(cont1);

            if(state) {
                if(state->buttons & CONT_DPAD_DOWN) {
                    ++selected;

                    if(selected == files) {
                        selected = files - 1;
                    }
                    else if(selected == 13) {
                        if(topItem + 13 < files) {
                            ++topItem;
                        }

                        selected = 12;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }
                else if(state->buttons & CONT_DPAD_UP) {
                    --selected;

                    if(selected == -1) {
                        if(topItem) {
                            --topItem;
                        }

                        selected = 0;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }

                if(state->buttons & CONT_DPAD_RIGHT) {
                    if(files < 13) {
                        selected = files - 1;
                        topItem = 0;
                    }
                    else if(topItem + 26 <= files) {
                        topItem += 13;
                    }
                    else {
                        selected = 12;
                        topItem = files - 13;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }
                else if(state->buttons & CONT_DPAD_LEFT) {
                    if(topItem < 13) {
                        topItem = 0;
                        selected = 0;
                    }
                    else {
                        topItem -= 13;
                    }

                    selectedType = -1;
                    timer = timer_ms_gettime64();
                }

                if(state->buttons & CONT_A) {
                    if(selectedType > 0) {
                        if(state->ltrig > 20)
                            *stretch = 1;

                        strcpy(filename, filenames[selected + topItem]);
                        break;
                    }
                    else {
                        /* Change to the new directory. This feels like cheating
                           a little bit to get where we are this way, but it
                           actually works, so why not? */
                        fs_chdir(filenames[selected + topItem]);
                        strcpy(curdir, fs_getwd());

                        /* Put a / at the end if we're not at / */
                        if(curdir[1] != '\0') {
                            strcat(curdir, "/");
                        }

                        /* Rebuild the list */
                        destroy_list();
                        create_list(curdir);
                        selected = topItem = 0;
                        selectedType = -1;
                        timer = timer_ms_gettime64() + 300;
                    }
                }
            }
        }
    
        if(selectedType == -1) {
            struct RomData *d = get_rom_entry(selected + topItem);
            selectedType = d->type;
        }

        pvr_wait_ready();
        pvr_scene_begin();

        pvr_list_begin(PVR_LIST_OP_POLY);
        draw_multicolored_quad(PVR_LIST_OP_POLY, 0.0f, 0.0f, 1.0f, 640.0f,
                               480.0f, 0xff44515a, 0xff4d5c66, 0xff44515a,
                               0xff4d5c66);
        pvr_list_finish();

        pvr_list_begin(PVR_LIST_TR_POLY);
#ifndef NO_GUI
        draw_textured_quad(&hdr_crab, 48.0f, 310.0f, 1.1f, 128.0f, 128.0f,
                           1.0f, 1.0f);
#endif
        draw_colored_quad(PVR_LIST_TR_POLY, 226.0f, 70.0f, 1.1f, 375.0f, 340.0f,
                          0x80000000);
        draw_colored_quad(PVR_LIST_TR_POLY, 226.0f, 44.0f, 1.1f, 375.0f, 26.0f,
                          0x40000000);

        draw_colored_quad(PVR_LIST_TR_POLY, 226.0f, 70.0f + 26 * selected,
                          1.2f, 375.0f, 26.0f, 0x80ffffff);

        for(i = 0; i < 13 && i + topItem < files; ++i) {
            font_draw_str(filenames[i + topItem], font, 250, 70 + 26 * i,
                          2);
        }

        /* Let the user know the current directory */
        font_draw_str(curdir, font, 226, 44, 2);
        pvr_list_finish();

        pvr_scene_finish();
    }

#ifndef NO_GUI
    pvr_mem_free(crab);
#endif

    destroy_list();

    if(selectedType == CONSOLE_COLECOVISION) {
        if(sms_initialized)
            sms_shutdown();
        else if(nes_initialized)
            nes_shutdown();

        if(!coleco_initialized)
            coleco_init(region);
        else
            coleco_reset();
    }
    else if(selectedType != CONSOLE_NES) {
        if(coleco_initialized)
            coleco_shutdown();
        else if(nes_initialized)
            nes_shutdown();

        if(!sms_initialized)
            sms_init(region, SMS_REGION_EXPORT);
        else
            sms_reset();
    }
    else if(selectedType == CONSOLE_NES) {
        if(sms_initialized)
            sms_shutdown();
        else if(coleco_initialized)
            coleco_shutdown();

        if(!nes_initialized)
            nes_init(region);
        else
            nes_reset();
    }

    if(selectedType == CONSOLE_SMS)
        load_bios_sms();
    else if(selectedType == CONSOLE_GG)
        load_bios_gg();
    else if(selectedType == CONSOLE_COLECOVISION)
        load_bios_coleco();

    *console = selectedType;

    return filename;
}
