/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007, 2008, 2009, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <unistd.h>

#include <OpenGL/OpenGL.h>

#include "NSCondition.h"
#include "EmuController.h"
#include "EmuOGLView.h"
#include "sms.h"
#include "colecovision.h"
#include "smsmem.h"
#include "colecomem.h"
#include "smsmem-gg.h"
#include "smsz80.h"
#include "sn76489.h"
#include "sound.h"
#include "smsvdp.h"
#include "gamepad.h"
#include "cheats.h"
#include "rom.h"
#include "nes.h"
#include "nesmem.h"

/* From main.m */
extern NSLock *memoryLock;
extern EmuOGLView *theview;
extern EmuController *thecont;
extern void emu_loop();
extern BOOL debuggerEnabled;
extern NSCondition *ss_sem;
extern BOOL ntscVideo;

/* from sms.c */
extern int sms_psg_enabled;
extern int sms_ym2413_enabled;
extern sn76489_t psg;
extern int sms_initialized;
extern int sms_console;
extern sms_vdp_t smsvdp;

/* from colecovision.c */
extern int coleco_initialized;

/* From nes.c */
extern int nes_initialized;

int started = 0;

/* For gamepad input... */
static void gp_cb(gamepad_t *gp, gp_elemdata_t *elem, int value) {
    [thecont eventFromGamepad:gp element:elem value:value];
}

@implementation EmuController

- (void)application:(NSApplication *)app openFiles:(NSArray *)filenames
{
    NSString *fn;
    int initted = sms_initialized || coleco_initialized || nes_initialized;

    /* Eventually we should do something different if for some reason someone
       tries to send us multiple files, but for now, punt and open the first
       file in the list */
    fn = [filenames objectAtIndex:0];

    if([self openRomFromFile:fn] == NO) {
        [NSApp replyToOpenOrPrint:NSApplicationDelegateReplyFailure];
        return;
    }

    if(!initted && [prefs boolForKey:@"Start Fullscreen"]) {
        [self toggleFullscreen:self];
    }

    [NSApp replyToOpenOrPrint:NSApplicationDelegateReplySuccess];
}

- (NSApplicationTerminateReply)applicationShouldTerminate:(NSApplication *)app
{
    if(sms_initialized) {
        sms_write_cartram_to_file([openRomSaveFn UTF8String]);
        sms_cheat_write([cheatFn UTF8String]);
    }
    else if(nes_initialized) {
        nes_mem_write_sram([openRomSaveFn UTF8String]);
    }

    if(j1map != nil) {
        [j1map release];
    }

    if(j2map != nil) {
        [j2map release];
    }

    /* Release all gamepads. This will close any open gamepads as well. */
    gp_cleanup();

    return NSTerminateNow;
}

- (BOOL)windowShouldClose:(id)sender
{
    int i;

    if(sms_initialized == 1) {
        sms_write_cartram_to_file([openRomSaveFn UTF8String]);
        sms_cheat_write([cheatFn UTF8String]);
    }
    else if(nes_initialized) {
        nes_mem_write_sram([openRomSaveFn UTF8String]);
    }

    running = NO;

    /* Disable stuff that should only be enabled while a ROM is running... */
    [screenshot setEnabled:NO];
    [fullscreen setEnabled:NO];
    [saveSRAM setEnabled:NO];

    for(i = 0; i < [saveStates numberOfItems]; ++i) {
        [[saveStates itemAtIndex:i] setEnabled:NO];
    }

    for(i = 0; i < [loadStates numberOfItems]; ++i) {
        [[loadStates itemAtIndex:i] setEnabled:NO];
    }

    [pause setEnabled:NO];
    [reset setEnabled:NO];
    [enableCheats setEnabled:NO];
    [editCheats setEnabled:NO];
    [debugger setEnabled:NO];
    [showVDPPalette setEnabled:NO];
    [showVDPPatterns setEnabled:NO];

    /* ...and enable those settings we didn't want available while a ROM is
       running. */
    [ntsc setEnabled:YES];
    [pal setEnabled:YES];
    [rExport setEnabled:YES];
    [rDomestic setEnabled:YES];

    [openRom release];
    [openRomSaveFn release];
    [cheatFn release];

    openRom = nil;
    openRomSaveFn = nil;
    cheatFn = nil;

    return YES;
}

- (void)awakeFromNib
{
    thecont = self;

    /* Make the pattern and palette panels not able to become key windows for
       the application. This fixes some potential problems with them and
       fullscreen mode. */
    [vdpPalettePanel setBecomesKeyOnlyIfNeeded:YES];
    [vdpPatternPanel setBecomesKeyOnlyIfNeeded:YES];
    [nesPatternPanel setBecomesKeyOnlyIfNeeded:YES];

    prefs = [[NSUserDefaults standardUserDefaults] retain];

    openRom = nil;
    openRomSaveFn = nil;
    cheatFn = nil;
    started = 0;

    [self readPrefs];

    if(CGLLockContext == NULL) {
        [fullscreen setEnabled:NO];
    }

    j1map = nil;
    j2map = nil;
    joy1 = NULL;
    joy2 = NULL;

    p1curinput = p1keyboard;
    p2curinput = p2keyboard;

    [prefsController scanJoysticks:nil];
    runningLock = [[NSLock alloc] init];
}

- (void)dealloc
{
    [runningLock release];
    [openRom release];
    [openRomSaveFn release];
    [cheatFn release];
    [prefs release];
    [keymap release];

    [super dealloc];
}

- (IBAction)enableNoise:(id)sender
{
    /* [noise state] is the state AFTER Cocoa changes it */
	if([noise state] == NSOffState)	{
		psg.enabled_channels &= ~ENABLE_NOISE;
	}
	else {
		psg.enabled_channels |= ENABLE_NOISE;
	}
}

- (IBAction)enablePSG:(id)sender
{
    /* [PSG state] is the state AFTER Cocoa changes it */
	if([PSG state] == NSOffState) {
		sms_psg_enabled = 0;
	}
	else {
		sms_psg_enabled = 1;
	}
}

- (IBAction)enableYM2413:(id)sender
{
    /* [YM2413 state] is the state AFTER Cocoa changes it */
	if([YM2413 state] == NSOffState) {
		sms_ym2413_enabled = 0;
	}
	else {
		sms_ym2413_enabled = 1;
	}
}

- (IBAction)enableTone0:(id)sender
{
    /* [tone0 state] is the state AFTER Cocoa changes it */
	if([tone0 state] == NSOffState)	{
		psg.enabled_channels &= ~ENABLE_TONE0;
	}
	else {
		psg.enabled_channels |= ENABLE_TONE0;
	}
}

- (IBAction)enableTone1:(id)sender
{
    /* [tone1 state] is the state AFTER Cocoa changes it */
	if([tone1 state] == NSOffState)	{
		psg.enabled_channels &= ~ENABLE_TONE1;
	}
	else {
		psg.enabled_channels |= ENABLE_TONE1;
	}
}

- (IBAction)enableTone2:(id)sender
{
    /* [tone2 state] is the state AFTER Cocoa changes it */
	if([tone2 state] == NSOffState)	{
		psg.enabled_channels &= ~ENABLE_TONE2;
	}
	else {
		psg.enabled_channels |= ENABLE_TONE2;
	}
}

- (IBAction)loadMasterRom:(id)sender
{
	NSOpenPanel *openPanel = [NSOpenPanel openPanel];
    NSArray *fileTypes =
        [NSArray arrayWithObjects:@"sms", @"gg", @"sc", @"sg", @"gz", @"bz2",
                                  @"zip", @"col", @"rom", @"nes", nil];
	[openPanel setTitle:@"Choose a Rom"];
	[openPanel setCanChooseDirectories:NO];
	
	if([openPanel runModalForTypes:fileTypes] == NSOKButton) {
        [self openRomFromFile:[openPanel filename]];
	}
}

- (void)eventFromGamepad:(gamepad_t *)pad element:(gp_elemdata_t *)elem
                   value:(int)value
{
    void (*pfunc)(int, int);
    void (*rfunc)(int, int);
    const uint32 *cont;
    int pl, len, i, range;
    uint32 val;

    /* Which button function do we need... */
    switch(sms_console) {
        case CONSOLE_SMS:
        case CONSOLE_GG:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
            pfunc = &sms_button_pressed;
            rfunc = &sms_button_released;
            len = 8;
            break;

        case CONSOLE_COLECOVISION:
            pfunc = &coleco_button_pressed;
            rfunc = &coleco_button_released;
            len = 18;
            break;

        case CONSOLE_NES:
            pfunc = &nes_button_pressed;
            rfunc = &nes_button_released;
            len = 8;
            break;

        default:
            return;
    }

    /* See if the controls are configured properly... */
    if(!(cont = [prefsController gamepadAssignmentForConsole:sms_console
                                                     gamepad:pad]))
        return;

    /* Figure out who's input we're looking at... */
    if(pad == joy1)
        pl = 1;
    else if(pad == joy2)
        pl = 2;
    else
        return;

    switch(elem->type) {
        case GP_TYPE_BUTTON:
            for(i = 0; i < len; ++i) {
                if(cont[i] == (uint32)elem->number) {
                    if(value)
                        pfunc(pl, i);
                    else
                        rfunc(pl, i);
                }
            }
            break;
            
        case GP_TYPE_HAT:
            val = 0x80000000 | elem->number;
            for(i = 0; i < len; ++i) {
                if((cont[i] & 0xFF0000FF) == val) {
                    if((cont[i] >> 8) & value)
                        pfunc(pl, i);
                    else
                        rfunc(pl, i);
                }
            }
            break;

        case GP_TYPE_X_AXIS:
        case GP_TYPE_Y_AXIS:
        case GP_TYPE_Z_AXIS:
        case GP_TYPE_RX_AXIS:
        case GP_TYPE_RY_AXIS:
        case GP_TYPE_RZ_AXIS:
            range = elem->max - elem->min + 1;
            value -= range / 2;
            val = 0x40000000 | elem->type;

            for(i = 0; i < len; ++i) {
                if((cont[i] & 0x400000FF) == val) {
                    if(cont[i] & 0x20000000) {
                        if(value < -(range / 4))
                            pfunc(pl, i);
                        else
                            rfunc(pl, i);
                    }
                    else {
                        if(value > (range / 4))
                            pfunc(pl, i);
                        else
                            rfunc(pl, i);
                    }
                }
            }
            break;
    }
}

- (void)updatePalette
{
    if([showVDPPalette state] == NSOnState) {
        [paletteView setNeedsDisplay:YES];
    }
}

- (void)updatePatterns
{
    if(sms_initialized && [showVDPPatterns state] == NSOnState) {
        [patternView setNeedsDisplay:YES];
    }
    else if(nes_initialized && [showNESPatterns state] == NSOnState) {
        [nesPatternView setNeedsDisplay:YES];
    }
}

- (void)updateSDSCTerminalWithString:(char[25][80 * 2])console
{
    [sdscView updateString:console];
}

- (void)masterEmuThread:(id)junk
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    [runningLock lock];

	emu_loop();

    if(sms_initialized)
        sms_shutdown();
    else if(coleco_initialized)
        coleco_shutdown();
    else if(nes_initialized)
        nes_shutdown();

    [runningLock unlock];

	[pool release];
}

- (void)pauseEmulation:(id)sender
{
    if(started == 0)
        return;

    if(sms_console == CONSOLE_SMS) {
        [memoryLock lock];
        sms_z80_nmi();
        [memoryLock unlock];
    }

    if([pause state] == NSOffState) {
        if(sms_console != CONSOLE_SMS) {
            [memoryLock lock];
            sound_pause();
        }

        [pause setState: NSOnState];
        
    }
    else {
        if(sms_console != CONSOLE_SMS) {
            sound_unpause();
            [memoryLock unlock];
        }

        [pause setState: NSOffState];
    }
}

- (IBAction)enableDebugger:(id)sender
{
    if([debugger state] == NSOffState) {
        [debugger setState:NSOnState];
        [smsVdpDebuggerWindow makeKeyAndOrderFront:self];
        [debuggerWindow makeKeyAndOrderFront:self];
        debuggerEnabled = YES;
    }
    else {
        [debugger setState:NSOffState];
        [smsVdpDebuggerWindow orderOut:self];
        [debuggerWindow orderOut:self];
        debuggerEnabled = NO;
        [ss_sem signal];
    }
}

- (NSString *)keyListForConsole:(int)console
                         player:(int)player
{
    if(player == 1 && [p1keyboard state] == NSOffState &&
       ![prefsController keyboardWithGamepadEnabled]) {
        return [NSString string];
    }
    else if(player == 2 && [p2keyboard state] == NSOffState &&
            ![prefsController keyboardWithGamepadEnabled]) {
        return [NSString string];
    }

    return [prefsController keyListForConsole:console player:player];
}

- (IBAction)saveSRAM:(id)sender
{
    if(sms_initialized)
        sms_write_cartram_to_file([openRomSaveFn UTF8String]);
    else if(nes_initialized)
        nes_mem_write_sram([openRomSaveFn UTF8String]);
}

- (IBAction)loadSaveState:(id)sender
{
    NSString *fn;
    NSString *ext;
    NSString *ffn;
    BOOL save = [sender tag] < 10;
    int slot = [sender tag] - (save ? 0 : 10);
    int (*save_func)(const char *);
    int (*load_func)(const char *);

    /* Which function do we want? */
    if(sms_initialized) {
        save_func = &sms_save_state;
        load_func = &sms_load_state;
    }
    else if(coleco_initialized) {
        save_func = &coleco_save_state;
        load_func = &coleco_load_state;
    }
    else if(nes_initialized) {
        save_func = &nes_save_state;
        load_func = &nes_load_state;
    }
    else {
        /* ?! */
        return;
    }

    ext = [NSString stringWithFormat:@"%d.cestate", slot];
    fn = [openRom stringByAppendingPathExtension:ext];
    ffn = [[saveStateDir stringValue] stringByAppendingPathComponent:fn];

    [memoryLock lock];

    if(save) {
        if(save_func([ffn UTF8String])) {
            NSString *title = [NSString stringWithFormat:@"Cannot open save "
                               "state file %@.", ffn];
            NSRunAlertPanel(title, @"CrabEmu cannot open this file for "
                            "writing. Verify that you have the permissions to "
                            "write to this directory and that the target "
                            "volume is not full.", @"OK", NULL, NULL);
        }
    }
    else {
        if(load_func([ffn UTF8String])) {
            NSString *title = [NSString stringWithFormat:@"Cannot open save "
                               "state file %@.", ffn];
            NSRunAlertPanel(title, @"CrabEmu cannot open this file for "
                            "reading. Verify that you have the permissions to "
                            "read to this directory and that the file exists.", 
                            @"OK", NULL, NULL);
        }
    }

    [memoryLock unlock];
}

- (IBAction)resetEmulation:(id)sender
{
    if(sms_console != CONSOLE_NES)
        sms_soft_reset();
    else
        nes_soft_reset();
}

- (IBAction)modeNTSC:(id)sender
{
    if(!sms_initialized && !coleco_initialized) {
        if([ntsc state] == NSOffState) {
            [ntsc setState:NSOnState];
            [pal setState:NSOffState];
            ntscVideo = YES;
        }

        [self savePrefs];
    }
}

- (IBAction)modePAL:(id)sender
{
    if(!sms_initialized && !coleco_initialized) {
        if([pal state] == NSOffState) {
            [pal setState:NSOnState];
            [ntsc setState:NSOffState];
            ntscVideo = NO;
        }

        [self savePrefs];
    }
}

- (IBAction)regionDomestic:(id)sender
{
    if(sms_initialized != 1) {
        if([rDomestic state] == NSOffState) {
            [rDomestic setState:NSOnState];
            [rExport setState:NSOffState];
            domestic = YES;
        }
        
        [self savePrefs];
    }
}

- (IBAction)regionExport:(id)sender
{
    if(sms_initialized != 1) {
        if([rExport state] == NSOffState) {
            [rExport setState:NSOnState];
            [rDomestic setState:NSOffState];
            domestic = NO;
        }
        
        [self savePrefs];
    }
}

- (IBAction)setInput1:(id)sender
{
    NSMenuItem *item = (NSMenuItem *)sender;
    int tag;
    NSString *str;
    gamepad_t *old = joy1;

    /* If we have tried to select the same input device we currently are using,
       ignore it. */
    if([item state] == NSOnState) {
        return;
    }

    /* Lock the memory lock, and grab the tag of the menu item. */
    [memoryLock lock];
    tag = [item tag];

    /* If the user has selected the keyboard, close the old gamepad if any,
       and release the old button mapping. */
    if(tag == -1) {
        if(old) {
            /* Release resources associated with the gamepad. */
            gp_close(old);
            [j1map release];
            j1map = nil;
            joy1 = NULL;

            /* Reenable the old source on player 2's menu. */
            [[p2input itemAtIndex:[p1curinput tag] + 1] setEnabled:YES];

            /* Deselect the old input source. */
            [p1curinput setState:NSOffState];

            /* Select the keyboard in the menu. */
            [p1keyboard setState:NSOnState];
            p1curinput = p1keyboard;
        }
    }
    else {
        /* Grab the specified gamepad. */
        joy1 = gp_get_gamepad(tag);

        /* Attempt to open the specified gamepad. */
        if(gp_open(joy1, &gp_cb)) {
            /* The gamepad has not been opened, show an error and keep the old
               selected input source. */
            NSString *title = [NSString stringWithFormat:@"Cannot open "
                "gamepad %s.", joy1->name];
            NSRunAlertPanel(title, @"CrabEmu cannot open the specified "
                            "gamepad. Verify that the gamepad is attached "
                            "to the system and is powered on (if applicable).", 
                            @"OK", NULL, NULL);

            joy1 = old;
        }
        else {
            /* Grab the button mapping for the specified gamepad. */
            if(j1map != nil) {
                [j1map release];
            }

            str = [NSString stringWithUTF8String:joy1->name];
            j1map = (NSMutableDictionary *)[[keymap objectForKey:str] retain];

            /* Close the old gamepad, if any. */
            if(old) {
                gp_close(old);
            }

            /* Disable the gamepad on the second player's menu. Reenable the
               old item. */
            [[p2input itemAtIndex:tag + 1] setEnabled:NO];
            [[p2input itemAtIndex:[p1curinput tag] + 1] setEnabled:YES];

            /* Set the old input source to the off state, and the new one to
               the on state. */
            [p1curinput setState:NSOffState];
            [item setState:NSOnState];
            p1curinput = item;
        }
    }

    /* We're done, unlock the memory lock. */
    [memoryLock unlock];
}

- (IBAction)setInput2:(id)sender
{
    NSMenuItem *item = (NSMenuItem *)sender;
    int tag;
    NSString *str;
    gamepad_t *old = joy2;

    /* If we have tried to select the same input device we currently are using,
       ignore it. */
    if([item state] == NSOnState) {
        return;
    }

    /* Lock the memory lock, and grab the tag of the menu item. */
    [memoryLock lock];
    tag = [item tag];

    /* If the user has selected the keyboard, close the old gamepad if any,
       and release the old button mapping. */
    if(tag == -1) {
        if(old) {
            /* Release resources associated with the gamepad. */
            gp_close(old);
            [j2map release];
            j2map = nil;
            joy2 = NULL;

            /* Reenable the old source on player 1's menu. */
            [[p1input itemAtIndex:[p2curinput tag] + 1] setEnabled:YES];

            /* Deselect the old input source. */
            [p2curinput setState:NSOffState];

            /* Select the keyboard in the menu. */
            [p2keyboard setState:NSOnState];
            p2curinput = p2keyboard;
        }
    }
    else {
        /* Grab the specified gamepad. */
        joy2 = gp_get_gamepad(tag);

        /* Attempt to open the specified gamepad. */
        if(gp_open(joy2, &gp_cb)) {
            /* The gamepad has not been opened, show an error and keep the old
               selected input source. */
            NSString *title = [NSString stringWithFormat:@"Cannot open "
                "gamepad %s.", joy2->name];
            NSRunAlertPanel(title, @"CrabEmu cannot open the specified "
                            "gamepad. Verify that the gamepad is attached "
                            "to the system and is powered on (if applicable).", 
                            @"OK", NULL, NULL);

            joy2 = old;
        }
        else {
            /* Grab the button mapping for the specified gamepad. */
            if(j2map != nil) {
                [j2map release];
            }

            str = [NSString stringWithUTF8String:joy2->name];
            j2map = (NSMutableDictionary *)[[keymap objectForKey:str] retain];

            /* Close the old gamepad, if any. */
            if(old) {
                gp_close(old);
            }

            /* Disable the gamepad on the first player's menu. Reenable the
               old item. */
            [[p1input itemAtIndex:tag + 1] setEnabled:NO];
            [[p1input itemAtIndex:[p2curinput tag] + 1] setEnabled:YES];

            /* Set the old input source to the off state, and the new one to
               the on state. */
            [p2curinput setState:NSOffState];
            [item setState:NSOnState];
            p2curinput = item;
        }
    }

    /* We're done, unlock the memory lock. */
    [memoryLock unlock];    
}

- (IBAction)takeScreenshot:(id)sender
{
    NSSavePanel *panel;
    NSBitmapImageRep *img;
    int w, h, i, j, offh, offw, idx;
    unsigned char *data;

    if(!sms_initialized && !coleco_initialized)
        return;

    [memoryLock lock];

    panel = [NSSavePanel savePanel];
    [panel setTitle:@"Save Screenshot As"];
    [panel setRequiredFileType:@"png"];

    if([panel runModal] != NSOKButton) {
        [memoryLock unlock];
        return;
    }

    if(sms_console == CONSOLE_GG) {
        w = 160;
        h = 144;
        offh = 24;
        offw = 48;
    }
    else {
        w = 256;
        h = smsvdp.lines;
        offh = offw = 0;
    }

    w = sms_console == CONSOLE_GG ? 160 : 256;
    h = sms_console == CONSOLE_GG ? 144 : smsvdp.lines;

    img = [[NSBitmapImageRep alloc]
            initWithBitmapDataPlanes:NULL
                          pixelsWide:w
                          pixelsHigh:h
                       bitsPerSample:8
                     samplesPerPixel:3
                            hasAlpha:NO
                            isPlanar:NO
                      colorSpaceName:NSDeviceRGBColorSpace
                         bytesPerRow:0
                        bitsPerPixel:32];

    if(img == nil) {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot create a NSBitmapImageRep"
                        "for taking a screenshot. Please file a bug report.",
                        @"OK", NULL, NULL);
        [memoryLock unlock];
        return;
    }

    data = [img bitmapData];

    for(i = 0; i < h; ++i) {
        for(j = 0; j < w; ++j) {
            idx = (i + offh) * 256 + j + offw;

            data[(i * w + j) * 4] = (smsvdp.framebuffer[idx] >> 16) & 0xFF;
            data[(i * w + j) * 4 + 1] = (smsvdp.framebuffer[idx] >> 8) & 0xFF;
            data[(i * w + j) * 4 + 2] = smsvdp.framebuffer[idx] & 0xFF;
            data[(i * w + j) * 4 + 3] = 0xFF;
        }
    }

    if(![[img representationUsingType:NSPNGFileType
                           properties:NULL] writeToFile:[panel filename]
                                             atomically:NO]) {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot open this file for writing. "
                        "Verify that you have the permissions to write to"
                        "this directory, and this file.", 
                        @"OK", NULL, NULL);
    }

    [memoryLock unlock];
}

- (IBAction)toggleFullscreen:(id)sender
{
    if(!sms_initialized && !coleco_initialized && !nes_initialized)
        return;

    [theview setKeepAspect:[preserveAspect state] == NSOnState];
    [theview toggleFullscreen];
}

- (IBAction)showVDPPalette:(id)sender
{
    if(!sms_initialized && !nes_initialized)
        return;

    if([showVDPPalette state] == NSOnState) {
        [showVDPPalette setState:NSOffState];
        [vdpPalettePanel orderOut:self];
    }
    else {
        [showVDPPalette setState:NSOnState];
        [vdpPalettePanel orderFront:self];
    }
}

- (IBAction)showVDPPatterns:(id)sender
{
    if(!sms_initialized)
        return;

    if([showVDPPatterns state] == NSOnState) {
        [showVDPPatterns setState:NSOffState];
        [vdpPatternPanel orderOut:self];
    }
    else {
        [showVDPPatterns setState:NSOnState];
        [vdpPatternPanel orderFront:self];
    }
}

- (IBAction)toggleCheats:(id)sender
{
    if([enableCheats state] == NSOnState) {
        [enableCheats setState:NSOffState];
        sms_cheat_disable();
    }
    else {
        [enableCheats setState:NSOnState];
        sms_cheat_enable();
    }
}

- (IBAction)editCheats:(id)sender
{
    [cheatController fillTable];
    [parPanel makeKeyAndOrderFront:self];
}

- (IBAction)showSDSCTerminal:(id)sender
{
    [sdscWindow orderFront:self];
}

- (IBAction)showNESPatterns:(id)sender
{
    if(!nes_initialized)
        return;

    if([showNESPatterns state] == NSOnState) {
        [showNESPatterns setState:NSOffState];
        [nesPatternPanel orderOut:self];
    }
    else {
        [showNESPatterns setState:NSOnState];
        [nesPatternPanel orderFront:self];
    }
}

- (IBAction)changeRenderer:(id)sender
{
    int tag = [sender tag];

    if([sender state] == NSOnState) {
        return;
    }

    switch(tag) {
        case 0:
            [normalRenderer setState:NSOnState];
            [hq2xRenderer setState:NSOffState];
            [hq4xRenderer setState:NSOffState];
            [scale2xRenderer setState:NSOffState];
            break;

        case 1:
            [normalRenderer setState:NSOffState];
            [hq2xRenderer setState:NSOffState];
            [hq4xRenderer setState:NSOffState];
            [scale2xRenderer setState:NSOnState];
            break;

        case 2:
            [normalRenderer setState:NSOffState];
            [hq2xRenderer setState:NSOnState];
            [hq4xRenderer setState:NSOffState];
            [scale2xRenderer setState:NSOffState];
            break;

        case 3:
            [normalRenderer setState:NSOffState];
            [hq2xRenderer setState:NSOffState];
            [hq4xRenderer setState:NSOnState];
            [scale2xRenderer setState:NSOffState];
            break;
    }

    [theview setRenderer:tag];
}

- (BOOL)initMasterSystem:(int)console
{
    int i;
    uint8 ch = 0x00;

    if(sms_init(ntscVideo == YES ? SMS_VIDEO_NTSC : SMS_VIDEO_PAL,
                domestic == YES ? SMS_REGION_DOMESTIC : SMS_REGION_EXPORT)) {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot initialize the Master "
                        "System emulation. Please report this error!", 
                        @"OK", NULL, NULL);
        return NO;
    }

    /* Restore the state of the audio channels to what the menu says they
       should be. */
    if([tone0 state] == NSOnState)
        ch |= ENABLE_TONE0;
    if([tone1 state] == NSOnState)
        ch |= ENABLE_TONE1;
    if([tone2 state] == NSOnState)
        ch |= ENABLE_TONE2;
    if([noise state] == NSOnState)
        ch |= ENABLE_NOISE;

    psg.enabled_channels = ch;

    if(console == CONSOLE_SMS && [smsBiosEnabled state] == NSOnState &&
       [[smsBios stringValue] length] > 0) {
        if(sms_mem_load_bios([[smsBios stringValue] UTF8String])) {
            NSString *s = [NSString stringWithFormat:@"CrabEmu cannot open\n%@"
                           @"\nPlease check that the file exists and that you\n"
                           @"have the appropriate permissions to access it.",
                           [smsBios stringValue]];
            NSRunAlertPanel(@"Cannot open BIOS ROM file", s, @"OK", NULL, NULL);
            sms_shutdown();
            return NO;
        }
    }
    else if(console == CONSOLE_GG && [ggBiosEnabled state] == NSOnState &&
            [[ggBios stringValue] length] > 0) {
        if(sms_mem_load_bios([[ggBios stringValue] UTF8String])) {
            NSString *s = [NSString stringWithFormat:@"CrabEmu cannot open\n%@"
                           @"\nPlease check that the file exists and that you\n"
                           @"have the appropriate permissions to access it.",
                           [ggBios stringValue]];
            NSRunAlertPanel(@"Cannot open BIOS ROM file", s, @"OK", NULL, NULL);
            sms_shutdown();
            return NO;
        }
    }

    [theview showWindow];

    /* Enable everything that needs to be enabled while a ROM is running... */
    [screenshot setEnabled:YES];

    if(CGLLockContext != NULL) {
        [fullscreen setEnabled:YES];
    }

    [saveSRAM setEnabled:YES];

    for(i = 0; i < [saveStates numberOfItems]; ++i) {
        [[saveStates itemAtIndex:i] setEnabled:YES];
    }
    
    for(i = 0; i < [loadStates numberOfItems]; ++i) {
        [[loadStates itemAtIndex:i] setEnabled:YES];
    }

    [pause setEnabled:YES];
    [reset setEnabled:YES];
    [enableCheats setEnabled:YES];
    [editCheats setEnabled:YES];
    [debugger setEnabled:YES];
    [showVDPPalette setEnabled:YES];
    [showVDPPatterns setEnabled:YES];

    /* ...and disable those settings we don't want available while a ROM is
       running. */
    [ntsc setEnabled:NO];
    [pal setEnabled:NO];
    [rExport setEnabled:NO];
    [rDomestic setEnabled:NO];

    return YES;
}

- (BOOL)loadMasterSystemRomFromFile:(NSString *)filename console:(int)c
{
    NSString *sfn;

    [openRom release];
    [openRomSaveFn release];
    [cheatFn release];

    openRom = [[filename lastPathComponent] retain];

    /* Load up the Rom itself. */
    if(sms_mem_load_rom([filename UTF8String], c)) {
        [openRom release];
        openRomSaveFn = nil;
        openRom = nil;

        return NO;
    }

    /* Load up any saves for this rom. */
    sfn = [openRom stringByAppendingPathExtension:@"ces"];
    openRomSaveFn = [[saveDir stringValue] stringByAppendingPathComponent:sfn];
    [openRomSaveFn retain];

    sms_read_cartram_from_file([openRomSaveFn UTF8String]);

    /* Load up any cheats for this rom. */
    sfn = [openRom stringByAppendingPathExtension:@"cht"];
    cheatFn = [[cheatDir stringValue] stringByAppendingPathComponent:sfn];
    [cheatFn retain];

    sms_cheat_read([cheatFn UTF8String]);

    if(sms_console == CONSOLE_SMS && [[smsBios stringValue] length] > 0 &&
       [smsBiosEnabled state] == NSOnState) {
        sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                              SMS_MEMCTL_RAM) ^ 0xFF);
    }
    else if(sms_console == CONSOLE_GG && [[ggBios stringValue] length] > 0 &&
            [ggBiosEnabled state] == NSOnState) {
        sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                 SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
    }

    return YES;
}

- (BOOL)initColecoVision
{
    uint8 ch = 0x00;
    int i;

    if(coleco_init(ntscVideo == YES ? SMS_VIDEO_NTSC : SMS_VIDEO_PAL)) {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot initialize the Coleco"
                        "Vision emulation. Please report this error!", 
                        @"OK", NULL, NULL);
        return NO;
    }

    /* Restore the state of the audio channels to what the menu says they
       should be. */
    if([tone0 state] == NSOnState)
        ch |= ENABLE_TONE0;
    if([tone1 state] == NSOnState)
        ch |= ENABLE_TONE1;
    if([tone2 state] == NSOnState)
        ch |= ENABLE_TONE2;
    if([noise state] == NSOnState)
        ch |= ENABLE_NOISE;

    psg.enabled_channels = ch;

    if(coleco_mem_load_bios([[cvBios stringValue] UTF8String])) {
        NSString *s = [NSString stringWithFormat:@"CrabEmu cannot open\n%@\n"
                       @"Please check that the file exists and that you\n"
                       "have the appropriate permissions to access it.",
                       [cvBios stringValue]];
        NSRunAlertPanel(@"Cannot open BIOS ROM file", s, @"OK", NULL, NULL);
        coleco_shutdown();
        return NO;
    }

    [theview showWindow];

    /* Enable everything that needs to be enabled while a ROM is running... */
    [screenshot setEnabled:YES];

    if(CGLLockContext != NULL) {
        [fullscreen setEnabled:YES];
    }

    for(i = 0; i < [saveStates numberOfItems]; ++i) {
        [[saveStates itemAtIndex:i] setEnabled:YES];
    }
    
    for(i = 0; i < [loadStates numberOfItems]; ++i) {
        [[loadStates itemAtIndex:i] setEnabled:YES];
    }

    [pause setEnabled:YES];
    [reset setEnabled:YES];

    /* ...and disable those settings we don't want available while a ROM is
       running. */
    [ntsc setEnabled:NO];
    [pal setEnabled:NO];
    [rExport setEnabled:NO];
    [rDomestic setEnabled:NO];

    return YES;
}

- (BOOL)loadColecoVisionRomFromFile:(NSString *)filename
{
    [openRom release];

    openRom = [[filename lastPathComponent] retain];

    /* Load up the Rom itself. */
    if(coleco_mem_load_rom([filename UTF8String])) {
        [openRom release];
        openRomSaveFn = nil;
        openRom = nil;

        return NO;
    }

    return YES;
}

- (BOOL)initNES
{
    int i;

    if(nes_init(ntscVideo == YES ? VIDEO_NTSC : VIDEO_PAL)) {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot initialize the NES "
                        "emulation. Please report this error!", 
                        @"OK", NULL, NULL);
        return NO;
    }

    [theview showWindow];

    /* Enable everything that needs to be enabled while a ROM is running... */
    [screenshot setEnabled:YES];

    if(CGLLockContext != NULL) {
        [fullscreen setEnabled:YES];
    }

    [saveSRAM setEnabled:YES];

    for(i = 0; i < [saveStates numberOfItems]; ++i) {
        [[saveStates itemAtIndex:i] setEnabled:YES];
    }
    
    for(i = 0; i < [loadStates numberOfItems]; ++i) {
        [[loadStates itemAtIndex:i] setEnabled:YES];
    }

    [pause setEnabled:YES];
    [reset setEnabled:YES];
    [showVDPPalette setEnabled:YES];
    [showNESPatterns setEnabled:YES];

    /* ...and disable those settings we don't want available while a ROM is
     running. */
    [ntsc setEnabled:NO];
    [pal setEnabled:NO];
    [rExport setEnabled:NO];
    [rDomestic setEnabled:NO];

    return YES;
}

- (BOOL)loadNESRomFromFile:(NSString *)filename
{
    NSString *sfn;

    [openRom release];
    [openRomSaveFn release];

    openRom = [[filename lastPathComponent] retain];

    /* Load up the Rom itself. */
    if(nes_mem_load_rom([filename UTF8String])) {
        [openRom release];
        openRomSaveFn = nil;
        openRom = nil;

        return NO;
    }

    /* Load up any saves for this rom. */
    sfn = [openRom stringByAppendingPathExtension:@"ces"];
    openRomSaveFn = [[saveDir stringValue] stringByAppendingPathComponent:sfn];
    [openRomSaveFn retain];

    nes_mem_read_sram([openRomSaveFn UTF8String]);

    return YES;
}

- (BOOL)openRomFromFile:(NSString *)fn
{
    int console;
    NSDocumentController *c;

    console = rom_detect_console([fn fileSystemRepresentation]);

    /* Are we dealing with a ColecoVision rom? */
    if(console == CONSOLE_COLECOVISION) {
        /* Make sure the user has a bios configured. */
        if([[cvBios stringValue] length] == 0) {
            NSRunAlertPanel(@"Error", @"You must configure a ColecoVision BIOS "
                            "Rom before opening a ColecoVision game.", @"OK",
                            nil, nil);
            return NO;
        }
    }
    else if(console < 0) {
        NSRunAlertPanel(@"Error", @"CrabEmu cannot detect the console for "
                        "the rom selected!", @"OK", nil, nil);
        return NO;
    }

    if(sms_initialized == 1) {
        sms_write_cartram_to_file([openRomSaveFn UTF8String]);
        sms_cheat_write([cheatFn UTF8String]);
    }
    else if(nes_initialized) {
        nes_mem_write_sram([openRomSaveFn UTF8String]);
    }

    running = NO;
    started = 0;

    /* This is just to make sure we actually finish with the thread... */
    [runningLock lock];
    [runningLock unlock];
    [memoryLock lock];

    if(console == CONSOLE_COLECOVISION) {
        if([self initColecoVision] == NO) {
            [memoryLock unlock];
            return NO;
        }

        if([self loadColecoVisionRomFromFile:fn] == NO) {
            NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
                               " file %@.", fn];
            NSRunAlertPanel(title, @"CrabEmu cannot open this file for "
                            "reading. Verify that you have the permissions to "
                            "read from this directory, and this file.", 
                            @"OK", NULL, NULL);

            [memoryLock unlock];

            return NO;
        }
    }
    else if(console == CONSOLE_NES) {
        if([self initNES] == NO) {
            [memoryLock unlock];
            return NO;
        }

        if([self loadNESRomFromFile:fn] == NO) {
            NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
                               " file %@.", fn];
            NSRunAlertPanel(title, @"CrabEmu cannot open this file for "
                            "reading. Verify that you have the permissions to "
                            "read from this directory, and this file.", 
                            @"OK", NULL, NULL);

            [memoryLock unlock];

            return NO;
        }
    }
    else {
        if([self initMasterSystem:console] == NO) {
            [memoryLock unlock];
            return NO;
        }

        if([self loadMasterSystemRomFromFile:fn console:console] == NO) {
            NSString *title = [NSString stringWithFormat:@"Cannot open ROM"
                               " file %@.", fn];
            NSRunAlertPanel(title, @"CrabEmu cannot open this file for "
                            "reading. Verify that you have the permissions to "
                            "read from this directory, and this file.", 
                            @"OK", NULL, NULL);

            [memoryLock unlock];

            return NO;
        }
    }

    sound_unpause();
    [memoryLock unlock];

    if(!running) {
        started = 1;
        running = YES;
        
        [NSThread detachNewThreadSelector:@selector(masterEmuThread:)
                                 toTarget:self
                               withObject:nil];
    }

    c = [NSDocumentController sharedDocumentController];
    [c noteNewRecentDocumentURL:[NSURL fileURLWithPath:fn]];

    return YES;
}

- (void)toggleScreensaver:(BOOL)on
{
    if(on) {
        if(timer == nil) {
            /* The user has newly selected to disable the screensaver, start
               the timer. */
            timer = [NSTimer scheduledTimerWithTimeInterval:15
                                                     target:self
                                                   selector:@selector(ssTimer:)
                                                   userInfo:nil
                                                    repeats:YES];
        }
    }
    else {
        if(timer != nil) {
            /* The user has selected to not disable the screensaver, invalidate
               the existing timer, and set it to nil. The run loop will
               release the object for us. */
            [timer invalidate];
            timer = nil;
        }
    }
}

- (void)savePrefs
{
    [prefs setBool:!ntscVideo forKey:@"SMSPAL"];
    [prefs setBool:domestic forKey:@"SMSDomestic"];

    [prefs synchronize];
}

- (void)readPrefs
{
    BOOL palMode, disableSaver;

    palMode = [prefs boolForKey:@"SMSPAL"];
    domestic = [prefs boolForKey:@"SMSDomestic"];
    disableSaver = [prefs boolForKey:@"DisableScreensaver"];

    [keymap release];

    /* Read the controller configuration */
    keymap = [NSMutableDictionary dictionaryWithDictionary:
              [prefs dictionaryForKey:@"SMS Buttonmap"]];
    [keymap retain];

    if(palMode) {
        [pal setState:NSOnState];
        [ntsc setState:NSOffState];
        ntscVideo = NO;
    }

    if(domestic) {
        [rDomestic setState:NSOnState];
        [rExport setState:NSOffState];
        domestic = YES;
    }

    if(disableSaver) {
        if(timer == nil) {
            timer = [NSTimer scheduledTimerWithTimeInterval:15
                                                     target:self
                                                   selector:@selector(ssTimer:)
                                                   userInfo:nil
                                                    repeats:YES];
        }
    }
}

- (void)ssTimer:(NSTimer *)timer
{
    UpdateSystemActivity(UsrActivity);
}

- (BOOL)running
{
    return running;
}

@end
