/*
	This file is part of CrabEmu.

	Copyright (C) 2005, 2006, 2007, 2008, 2009, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "EmuOGLView.h"
#include "EmuController.h"

#include "sms.h"
#include "colecovision.h"
#include "nes.h"
#include "smsvdp.h"
#include "nesppu.h"
#include "terebi.h"
#include "pastlib.h"
#include "scale2x.h"

extern EmuOGLView *theview;
extern int sms_initialized;
extern int coleco_initialized;
extern int nes_initialized;
extern int sms_console;

@implementation EmuOGLView

- (id)initWithFrame:(NSRect)frameRect
{
    CGLContextObj cxt = CGLGetCurrentContext();

    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAWindow,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        0
    };

    if(CGLLockContext != NULL) {
        CGLLockContext(cxt);
    }

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];

    if(fmt == nil) {
        if(CGLUnlockContext != NULL) {
            CGLUnlockContext(cxt);
        }

        [fmt release];
        return nil;
    }

    if((self = [super initWithFrame:frameRect pixelFormat:fmt]) != nil) {
        GLint swapInterval = 1;
        theview = self;
        [[self openGLContext] setValues:&swapInterval
                           forParameter:NSOpenGLCPSwapInterval];
    }
    else {
        if(CGLUnlockContext != NULL) {
            CGLUnlockContext(cxt);
        }

        [fmt release];
        return nil;
    }

    /* Set up the full screen context. */
    disp = [self screenID];
    [self createFullscreenContext];

    resized = NO;
    isFullscreen = NO;
    keepAspect = NO;
    useHQ2x = NO;
    useScale2x = NO;
    useHQ4x = NO;
    txrChanged = YES;
    texwidth = 256;
    texheight = 256;
    fbTrans = NULL;

    if(CGLUnlockContext != NULL) {
        CGLUnlockContext(cxt);
    }

    [fmt release];

    return self;
}

- (void)dealloc
{
    if(useHQ2x || useHQ4x || useScale2x) {
        free(fbTrans);
    }

    [fsCxt release];
    [super dealloc];
}

- (void)renderScene
{
    if(txrChanged) {
        if(!sms_initialized && !coleco_initialized && !nes_initialized)
            return;

        if(glInitted) {
            glDeleteTextures(1, &glTexture);
        }
        else {
            glInitted = YES;
            resized = YES;
        }

        glGenTextures(1, &glTexture);
        glBindTexture(GL_TEXTURE_2D, glTexture);
        glTexImage2D(GL_TEXTURE_2D, 0, 4, texwidth, texheight, 0, GL_BGRA,
                     GL_UNSIGNED_SHORT_1_5_5_5_REV, NULL);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    }

    if(resized) {
        glShadeModel(GL_SMOOTH);
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        glClearDepth(1.0f);

        if(isFullscreen && keepAspect) {
            float sysw, sysh, ratio, x, y;

            if(sms_console == CONSOLE_NES) {
                sysw = 256.0f;
                sysh = 240.0f;
            }
            else if(sms_console == CONSOLE_GG) {
                sysw = 160.0f;
                sysh = 144.0f;
            }
            else {
                sysw = 256.0f;
                sysh = (float)smsvdp.lines;
            }

            if([self width] / sysw < [self height] / sysh) {
                ratio = [self width] / sysw;
                x = 0.0f;
                y = ([self height] / 2.0f) - (sysh * ratio / 2.0f);
            }
            else {
                ratio = [self height] / sysh;
                x = ([self width] / 2.0f) - (sysw * ratio / 2.0f);
                y = 0.0f;
            }

            glViewport(x, y, sysw * ratio, sysh * ratio);
        }
        else {
            glViewport(0.0f, 0.0f, [self width], [self height]);
        }

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        if(sms_console == CONSOLE_NES)
            glOrtho(0.0f, 255.0f, 239.0f, 0.0f, 1.0f, 0.0f);
        else if(sms_console != CONSOLE_GG)
            glOrtho(0.0f, 255.0f, smsvdp.lines - 1, 0.0f, 1.0f, 0.0f);
        else
            glOrtho(0.0f, 159.0f, 143.0f, 0.0f, 1.0f, 0.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        resized = NO;
    }

    if(sms_console != CONSOLE_NES) {
        if(useHQ2x) {
            ProcessHQ2x((pixel *)smsvdp.framebuffer, (pixel *)fbTrans,
                        smsvdp.lines);
        }
        else if(useHQ4x) {
            ProcessHQ4x((pixel *)smsvdp.framebuffer, (pixel *)fbTrans,
                        smsvdp.lines);
        }
        else if(useScale2x) {
            scale2x_scale(smsvdp.framebuffer, fbTrans, 256, smsvdp.lines);
        }
    }
    else {
        if(useHQ2x) {
            ProcessHQ2x((pixel *)nes_ppu_framebuffer(), (pixel *)fbTrans, 240);
        }
        else if(useHQ4x) {
            ProcessHQ4x((pixel *)nes_ppu_framebuffer(), (pixel *)fbTrans, 240);
        }
        else if(useScale2x) {
            scale2x_scale(nes_ppu_framebuffer(), fbTrans, 256, 240);
        }
    }

    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();

    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, glTexture);

    if(fbTrans) {
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texwidth, texheight, GL_BGRA,
                        GL_UNSIGNED_SHORT_1_5_5_5_REV, fbTrans);
    }
    else if(sms_console != CONSOLE_NES) {
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texwidth, texheight, GL_BGRA,
                        GL_UNSIGNED_SHORT_1_5_5_5_REV, smsvdp.framebuffer);
    }
    else {
        glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, texwidth, texheight, GL_BGRA,
                        GL_UNSIGNED_SHORT_1_5_5_5_REV, nes_ppu_framebuffer());
    }

    if(sms_console == CONSOLE_NES) {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(0.0f, 240.0f / 256.0f);
        glVertex2f(0.0f, 239.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(1.0f, 240.0f / 256.0f);
        glVertex2f(255.0f, 239.0f);
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(255.0f, 0.0f);
        glEnd();
    }
    else if(sms_console != CONSOLE_GG) {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(0.0f, smsvdp.lines / 256.0f);
        glVertex2f(0.0f, smsvdp.lines - 1);
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(1.0f, smsvdp.lines / 256.0f);
        glVertex2f(255.0f, smsvdp.lines - 1);
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(255.0f, 0.0f);
        glEnd();
    }
    else {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(48.0f / 256.0f, 168.0f / 256.0f);
        glVertex2f(0.0f, 143.0f);
        glTexCoord2f(48.0f / 256.0f, 24.0f / 256.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(208.0f / 256.0f, 168.0f / 256.0f);
        glVertex2f(159.0f, 143.0f);
        glTexCoord2f(208.0f / 256.0f, 24.0f / 256.0f);
        glVertex2f(159.0f, 0.0f);
        glEnd();
    }

    glFlush();
}

- (void)drawScene
{
    CGLContextObj cxt = CGLGetCurrentContext();
    static BOOL wasFullscreen = NO;
    static int count = 2;

    if(CGLLockContext != NULL) {
        CGLLockContext(cxt);
    }

    if(resized) {
        glViewport(0.0f, 0.0f, [self width], [self height]);
    }

    if((!wasFullscreen && isFullscreen) ||
       (wasFullscreen && !isFullscreen)) {
        resized = YES;

        /* UGLY UGLY UGLY HACK. For some reason, fullscreen mode seems to need
           some extra encouragement when switching into it sometimes. Make sure
           that the resize functionality gets called at least a few times with
           the new sizes set up. */
        if(!(--count)) {
            wasFullscreen = isFullscreen;
            count = 2;
        }
    }

    [self renderScene];

    if(!isFullscreen) {
        [[self openGLContext] flushBuffer];
    }
    else {
        [fsCxt flushBuffer];
    }

    if(CGLUnlockContext != NULL) {
        CGLUnlockContext(cxt);
    }

    if(!isFullscreen || [[NSScreen screens] count] > 1) {
        [controller updatePalette];
        [controller updatePatterns];
    }
}

- (void)drawRect:(NSRect)rect
{
    [self drawScene];
}

- (void)setNeedsDisplay:(BOOL)display
{
    if(isFullscreen) {
        if(display) {
            [fsCxt makeCurrentContext];
            [self drawScene];
        }
    }
    else {
        [super setNeedsDisplay:display];
    }
}

- (BOOL)acceptsFirstResponder
{
    return YES;
}

- (void)showWindow
{
    [window makeKeyAndOrderFront:self];
}

- (void)keyDown:(NSEvent *)event
{
    unichar buttonmap[32];
    NSString *str = [event charactersIgnoringModifiers];
    unichar character = [str characterAtIndex:0];
    int len, i;
    void (*bfunc)(int, int);

    /* Which button function do we need... */
    switch(sms_console) {
        case CONSOLE_SMS:
        case CONSOLE_GG:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
            bfunc = &sms_button_pressed;
            break;

        case CONSOLE_COLECOVISION:
            bfunc = &coleco_button_pressed;
            break;

        case CONSOLE_NES:
            bfunc = &nes_button_pressed;
            break;

        default:
            return;
    }

    str = [controller keyListForConsole:sms_console player:1];
    len = (int)[str length];

    if(len > 0 && len <= 32) {
        [str getCharacters:buttonmap range:NSMakeRange(0, len)];

        for(i = 0; i < len; ++i) {
            if(character == buttonmap[i]) {
                bfunc(1, i);
                return;
            }
        }
    }

    str = [controller keyListForConsole:sms_console player:2];
    len = (int)[str length];
    
    if(len > 0 && len <= 32) {
        [str getCharacters:buttonmap range:NSMakeRange(0, len)];

        for(i = 0; i < len; ++i) {
            if(character == buttonmap[i]) {
                bfunc(2, i);
                return;
            }
        }
    }
}

- (void)keyUp:(NSEvent *)event
{
    unichar buttonmap[32];
    NSString *str = [event charactersIgnoringModifiers];
    unichar character = [str characterAtIndex:0];
    int len, i;
    void (*bfunc)(int, int);

    /* Which button function do we need... */
    switch(sms_console) {
        case CONSOLE_SMS:
        case CONSOLE_GG:
        case CONSOLE_SG1000:
        case CONSOLE_SC3000:
            bfunc = &sms_button_released;
            break;
            
        case CONSOLE_COLECOVISION:
            bfunc = &coleco_button_released;
            break;

        case CONSOLE_NES:
            bfunc = &nes_button_released;
            break;

        default:
            return;
    }

    str = [controller keyListForConsole:sms_console player:1];
    len = (int)[str length];

    if(len > 0 && len <= 32) {
        [str getCharacters:buttonmap range:NSMakeRange(0, len)];

        for(i = 0; i < len; ++i) {
            if(character == buttonmap[i]) {
                bfunc(1, i);
                return;
            }
        }
    }

    str = [controller keyListForConsole:sms_console player:2];
    len = (int)[str length];

    if(len > 0 && len <= 32) {
        [str getCharacters:buttonmap range:NSMakeRange(0, len)];

        for(i = 0; i < len; ++i) {
            if(character == buttonmap[i]) {
                bfunc(2, i);
                return;
            }
        }
    }
}

- (void)mouseDown:(NSEvent *)event
{
    NSPoint loc = [self convertPoint:[event locationInWindow]
                            fromView:nil];
    NSRect bounds = [self bounds];
    int x, y;

    /* Scale the position to [0.0, 1.0], and multiply by the SG-1000's
       resolution. */
    x = (int)((loc.x / bounds.size.width) * 256);
    y = (int)((loc.y / bounds.size.height) * 192);

    /* Mac OS X takes positions from the lower-left corner, we want it from the
       upper left. Flip the y value so that we get that. */
    y = (192 - y) > 0 ? 192 - y : 0;

    terebi_update(x, y, 1);
}

- (void)mouseUp:(NSEvent *)event
{
    terebi_update(0, 0, 0);
}

- (void)mouseDragged:(NSEvent *)event
{
    [self mouseDown:event];
}

- (float)width
{
    if(!isFullscreen) {
        return [self bounds].size.width;
    }
    else {
        return [[NSScreen mainScreen] frame].size.width;
    }
}

- (float)height
{
    if(!isFullscreen) {
        return [self bounds].size.height;
    }
    else {
        return [[NSScreen mainScreen] frame].size.height;
    }
}

- (void)setAspectRatioX:(float)x Y:(float)y
{
    float aspect = y / x;
    float nh = [self width] * aspect;
    float diff = nh - [self height];
    NSRect frame = [window frame];

    frame.size.height += diff;
    frame.origin.y -= diff;

    [window setContentAspectRatio:NSMakeSize(x, y)];
    [window setFrame:frame display:YES];
    resized = YES;
}

- (void)update
{
    /* Set a flag here, and actually handle the resizing in drawRect, since
       the bounds have not changed yet when update is called. */
    resized = YES;

    [super update];
}

- (void)setTitle:(NSString *)str
{
    [window setTitle:str];
}

- (BOOL)isFullscreen
{
    return isFullscreen;
}

- (void)toggleFullscreen
{
    CGError err;
    CGDisplayFadeReservationToken token;

    /* This probably will crash on Mac OS X 10.3, so make sure not to crash by
       not allowing fullscreen mode there. */
    if(CGLLockContext == NULL || CGLUnlockContext == NULL)
        return;

    err = CGAcquireDisplayFadeReservation(kCGMaxDisplayReservationInterval,
                                          &token);

    if(err == kCGErrorSuccess) {
        CGDisplayFade(token, 0.5, kCGDisplayBlendNormal,
                      kCGDisplayBlendSolidColor, 0, 0, 0, 1);
    }

    if(isFullscreen) {
        CGPoint mousePoint;
        int height = CGDisplayPixelsHigh(disp);

        mousePoint.x = mouseLoc.x;
        mousePoint.y = height - mouseLoc.y;

        [fsCxt clearDrawable];
        [[self openGLContext] makeCurrentContext];

        /* Show the mouse pointer, and reassociate it with the mouse. */
        CGAssociateMouseAndMouseCursorPosition(TRUE);
        CGDisplayMoveCursorToPoint(disp, mousePoint);
        CGDisplayShowCursor(disp);

        CGReleaseAllDisplays();
    }
    else {
        CGDirectDisplayID d = [self screenID];

        if(d != disp) {
            [fsCxt release];
            disp = d;
            [self createFullscreenContext];
        }

        [fsCxt clearDrawable];
        CGDisplayCapture(disp);

        /* Hide the cursor, but store its location so we can restore it later.
           Also, disassociate the mouse and the cursor position. */
        CGDisplayHideCursor(disp);
        mouseLoc = [NSEvent mouseLocation];
        CGDisplayMoveCursorToPoint(disp, CGPointZero);
        CGAssociateMouseAndMouseCursorPosition(FALSE);

        [fsCxt setFullScreen];
        [fsCxt makeCurrentContext];
    }

    if(err == kCGErrorSuccess) {
        CGDisplayFade(token, 0.5, kCGDisplayBlendNormal,
                      kCGDisplayBlendSolidColor, 0, 0, 0, 0);
        CGReleaseDisplayFadeReservation(token);
    }

    isFullscreen = !isFullscreen;
}

- (CGDirectDisplayID)screenID
{
    NSArray *screens = [NSScreen screens];
    NSEnumerator *i = [screens objectEnumerator];
    NSScreen *obj;
    NSRect f = [window frame];
    NSRect sf;

    while((obj = (NSScreen *)[i nextObject])) {
        sf = [obj frame];

        if(f.origin.x >= sf.origin.x && f.origin.y >= sf.origin.y &&
           f.origin.x <= sf.origin.x + sf.size.width &&
           f.origin.y <= sf.origin.y + sf.size.height) {
            NSDictionary *d = [obj deviceDescription];
            NSNumber *n = (NSNumber *)[d objectForKey:@"NSScreenNumber"];

            return (CGDirectDisplayID)[n unsignedIntValue];
        }
    }

    /* Punt. */
    return kCGDirectMainDisplay;
}

- (void)createFullscreenContext
{
    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAFullScreen,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        NSOpenGLPFAScreenMask,
        CGDisplayIDToOpenGLDisplayMask(disp),
        0
    };

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];

    if(fmt != nil) {
        GLint sync = 1;

        fsCxt = [[NSOpenGLContext alloc] initWithFormat:fmt
                                           shareContext:[self openGLContext]];
        [fmt release];
        [fsCxt setValues:&sync forParameter:NSOpenGLCPSwapInterval];
    }
    else {
        fsCxt = nil;
        [fmt release];
    }
}

- (void)setKeepAspect:(BOOL)ka
{
    keepAspect = ka;
}

- (void)setRenderer:(int)renderer
{
    if(useHQ2x || useHQ4x || useScale2x) {
        free(fbTrans);
    }

    switch(renderer) {
        case 1:
            useHQ2x = NO;
            useHQ4x = NO;
            useScale2x = YES;
            txrChanged = YES;
            texwidth = texheight = 512;
            break;

        case 2:
            useHQ2x = YES;
            useHQ4x = NO;
            useScale2x = NO;
            txrChanged = YES;
            texwidth = texheight = 512;
            break;

        case 3:
            useHQ2x = NO;
            useHQ4x = YES;
            useScale2x = NO;
            txrChanged = YES;
            texwidth = texheight = 1024;
            break;

        default:
            useHQ2x = NO;
            useHQ4x = NO;
            useScale2x = NO;
            txrChanged = YES;
            fbTrans = NULL;
            texwidth = texheight = 256;
            return;
    }

    fbTrans = (pixel_t *)malloc(texwidth * texheight * sizeof(pixel_t));
}

@end
