/*
    This file is part of CrabEmu.

    Copyright (C) 2009 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <OpenGL/GL.h>

#include "CrabEmu.h"
#include "OpenGLFont.h"

static inline int next_pow2(int num) {
    int rv = 8;

    while(rv < num) {
        rv <<= 1;
    }

    return rv;
}

@interface NSFont (CrabEmu_GL_Ext)

- (GLuint *)makeOpenGLTexturesForStartChar:(unichar)start
                                     count:(int)count
                                     sizes:(NSSize **)sizes
                                       uvs:(NSPoint **)uvs;
- (GLuint)makeOpenGLTextureForChar:(unichar)ch
                              size:(NSSize *)size
                                uv:(NSPoint *)uv;

@end

@implementation NSFont (CrabEmu_GL_Ext)

- (GLuint *)makeOpenGLTexturesForStartChar:(unichar)start
                                     count:(int)count
                                     sizes:(NSSize **)sizes
                                       uvs:(NSPoint **)uvs
{
    unichar i;
    GLuint *txrs;
    NSSize *size;
    NSPoint *uv;

    /* Allocate space for the textures/sizes */
    txrs = (GLuint *)malloc(count * sizeof(GLuint));
    if(!txrs)
        return NULL;

    size = (NSSize *)malloc(count * sizeof(NSSize));
    if(!size) {
        free(txrs);
        return NULL;
    }

    uv = (NSPoint *)malloc(count * sizeof(NSPoint));
    if(!uv) {
        free(size);
        free(txrs);
        return NULL;
    }

    /* Create each texture. */
    for(i = 0; i < count; ++i) {
        txrs[i] = [self makeOpenGLTextureForChar:(i + start)
                                            size:size + i
                                              uv:uv + i];
    }

    *sizes = size;
    *uvs = uv;
    return txrs;
}

- (GLuint)makeOpenGLTextureForChar:(unichar)ch
                              size:(NSSize *)size
                                uv:(NSPoint *)uv
{
    NSImage *img;
    NSString *str = [NSString stringWithCharacters:&ch
                                            length:1];
    NSColor *bg = [NSColor colorWithDeviceRed:0.0 green:0.0 blue:0.0 alpha:0.0];
    NSDictionary *d = [NSDictionary dictionaryWithObjectsAndKeys:
        self, NSFontAttributeName,
        bg, NSBackgroundColorAttributeName,
        [NSColor whiteColor], NSForegroundColorAttributeName, nil];
    NSRect rc;
    GLuint txr = 0;
    NSData *tiff;
    NSBitmapImageRep *bmp;
    int w, h;
    int i, j, rb, pb;
    uint8 *tex, *data;

    /* Set up the rectangle for drawing, and make sure it is of an integral
       size. */
    rc.origin.x = 0.0;
    rc.origin.y = 0.0;
    rc.size = [str sizeWithAttributes:d];
    rc = NSIntegralRect(rc);
    *size = rc.size;

    /* Calculate the U/V coordinates. */
    w = next_pow2((int)rc.size.width);
    h = next_pow2((int)rc.size.height);

    uv->x = rc.size.width / w;
    uv->y = rc.size.height / h;

    rc.size.width = w;
    rc.size.height = h;

    /* Allocate our storage space. */
    tex = (uint8 *)malloc(sizeof(uint8) * 4 * w * h);
    memset(tex, 0, sizeof(uint8) * 4 * w * h);

    /* Allocate the image we'll draw into. */
    img = [[NSImage alloc] initWithSize:rc.size];

    /* Draw the image. */
    [img lockFocus];
    [bg set];
    [NSBezierPath fillRect:rc];
    [str drawInRect:rc withAttributes:d];
    [img unlockFocus];

    tiff = [img TIFFRepresentationUsingCompression:NSTIFFCompressionNone
                                            factor:0];
    bmp = [NSBitmapImageRep imageRepWithData:tiff];
    data = [bmp bitmapData];
    rb = [bmp bytesPerRow];
    pb = [bmp bitsPerPixel] / 8;

    /* Transform it to the OpenGL texture. */
    for(i = 0; i < size->height; ++i) {
        for(j = 0; j < size->width; ++j) {
            tex[(i * w + j) * 4] = data[(i * rb) + j * pb + 3];
            tex[(i * w + j) * 4 + 1] = data[(i * rb) + j * pb];
            tex[(i * w + j) * 4 + 2] = data[(i * rb) + j * pb + 1];
            tex[(i * w + j) * 4 + 3] = data[(i * rb) + j * pb + 2];
        }
    }

    /* Clean up the NSImage object. */
    [img release];

    /* Set up the texture for OpenGL. */
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &txr);
    glBindTexture(GL_TEXTURE_2D, txr);
    glTexImage2D(GL_TEXTURE_2D, 0, 4, w, h, 0, GL_BGRA,
                 GL_UNSIGNED_INT_8_8_8_8_REV, tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glDisable(GL_TEXTURE_2D);

    /* Clean up the memory used by the texture. */
    free(tex);

    return txr;
}

@end

@implementation OpenGLFont

- (id)initWithFont:(NSFont *)f
{
    if(!(self = [super init]))
        return nil;

    if(f == nil)
        return nil;

    chars = [f makeOpenGLTexturesForStartChar:32
                                        count:95
                                        sizes:&sizes
                                          uvs:&uvs];

    if(chars == NULL)
        return nil;

    font = [f retain];

    return self;
}

- (void)dealloc
{
    if(chars) {
        glDeleteTextures(95, chars);
        free(chars);
    }

    if(sizes)
        free(sizes);

    if(uvs)
        free(uvs);

    [font release];

    [super dealloc];
}

- (void)drawString:(NSString *)str at:(NSPoint)pt red:(float)r green:(float)g
              blue:(float)b alpha:(float)a
{
    int i;
    unichar ch;

    for(i = 0; i < (int)[str length]; ++i) {
        ch = [str characterAtIndex:i];
        pt = [self drawCharacter:(char)ch at:pt red:r green:g blue:b alpha:a];
    }
}

- (NSPoint)drawCharacter:(char)ch at:(NSPoint)pt red:(float)r green:(float)g
                    blue:(float)b alpha:(float)a
{
    int i;

    if(ch < 32 || ch > 126)
        return pt;

    i = ch - 32;

    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, chars[i]);

    glColor4f(r, g, b, a);

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0.0f, uvs[i].y);
    glVertex2f(pt.x, pt.y + sizes[i].height);
    glTexCoord2f(0.0f, 0.0f);
    glVertex2f(pt.x, pt.y);
    glTexCoord2f(uvs[i].x, uvs[i].y);
    glVertex2f(pt.x + sizes[i].width, pt.y + sizes[i].height);
    glTexCoord2f(uvs[i].x, 0.0f);
    glVertex2f(pt.x + sizes[i].width, pt.y);
    glEnd();

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

    glDisable(GL_TEXTURE_2D);

    pt.x += sizes[i].width;

    return pt;
}

- (void)drawString:(NSString *)str at:(NSPoint)pt red:(float)r green:(float)g
              blue:(float)b alpha:(float)a bgRed:(float)bgr
           bgGreen:(float)bgg bgBlue:(float)bgb bgAlpha:(float)bga
{
    int i;
    unichar ch;

    for(i = 0; i < (int)[str length]; ++i) {
        ch = [str characterAtIndex:i];
        pt = [self drawCharacter:(char)ch
                              at:pt
                             red:r
                           green:g
                            blue:b
                           alpha:a
                           bgRed:bgr
                         bgGreen:bgg
                          bgBlue:bgb
                         bgAlpha:bga];
    }
}

- (NSPoint)drawCharacter:(char)ch at:(NSPoint)pt red:(float)r green:(float)g
                    blue:(float)b alpha:(float)a bgRed:(float)bgr
                 bgGreen:(float)bgg bgBlue:(float)bgb bgAlpha:(float)bga
{
    int i;

    if(ch < 32 || ch > 126)
        return pt;

    i = ch - 32;

    glDisable(GL_TEXTURE_2D);

    glColor4f(bgr, bgg, bgb, bga);

    glBegin(GL_TRIANGLE_STRIP);
    glVertex2f(pt.x, pt.y + sizes[i].height);
    glVertex2f(pt.x, pt.y);
    glVertex2f(pt.x + sizes[i].width, pt.y + sizes[i].height);
    glVertex2f(pt.x + sizes[i].width, pt.y);
    glEnd();

    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, chars[i]);

    glColor4f(r, g, b, a);

    glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0.0f, uvs[i].y);
    glVertex2f(pt.x, pt.y + sizes[i].height);
    glTexCoord2f(0.0f, 0.0f);
    glVertex2f(pt.x, pt.y);
    glTexCoord2f(uvs[i].x, uvs[i].y);
    glVertex2f(pt.x + sizes[i].width, pt.y + sizes[i].height);
    glTexCoord2f(uvs[i].x, 0.0f);
    glVertex2f(pt.x + sizes[i].width, pt.y);
    glEnd();

    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);

    glDisable(GL_TEXTURE_2D);

    pt.x += sizes[i].width;

    return pt;
}

- (NSSize)metricsForCharacter:(char)ch
{
    if(ch >= 32 && ch < 127) {
        return sizes[ch - 32];
    }
    else {
        return NSMakeSize(-1, -1);
    }
}

@end
