/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <OpenGL/GL.h>
#include "PaletteGLView.h"

#include "sms.h"
#include "smsvdp.h"
#include "nesppu.h"

/* From sms.c */
extern int sms_initialized;
extern int nes_initialized;
extern int sms_console;

#define SMS_COLOR_NAME @"SMS Color Value (BGR222)"
#define GG_COLOR_NAME  @"GG Color Value (BGR444)"

#define FLOAT_TO_UINT32(r, g, b) \
    ((uint32)(r * 255.0f) << 16) | \
    ((uint32)(g * 255.0f) <<  8) | \
    ((uint32)(b * 255.0f) <<  0)

@implementation PaletteGLView

- (id)initWithFrame:(NSRect)frameRect
{
    NSOpenGLPixelFormatAttribute attrs[] = {
        NSOpenGLPFAWindow,
        NSOpenGLPFANoRecovery,
        NSOpenGLPFAColorSize, 32,
        NSOpenGLPFADepthSize, 32,
        NSOpenGLPFADoubleBuffer,
        0
    };

    NSOpenGLPixelFormat *fmt =
        [[NSOpenGLPixelFormat alloc] initWithAttributes:attrs];
    
    if(fmt == nil)
        return nil;

    if((self = [super initWithFrame:frameRect pixelFormat:fmt]) == nil) {
        [fmt release];
        return nil;
    }

    colorPanelVisible = NO;
    [fmt release];

    return self;
}

- (void)drawRect:(NSRect)rect
{
    static BOOL first = YES;
    int i;
    pixel_t col;

    if(first) {
        glShadeModel(GL_SMOOTH);
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
        glClearDepth(1.0f);

        glViewport(0.0f, 0.0f, [self bounds].size.width,
                   [self bounds].size.height);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        glOrtho(0.0f, 320.0f, 120.0f, 0.0f, 1.0f, 0.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        first = NO;
    }

    if(sms_initialized) {
        for(i = 0; i < 32; ++i) {
            col = smsvdp.pal[i];

            glBegin(GL_QUADS);
            glColor3ub(((col >> 10) & 0x1F) << 3, 
                       ((col >>  5) & 0x1F) << 3,
                       ((col >>  0) & 0x1F) << 3);
            glVertex2f(i * 10.0f, 0.0f);
            glVertex2f(i * 10.0f + 10.0f, 0.0f);
            glVertex2f(i * 10.0f + 10.0f, 120.0f);
            glVertex2f(i * 10.0f, 120.0f);
            glEnd();
        }
    }
    else if(nes_initialized) {
        uint8 bg_pal[16], spr_pal[16];

        nes_ppu_fetch_bg_pal(bg_pal);
        nes_ppu_fetch_spr_pal(spr_pal);

        for(i = 0; i < 16; ++i) {
            col = nes_pal[bg_pal[i]];

            glBegin(GL_QUADS);
            glColor3ub(((col >> 10) & 0x1F) << 3, 
                       ((col >>  5) & 0x1F) << 3,
                       ((col >>  0) & 0x1F) << 3);
            glVertex2f(i * 10.0f, 0.0f);
            glVertex2f(i * 10.0f + 10.0f, 0.0f);
            glVertex2f(i * 10.0f + 10.0f, 120.0f);
            glVertex2f(i * 10.0f, 120.0f);
            glEnd();
        }

        for(i = 0; i < 16; ++i) {
            col = nes_pal[spr_pal[i]];

            glBegin(GL_QUADS);
            glColor3ub(((col >> 10) & 0x1F) << 3, 
                       ((col >>  5) & 0x1F) << 3,
                       ((col >>  0) & 0x1F) << 3);
            glVertex2f((i + 16) * 10.0f, 0.0f);
            glVertex2f((i + 16) * 10.0f + 10.0f, 0.0f);
            glVertex2f((i + 16) * 10.0f + 10.0f, 120.0f);
            glVertex2f((i + 16) * 10.0f, 120.0f);
            glEnd();
        }
    }

    glFlush();
    [[self openGLContext] flushBuffer];
}

- (void)mouseDown:(NSEvent *)event
{
    if(sms_initialized) {
        NSPoint loc = [self convertPoint:[event locationInWindow]
                                fromView:nil];
        int index;
        char text[7];
        float r, g, b;

        if(colorPanelVisible == NO) {
            [colorPanel orderFront:self];
        }

        index = loc.x / 10;

        r = (float)((smsvdp.pal[index] >> 10) & 0x1F) / 31.0f;
        g = (float)((smsvdp.pal[index] >>  5) & 0x1F) / 31.0f;
        b = (float)((smsvdp.pal[index] >>  0) & 0x1F) / 31.0f;

        if(sms_console == CONSOLE_GG) {
            int col = smsvdp.cram[index << 1] |
                (smsvdp.cram[(index << 1) + 1] << 8);
            snprintf(text, 7, "%03X", col & 0x0FFF);

            [smsColorLabel setStringValue:GG_COLOR_NAME];
        }
        else {
            snprintf(text, 7, "%02X", smsvdp.cram[index] & 0x3F);
            [smsColorLabel setStringValue:SMS_COLOR_NAME];
        }

        [smsColor setStringValue:[NSString stringWithUTF8String:text]];

        snprintf(text, 7, "%06X", FLOAT_TO_UINT32(r, g, b));
        [rgbColor setStringValue:[NSString stringWithUTF8String:text]];

        [colorWell setColor:[NSColor colorWithDeviceRed:r
                                                  green:g
                                                   blue:b
                                                  alpha:1.0f]];
    }
}

@end
