/*
    This file is part of CrabEmu.

    Copyright (C) 2008, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/hid/IOHIDLib.h>
#include <IOKit/hid/IOHIDUsageTables.h>
#include <CoreFoundation/CoreFoundation.h>

#include "gamepad.h"

static int pad_count = 0;
static gamepad_t *pads = NULL;

static void gp_find_elements(CFMutableDictionaryRef prop, gamepad_t *pad);

static void gp_cb(void *target, IOReturn result, void *refcon, void *sender) {
    IOReturn err = 0;
    AbsoluteTime zero = { 0, 0 };
    IOHIDEventStruct event;
    gamepad_t *pad = (gamepad_t *)refcon;
    int i;

    /* Yes, GCC... I know these are unused... */
    (void)target;
    (void)result;
    (void)sender;

    while(!err) {
        err = (*pad->queue)->getNextEvent(pad->queue, &event, zero, 0);

        if (!err) {
            if(event.type == kIOHIDElementTypeInput_Button) {
                for(i = 0; i < pad->button_count; ++i) {
                    if(pad->buttons[i].cookie == event.elementCookie) {
                        pad->cb(pad, &pad->buttons[i], (int)event.value);
                        break;
                    }
                }
            }
            else if(event.type == kIOHIDElementTypeInput_Axis) {
                for(i = 0; i < pad->axis_count; ++i) {
                    if(pad->axes[i].cookie == event.elementCookie) {
                        pad->cb(pad, &pad->axes[i], (int)event.value);
                        break;
                    }
                }
            }
            else if(event.type == kIOHIDElementTypeInput_Misc) {
                /* Ugh... This seems to include both axes and hats... */
                for(i = 0; i < pad->axis_count; ++i) {
                    if(pad->axes[i].cookie == event.elementCookie) {
                        pad->cb(pad, &pad->axes[i], (int)event.value);
                        goto out;
                    }
                }

                for(i = 0; i < pad->hat_count; ++i) {
                    if(pad->hats[i].cookie == event.elementCookie) {
                        pad->cb(pad, &pad->hats[i], (int)event.value);
                        break;
                    }
                }
out:
                ;
            }
        }
    }
}

/* Get the io_iterator_t object needed to iterate through the list of HID
   devices. */
static int gp_get_iterator(mach_port_t port, io_iterator_t *iter) {
    CFMutableDictionaryRef d;
    IOReturn rv;

    /* Create a matching dictionary that will be used to search the device
     tree. */
    if(!(d = IOServiceMatching(kIOHIDDeviceKey)))
        return -1;

    /* Get all matching devices from IOKit. */
    rv = IOServiceGetMatchingServices(port, d, iter);

    if(rv != kIOReturnSuccess || !(*iter))
        return -1;

    return 0;
}

/* Create the interface needed to do stuff with the device. */
static int gp_create_interface(io_object_t hidDevice, gamepad_t *pad) {
    IOCFPlugInInterface **plugin;
    SInt32 score = 0;

    /* Create the plugin that we will use to actually get the device interface
       that is needed. */
    if(IOCreatePlugInInterfaceForService(hidDevice,
                                         kIOHIDDeviceUserClientTypeID,
                                         kIOCFPlugInInterfaceID, &plugin,
                                         &score) != kIOReturnSuccess)
        return -1;

    /* Grab the device interface from the plugin. */
    if((*plugin)->QueryInterface(plugin,
                                 CFUUIDGetUUIDBytes(kIOHIDDeviceInterfaceID),
                                 (LPVOID)&pad->iface) != S_OK)
        return -1;

    /* The plugin has done all it needs to, release it. */
    (*plugin)->Release(plugin);

    return 0;
}

/* Fill in a elemdata_t structure with information about the given physical
   element. */
static void gp_fill_elem(CFTypeRef elem, gp_elemdata_t *ptr) {
    CFTypeRef ref;
    int num;

    memset(ptr, 0, sizeof(gp_elemdata_t));

    /* Grab the element cookie. */
    if((ref = CFDictionaryGetValue(elem, CFSTR(kIOHIDElementCookieKey)))) {
        if(CFNumberGetValue(ref, kCFNumberIntType, &num))
            ptr->cookie = (IOHIDElementCookie)num;
    }

    /* Grab the element's minimum value. */
    if((ref = CFDictionaryGetValue(elem, CFSTR(kIOHIDElementMinKey)))) {
        if(CFNumberGetValue(ref, kCFNumberIntType, &num))
            ptr->min = num;
    }
    
    /* Grab the element's maximum value. */
    if((ref = CFDictionaryGetValue(elem, CFSTR(kIOHIDElementMaxKey)))) {
        if(CFNumberGetValue(ref, kCFNumberIntType, &num))
            ptr->max = num;
    }
}

/* Callback function to handle each sub-element of the controller class. */
static void gp_elem_array_hnd(const void *value, void *parameter) {
    CFTypeRef elem = (CFTypeRef)value;
    CFTypeID t = CFGetTypeID(value);
    gamepad_t *pad = (gamepad_t *)parameter;
    long elemType, elemPage, elemUsage;
    CFTypeRef type, page, usage;
    void *ptr;
    int et;

    /* Make sure we're dealing with a dictionary. */
    if(t == CFDictionaryGetTypeID()) {
        /* Grab the type of the element. */
        if(!(type = CFDictionaryGetValue(elem, CFSTR(kIOHIDElementTypeKey))))
            return;

        /* Grab the HID usage page. */
        if(!(page = CFDictionaryGetValue(elem,
                                         CFSTR(kIOHIDElementUsagePageKey))))
            return;

        /* Grab the HID usage type. */
        if(!(usage = CFDictionaryGetValue(elem, CFSTR(kIOHIDElementUsageKey))))
            return;

        /* Get the integer values from the data grabbed above. */
        if(!CFNumberGetValue(type, kCFNumberLongType, &elemType))
            return;

        if(!CFNumberGetValue(page, kCFNumberLongType, &elemPage))
            return;

        if(!CFNumberGetValue(usage, kCFNumberLongType, &elemUsage))
            return;

        /* If the element is listed as a button, axis, or misc input, check
           what it actually is. */
        if(elemType == kIOHIDElementTypeInput_Button) {
            if(elemPage == kHIDPage_Button) {
                ptr = realloc(pad->buttons, (pad->button_count + 1) *
                              sizeof(gp_elemdata_t));

                if(ptr) {
                    pad->buttons = (gp_elemdata_t *)ptr;
                    gp_fill_elem(elem, pad->buttons + pad->button_count);
                    pad->buttons[pad->button_count].number = elemUsage;
                    pad->buttons[pad->button_count].type = GP_TYPE_BUTTON;
                    ++pad->button_count;
                }
            }
        }
        else if(elemType == kIOHIDElementTypeInput_Axis ||
                elemType == kIOHIDElementTypeInput_Misc) {
            if(elemPage == kHIDPage_GenericDesktop) {
                switch(elemUsage) {
                    case kHIDUsage_GD_X:
                        et = GP_TYPE_X_AXIS;
                        goto axis;

                    case kHIDUsage_GD_Y:
                        et = GP_TYPE_Y_AXIS;
                        goto axis;

                    case kHIDUsage_GD_Z:
                        et = GP_TYPE_Z_AXIS;
                        goto axis;

                    case kHIDUsage_GD_Rx:
                        et = GP_TYPE_RX_AXIS;
                        goto axis;

                    case kHIDUsage_GD_Ry:
                        et = GP_TYPE_RY_AXIS;
                        goto axis;

                    case kHIDUsage_GD_Rz:
                        et = GP_TYPE_RZ_AXIS;
axis:
                        ptr = realloc(pad->axes, (pad->axis_count + 1) *
                                      sizeof(gp_elemdata_t));

                        if(ptr) {
                            pad->axes = (gp_elemdata_t *)ptr;
                            gp_fill_elem(elem, pad->axes + pad->axis_count);
                            pad->axes[pad->axis_count].type = et;
                            ++pad->axis_count;
                        }
                        break;

                    case kHIDUsage_GD_Hatswitch:
                        ptr = realloc(pad->hats, (pad->hat_count + 1) *
                                      sizeof(gp_elemdata_t));
                        
                        if(ptr) {
                            pad->hats = (gp_elemdata_t *)ptr;
                            gp_fill_elem(elem, pad->hats + pad->hat_count);
                            pad->hats[pad->hat_count].type = GP_TYPE_HAT;
                            pad->hats[pad->hat_count].number = pad->hat_count;
                            ++pad->hat_count;
                        }
                        break;
                }
            }
        }
        /* If we've found another element array, effectively recurse. */
        else if(elemType == kIOHIDElementTypeCollection) {
            gp_find_elements((CFMutableDictionaryRef)elem, pad);
        }
    }
}

/* Process all the sub-elements of a given property list. */
static void gp_find_elements(CFMutableDictionaryRef prop, gamepad_t *pad) {
    CFTypeRef elem;
    CFTypeID type;

    if((elem = CFDictionaryGetValue(prop, CFSTR(kIOHIDElementKey)))) {
        type = CFGetTypeID(elem);

        if(type == CFArrayGetTypeID()) {
            /* Call our function on each element of the array. */
            CFRange r = { 0, CFArrayGetCount(elem) };
            CFArrayApplyFunction(elem, r, &gp_elem_array_hnd, (void *)pad);
        }
    }
}

/* Compare two buttons for sorting. */
static int gp_cmp_buttons(const void *e1, const void *e2) {
    const gp_elemdata_t *b1 = (const gp_elemdata_t *)e1;
    const gp_elemdata_t *b2 = (const gp_elemdata_t *)e2;

    return b1->number < b2->number ? -1 : (b1->number > b2->number ? 1 : 0);
}

/* Read the device passed in, and add it to our game pad list if appropriate. */
static void gp_read_device(io_object_t dev) {
    CFMutableDictionaryRef props = NULL;
    CFTypeRef inf;
    SInt32 page, usage;
    void *ptr;

    /* Create a dictionary to read the device's properties. */
    if(IORegistryEntryCreateCFProperties(dev, &props, kCFAllocatorDefault,
                                         kNilOptions) == KERN_SUCCESS) {
        /* Grab the primary usage page of the device. */
        inf = CFDictionaryGetValue(props, CFSTR(kIOHIDPrimaryUsagePageKey));

        if(!inf || !CFNumberGetValue((CFNumberRef)inf, kCFNumberSInt32Type,
                                     &page))
            goto out;

        /* Ignore devices that are not in the Generic Desktop page. */
        if(page != kHIDPage_GenericDesktop)
            goto out;

        /* Grab the primary device usage. */
        inf = CFDictionaryGetValue(props, CFSTR(kIOHIDPrimaryUsageKey));

        if(!inf || !CFNumberGetValue((CFNumberRef)inf, kCFNumberSInt32Type,
                                     &usage))
            goto out;

        /* Ignore devices that are not either a Game Pad or Joystick. */
        if(usage != kHIDUsage_GD_GamePad && usage != kHIDUsage_GD_Joystick)
            goto out;

        /* Allocate space for the new gamepad structure. */
        ptr = realloc(pads, (pad_count + 1) * sizeof(gamepad_t));

        if(!ptr)
            goto out;

        pads = (gamepad_t *)ptr;
        memset(pads + pad_count, 0, sizeof(gamepad_t));

        /* Grab and store the name of the device. */
        inf = CFDictionaryGetValue(props, CFSTR(kIOHIDProductKey));

        if(!CFStringGetCString((CFStringRef)inf, pads[pad_count].name, 256,
                               kCFStringEncodingUTF8))
            goto out;

        /* Create the device interface needed to interact with the device. */
        if(gp_create_interface(dev, pads + pad_count))
            goto out;

        /* Find all elements of the device. */
        gp_find_elements(props, pads + pad_count);

        qsort(pads[pad_count].buttons, pads[pad_count].button_count,
              sizeof(gp_elemdata_t), &gp_cmp_buttons);

        ++pad_count;
    }

out:
    CFRelease(props);
}

/* Scan the system for any connected gamepads/joysticks/etc. */
int gp_scan(void) {
    io_iterator_t iter;
    io_object_t dev;

    /* Make sure we don't re-initialize without cleaning up first... */
    if(pads)
        return -1;

    /* Get the iterator needed for looping through the list of devices. If this
       returns < 0, then there are either no devices or IOKit is being wonky.
       We'll assume the former. */
    if(gp_get_iterator(kIOMasterPortDefault, &iter) < 0)
        return 0;

    while((dev = IOIteratorNext(iter))) {
        gp_read_device(dev);
        IOObjectRelease(dev);
    }

    /* Clean up. */
    IOObjectRelease(iter);
    return pad_count;
}

/* Release the given gamepad's interface and clean up any memory used by it. */
static void gp_release(gamepad_t *pad) {
    (*pad->iface)->Release(pad->iface);

    free(pad->buttons);
    free(pad->axes);
    free(pad->hats);

    pad->iface = NULL;
    pad->buttons = pad->axes = pad->hats = NULL;
}

void gp_cleanup(void) {
    int i;

    for(i = 0; i < pad_count; ++i) {
        gp_close(pads + i);
        gp_release(pads + i);
    }

    free(pads);
    pads = NULL;

    pad_count = 0;
}

/* Get the number of gamepads connected. */
int gp_count(void) {
    return pad_count;
}

/* Get the gamepad at a given index in our list. */
gamepad_t *gp_get_gamepad(int index) {
    if(index < 0 || index >= pad_count)
        return NULL;

    return &pads[index];
}

/* Open a device, optionally associating a callback with it for when any events
   occur. */
int gp_open(gamepad_t *pad, gp_callback cb) {
    int qlen, i;
    CFRunLoopSourceRef src;

    /* Make sure we're not already open */
    if(pad->open)
        return -2;

    /* Open the device... */
    if((*pad->iface)->open(pad->iface, 0))
        return -1;

    pad->open = 1;

    if(cb) {
        pad->cb = cb;

        /* Allocate the queue and set it up... */
        if(!(pad->queue = (*pad->iface)->allocQueue(pad->iface)))
            goto err_iface;

        qlen = (pad->button_count + pad->hat_count + pad->axis_count) << 1;
        if((*pad->queue)->create(pad->queue, 0, qlen) != kIOReturnSuccess)
            goto err_queue;

        /* Add each element to the queue. */
        for(i = 0; i < pad->button_count; ++i) {
            if((*pad->queue)->addElement(pad->queue, pad->buttons[i].cookie,
                                         0) != kIOReturnSuccess)
                goto err_queue;
        }

        for(i = 0; i < pad->hat_count; ++i) {
            if((*pad->queue)->addElement(pad->queue, pad->hats[i].cookie,
                                         0) != kIOReturnSuccess)
                goto err_queue;
        }

        for(i = 0; i < pad->axis_count; ++i) {
            if((*pad->queue)->addElement(pad->queue, pad->axes[i].cookie,
                                         0) != kIOReturnSuccess)
                goto err_queue;
        }

        /* Start the queue up... */
        if((*pad->queue)->createAsyncEventSource(pad->queue,
                                                 &src) != kIOReturnSuccess)
            goto err_queue;

        if((*pad->queue)->setEventCallout(pad->queue, gp_cb, NULL,
                                          pad) != kIOReturnSuccess) {
            CFRelease(src);
            goto err_queue;
        }

        /* This call will retain src, so we can release it below... */
        CFRunLoopAddSource(CFRunLoopGetCurrent(), src, kCFRunLoopDefaultMode);
        pad->ev_src = src;
        CFRelease(src);

        if((*pad->queue)->start(pad->queue) != kIOReturnSuccess) {
            CFRelease(src);
            goto err_queue;
        }
    }

    return 0;

err_queue:
    (*pad->queue)->Release(pad->queue);
    pad->queue = NULL;
err_iface:
    (*pad->iface)->close(pad->iface);
    pad->cb = NULL;
    pad->ev_src = NULL;
    pad->open = 0;
    return -1;
}

/* Close the device and return its resources to the system. */
int gp_close(gamepad_t *pad) {
    IOReturn rv;

    if(!pad->open)
        return 0;

    /* Clean up the queue and the callback, if any is registered. */
    if(pad->queue) {
        (*pad->queue)->stop(pad->queue);
        CFRunLoopRemoveSource(CFRunLoopGetCurrent(), pad->ev_src,
                              kCFRunLoopDefaultMode);
        (*pad->queue)->dispose(pad->queue);
        (*pad->queue)->Release(pad->queue);
        pad->queue = NULL;
        pad->ev_src = NULL;
        pad->cb = NULL;
    }

    rv = (*pad->iface)->close(pad->iface);
    pad->open = 0;

    if(rv != kIOReturnSuccess)
        return -1;

    return 0;
}

/* Read a given element from the gamepad. The gamepad must be open for this
   function to actually do anything useful. */
int gp_read_element(gamepad_t *pad, gp_elemdata_t *elem) {
    IOHIDEventStruct ev;

    memset(&ev, 0, sizeof(ev));
    (*pad->iface)->getElementValue(pad->iface, elem->cookie, &ev);

    return ev.value;
}
