/*
    This file is part of CrabEmu.

    Copyright (C) 2008, 2012 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef GAMEPAD_H
#define GAMEPAD_H

#include <IOKit/hid/IOHIDLib.h>
#include <IOKit/hid/IOHIDUsageTables.h>

struct gp_elemdata_s;
struct gamepad_s;
typedef struct gp_elemdata_s gp_elemdata_t;
typedef struct gamepad_s gamepad_t;

/* Type for callbacks for button presses and joystick movement, and whatnot. */
typedef void (*gp_callback)(gamepad_t *gp, gp_elemdata_t *elem, int value);

/* Values for the gp_elemdata_t type element. */
#define GP_TYPE_BUTTON          0
#define GP_TYPE_X_AXIS          1
#define GP_TYPE_Y_AXIS          2
#define GP_TYPE_Z_AXIS          3
#define GP_TYPE_RX_AXIS         4
#define GP_TYPE_RY_AXIS         5
#define GP_TYPE_RZ_AXIS         6
#define GP_TYPE_HAT             7

/* Structure for holding basic information about each element. */
struct gp_elemdata_s {
    IOHIDElementCookie cookie;
    int type;
    int number;
    int min;
    int max;
};

/* Structure for holding information about each gamepad connected. */
struct gamepad_s {
    IOHIDDeviceInterface **iface;
    IOHIDQueueInterface **queue;
    CFRunLoopSourceRef ev_src;
    gp_callback cb;

    int open;
    int button_count;
    int axis_count;
    int hat_count;

    gp_elemdata_t *buttons;
    gp_elemdata_t *axes;
    gp_elemdata_t *hats;

    char name[256];
};

/* Scan the system for any gamepads (or other similar devices) connected.
   Returns the total number of devices found on success.
   Returns -1 if already initialized. */
int gp_scan(void);

/* Clean up the data allocated by the pad_scan_pads() function. */
void gp_cleanup(void);

/* You must have called gp_scan() already to use any of the rest of these
   functions. */

/* Open a device, optionally associating a callback with it for when any events
   occur.
   Returns 0 on success.
   Returns -2 if the device is already opened.
   Returns -1 on failure. */
int gp_open(gamepad_t *pad, gp_callback cb);

/* Close the device and return its resources to the system.
   Returns 0 on success (or if the device wasn't open).
   Returns -1 on failure. */
int gp_close(gamepad_t *pad);

/* Get the number of gamepads connected. */
int gp_count(void);

/* Get the gamepad at a given index in our list.
   Returns a pointer to the gamepad struct on success.
   Returns NULL if index does not refer to a valid gamepad. */
gamepad_t *gp_get_gamepad(int index);

/* Read a given element from the gamepad. The gamepad must be open for this
   function to actually do anything useful. */
int gp_read_element(gamepad_t *pad, gp_elemdata_t *elem);

#endif /* !GAMEPAD_H */
