/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008, 2009, 2012, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#import <Cocoa/Cocoa.h>
#include <CoreFoundation/CFDate.h>
#include "EmuOGLView.h"
#include "EmuController.h"
#include "Z80DebugController.h"
#include "SMSVDPDebugController.h"
#include "NSCondition.h"

#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include "sms.h"
#include "colecovision.h"
#include "smsvdp.h"
#include "smsz80.h"
#include "nes.h"
#include "sound.h"

EmuOGLView *theview;
EmuController *thecont;
Z80DebugController *debugger;
SMSVDPDebugController *vdpDebugger;
NSLock *memoryLock;
BOOL debuggerEnabled = NO;
int debuggerMode = 0;
NSCondition *ss_sem;
BOOL ntscVideo = YES;

extern int sms_console;

void emu_loop(void) {
    int oldrun = 0;
#ifdef TIMING_TEST
    time_t thetime = time(NULL);
    uint32 frames = 0;
#endif

    while([thecont running]) {
        if(debuggerEnabled == NO) {
#ifndef TIMING_TEST
            sound_wait();
#endif

            [memoryLock lock];

            switch(sms_console) {
                case CONSOLE_COLECOVISION:
                    oldrun = coleco_frame(oldrun, 0);
                    break;

                case CONSOLE_NES:
                    oldrun = nes_frame(oldrun, 0);
                    break;

                default:
                    oldrun = sms_frame(oldrun, 0);
                    break;
            }

            [theview setNeedsDisplay:YES];

            [memoryLock unlock];

#ifdef TIMING_TEST
            ++frames;

            if(thetime + 5 <= time(NULL)) {
                printf("%lu frames in %d seconds = %f FPS\n",
                       (long unsigned int)frames,
                       (int)(time(NULL) - thetime),
                       (float) frames / (time(NULL) - thetime));
                thetime = time(NULL);
                frames = 0;
            }
#endif
        }
        else {
            int line;
            int cycles;
            int lines = ntscVideo == YES ? 262 : 313;

            [memoryLock lock];

            for(line = 0; line < lines; ++line) {
                for(cycles = 228; cycles > 0; cycles -= sms_z80_run(1)) {
                    if(debuggerEnabled == YES && debuggerMode == 0) {
                        [vdpDebugger update];
                        [debugger updateWithCycles:cycles scanline:line];
                        [ss_sem lock];
                        [ss_sem wait];
                        [ss_sem unlock];
                    }
                }

                if(sms_console != CONSOLE_COLECOVISION)
                    sms_vdp_execute(line, 0);
                else
                    tms9918a_vdp_execute(line, &sms_z80_nmi, 0);

                if(debuggerEnabled == YES && debuggerMode == 1) {
                    [vdpDebugger update];
                    [debugger updateWithCycles:228 scanline:line];
                    [ss_sem lock];
                    [ss_sem wait];
                    [ss_sem unlock];
                }
            }

            [theview display];
            [memoryLock unlock];

            if(debuggerEnabled == YES && debuggerMode == 2) {
                [vdpDebugger update];
                [debugger updateWithCycles:228 scanline:226];
                [ss_sem lock];
                [ss_sem wait];
                [ss_sem unlock];
            }
        }
    }
}

void gui_set_aspect(float x, float y) {
    [theview setAspectRatioX:x Y:y];
}

void gui_set_title(const char *str) {
    [theview setTitle:[NSString stringWithUTF8String:str]];
}

#ifdef ENABLE_SDSC_TERMINAL

void gui_update_sdsc_terminal(char console[25][80 * 2]) {
    [thecont updateSDSCTerminalWithString:console];
}

#endif

int main(int argc, char *argv[]) {
    int retval;

    memoryLock = [[NSLock alloc] init];
    ss_sem = [[NSCondition alloc] init];

    retval = NSApplicationMain(argc,  (const char **)argv);

    [ss_sem release];
    [memoryLock release];
}
