/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QPainter>
#include <QMouseEvent>

#include "EmuWidget.h"
#include "CrabEmuThread.h"
#include "CrabEmu.h"

#include "sms.h"
#include "smsvdp.h"
#include "nes.h"
#include "nesppu.h"
#include "terebi.h"
#include "pastlib.h"
#include "scale2x.h"

#if defined(_WIN32) && defined(ENABLE_OPENGL)
#include "glext.h"
#endif

extern "C" int sms_console;
extern "C" sms_vdp_t smsvdp;
extern "C" int sms_initialized, coleco_initialized, nes_initialized;

#if defined(ENABLE_OPENGL) && defined(CRABEMU_32BIT_COLOR)
#define PXTYPE GL_BGRA
#define DTTYPE GL_UNSIGNED_INT_8_8_8_8_REV
#elif defined(ENABLE_OPENGL)
#define PXTYPE GL_BGRA
#define DTTYPE GL_UNSIGNED_SHORT_1_5_5_5_REV
#endif

#ifndef ENABLE_OPENGL

EmuWidget::EmuWidget(CrabEmuThread *thd, QWidget *parent) : QWidget(parent),
#ifdef CRABEMU_32BIT_COLOR
        img(256, 256, QImage::Format_RGB32) {
#else
        img(256, 256, QImage::Format_RGB555) {
#endif
    connect(thd, SIGNAL(frameRendered()), this, SLOT(update()));
    renderer = Render_Normal;
}

#else

EmuWidget::EmuWidget(CrabEmuThread *thd, QWidget *parent) : QGLWidget(parent) {
    connect(thd, SIGNAL(frameRendered()), this, SLOT(updateGL()));
    txrChanged = true;
    txrCreated = false;
    txrWidth = 256;
    txrHeight = 256;
    fbTrans = NULL;
    renderer = Render_Normal;
}

#endif

#ifndef ENABLE_OPENGL

QImage *EmuWidget::retainImage() {
    mutex.lock();
    return &img;
}

void EmuWidget::releaseImage() {
    mutex.unlock();
}

#else

QImage *EmuWidget::retainImage() {
    mutex.lock();
    img = grabFrameBuffer();
    return &img;
}

void EmuWidget::releaseImage() {
    mutex.unlock();
}

#endif

#ifndef ENABLE_OPENGL

void EmuWidget::paintEvent(QPaintEvent *event __UNUSED__) {
    QMutexLocker locker(&mutex);
    QPainter painter(this);
    QRect r;

    if(sms_initialized || coleco_initialized || nes_initialized) {
        if(sms_console == CONSOLE_GG) {
            r.setRect(48, 24, 160, 144);
        }
        else if(sms_console == CONSOLE_NES) {
            r.setRect(0, 0, 256, 240);
        }
        else {
            r.setRect(0, 0, 256, smsvdp.lines);
        }

        painter.drawImage(rect(), img, r);
    }
    else {
        img.fill(0);
        painter.drawRect(QRect());
    }
}

#else

void EmuWidget::resizeGL(int width, int height) {
    cxtWidth = width;
    cxtHeight = height;

    if(!sms_initialized && !coleco_initialized && !nes_initialized)
        return;

    glShadeModel(GL_SMOOTH);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0f);

    glViewport(0.0f, 0.0f, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if(sms_console == CONSOLE_NES)
        glOrtho(0.0f, 255.0f, 239.0f, 0.0f, 1.0f, 0.0f);
    else if(sms_console != CONSOLE_GG)
        glOrtho(0.0f, 255.0f, smsvdp.lines - 1, 0.0f, 1.0f, 0.0f);
    else
        glOrtho(0.0f, 159.0f, 143.0f, 0.0f, 1.0f, 0.0f);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}

void EmuWidget::paintGL() {    
    if(!sms_initialized && !coleco_initialized && !nes_initialized) {
        glClear(GL_COLOR_BUFFER_BIT);
        return;
    }

    if(txrChanged) {
        if(txrCreated) {
            glDeleteTextures(1, &glTexture);
        }
        else {
            /* Make sure the context has been set up correctly... */
            resizeGL(cxtWidth, cxtHeight);
        }

        glGenTextures(1, &glTexture);
        glBindTexture(GL_TEXTURE_2D, glTexture);
        glTexImage2D(GL_TEXTURE_2D, 0, 4, txrWidth, txrHeight, 0, PXTYPE,
                     DTTYPE, NULL);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        txrCreated = true;
        txrChanged = false;
    }

    glClear(GL_COLOR_BUFFER_BIT);
    glLoadIdentity();

    glEnable(GL_TEXTURE_2D);

    glBindTexture(GL_TEXTURE_2D, glTexture);

    if(sms_console != CONSOLE_NES) {
        switch(renderer) {
            case Render_Normal:
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, smsvdp.framebuffer);
                break;

            case Render_Scale2x:
                scale2x_scale(smsvdp.framebuffer, fbTrans, 256, smsvdp.lines);
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, fbTrans);
                break;

            case Render_HQ2x:
                ProcessHQ2x((pixel *)smsvdp.framebuffer, (pixel *)fbTrans,
                            smsvdp.lines);
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, fbTrans);
                break;

            case Render_HQ4x:
                ProcessHQ4x((pixel *)smsvdp.framebuffer, (pixel *)fbTrans,
                            smsvdp.lines);
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, fbTrans);
                break;
        }
    }
    else {
        switch(renderer) {
            case Render_Normal:
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, nes_ppu_framebuffer());
                break;

            case Render_Scale2x:
                scale2x_scale((pixel_t *)nes_ppu_framebuffer(), fbTrans, 256,
                              240);
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, fbTrans);
                break;

            case Render_HQ2x:
                ProcessHQ2x((pixel *)nes_ppu_framebuffer(), (pixel *)fbTrans,
                            240);
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, fbTrans);
                break;

            case Render_HQ4x:
                ProcessHQ4x((pixel *)nes_ppu_framebuffer(), (pixel *)fbTrans,
                            240);
                glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, txrWidth, txrHeight,
                                PXTYPE, DTTYPE, fbTrans);
                break;
        }
    }

    if(sms_console == CONSOLE_NES) {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(0.0f, 240.0f / 256.0f);
        glVertex2f(0.0f, 239.0f);
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(1.0f, 240.0f / 256.0f);
        glVertex2f(255.0f, 239.0f);
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(255.0f, 0.0f);
        glEnd();
    }
    else if(sms_console != CONSOLE_GG) {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(0.0f, smsvdp.lines / 256.0f);
        glVertex2f(0.0f, smsvdp.lines - 1);
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(1.0f, smsvdp.lines / 256.0f);
        glVertex2f(255.0f, smsvdp.lines - 1);
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(255.0f, 0.0f);
        glEnd();
    }
    else {
        glBegin(GL_TRIANGLE_STRIP);
        glTexCoord2f(48.0f / 256.0f, 168.0f / 256.0f);
        glVertex2f(0.0f, 143.0f);
        glTexCoord2f(48.0f / 256.0f, 24.0f / 256.0f);
        glVertex2f(0.0f, 0.0f);
        glTexCoord2f(208.0f / 256.0f, 168.0f / 256.0f);
        glVertex2f(159.0f, 143.0f);
        glTexCoord2f(208.0f / 256.0f, 24.0f / 256.0f);
        glVertex2f(159.0f, 0.0f);
        glEnd();
    }

    glFlush();
}

#endif

void EmuWidget::mousePressEvent(QMouseEvent *event) {
    int x, y;

    /* Scale the position to [0.0, 1.0], and multiply by the SG-1000's
       resolution. */
    x = (int)(((float)event->x() / width()) * 256);
    y = (int)(((float)event->y() / height()) * 192);

    terebi_update(x, y, 1);
    event->accept();
}

void EmuWidget::mouseReleaseEvent(QMouseEvent *event) {
    terebi_update(0, 0, 0);
    event->accept();
}

void EmuWidget::mouseMoveEvent(QMouseEvent *event) {
    mousePressEvent(event);
}

void EmuWidget::setRenderer(EmuRenderer r) {
#if defined(ENABLE_OPENGL) && !defined(CRABEMU_32BIT_COLOR)
    QMutexLocker locker(&mutex);

    switch(r) {
        case Render_Normal:
            free(fbTrans);
            fbTrans = NULL;
            txrWidth = txrHeight = 256;
            txrChanged = true;
            break;

        case Render_Scale2x:
        case Render_HQ2x:
            free(fbTrans);
            fbTrans = NULL;
            txrWidth = txrHeight = 512;
            txrChanged = true;
            fbTrans = (pixel_t *)malloc(sizeof(pixel_t) * txrWidth * txrHeight);
            break;

        case Render_HQ4x:
            free(fbTrans);
            fbTrans = NULL;
            txrWidth = txrHeight = 1024;
            txrChanged = true;
            fbTrans = (pixel_t *)malloc(sizeof(pixel_t) * txrWidth * txrHeight);
            break;

        default:
            return;
    }

    renderer = r;
#endif
}
