/*
    This file is part of CrabEmu.

    Copyright (C) 2009, 2012, 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <QApplication>
#include <QMessageBox>
#include <QFileDialog>
#include <QKeyEvent>
#include <QSettings>
#include <QFile>

#include "UICrabEmu.h"
#include "UIAboutDialog.h"
#include "UIPreferences.h"
#include "UIDebugWindow.h"
#include "../EmuWidget.h"
#include "sms.h"
#include "colecovision.h"
#include "nes.h"
#include "smsmem.h"
#include "smsmem-gg.h"
#include "colecomem.h"
#include "nesmem.h"
#include "rom.h"
#include "smsvdp.h"
#include "sn76489.h"
#include "smsz80.h"
#include "sound.h"
#include "cheats.h"

extern "C" int sms_initialized;
extern "C" int coleco_initialized;
extern "C" int nes_initialized;
extern "C" int sms_psg_enabled;
extern "C" int sms_ym2413_enabled;
extern "C" sn76489_t psg;
extern "C" int sms_console;
extern "C" int sms_region;
extern QMutex memoryLock;

UICrabEmu::UICrabEmu(QWidget *parent) : QMainWindow(parent), videoMultiple(1) {
    QSettings settings;
    bool pal, domestic;
    QActionGroup *videoActionGroup = new QActionGroup(this);

    setupUi(this);
    setAttribute(Qt::WA_DeleteOnClose);

    actionNormal->setActionGroup(videoActionGroup);
    actionScale2x->setActionGroup(videoActionGroup);
    actionHQ2x->setActionGroup(videoActionGroup);
    actionHQ4x->setActionGroup(videoActionGroup);

#if !defined(ENABLE_OPENGL) || defined(CRABEMU_32BIT_COLOR)
    actionScale2x->setDisabled(true);
    actionHQ2x->setDisabled(true);
    actionHQ4x->setDisabled(true);
#endif

    thd = new CrabEmuThread;

    emuwidget = new EmuWidget(thd, this);
    setCentralWidget(emuwidget);
    emuwidget->setFocusProxy(this);

    paused = false;
    debugger = false;
    debugWindow = NULL;

    /* If the user doesn't currently have any settings, set up the defaults. */
    if(!settings.allKeys().contains("SMS Buttonmap")) {
        initializeSettings();
        settings.sync();
    }

    if(!settings.allKeys().contains("Coleco Buttonmap")) {
        initializeSettings();
        settings.sync();
    }

    if(!settings.allKeys().contains("NES Buttonmap")) {
        initializeSettings();
        settings.sync();
    }

    pal = settings.value("SMS PAL").toBool();

    domestic = settings.value("SMS Domestic").toBool();

    videoSystem = pal ? SMS_VIDEO_PAL : SMS_VIDEO_NTSC;
    actionNTSC->setChecked(!pal);
    actionPAL->setChecked(pal);

    region = domestic ? SMS_REGION_DOMESTIC : SMS_REGION_EXPORT;
    actionExport->setChecked(!domestic);
    actionDomestic->setChecked(domestic);

    actionEnableCheats->setVisible(false);
}

UICrabEmu::~UICrabEmu() {
    // the emulator can't be paused otherwise the thread will just wait and the
    // application never exits cleanly
    unpause();

    delete thd;

    if(sms_initialized) {
        saveSRAM();
        sms_shutdown();
    }
    else if(coleco_initialized) {
        coleco_shutdown();
    }
    else if(nes_initialized) {
        saveSRAM();
        nes_shutdown();
    }
}

void UICrabEmu::restoreSettings() {
    QSettings settings;
    settings.beginGroup("Window");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
}

void UICrabEmu::closeEvent(QCloseEvent *event) {
    QSettings settings;
    settings.beginGroup("Window");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    QMainWindow::closeEvent(event);
}

CrabEmuThread *UICrabEmu::thread() {
    return thd;
}

UIDebugWindow* UICrabEmu::debug() {
    return debugWindow;
}

int UICrabEmu::initializeCV(QString filename) {
    QString biosfn;
    QSettings settings;

    unpause();

    if(sms_initialized) {
        saveSRAM();
        if(sms_shutdown()) {
            return -1;
        }
    }
    else if(nes_initialized) {
        saveSRAM();
        if(nes_shutdown())
            return -1;
    }

    if(debugger)
        on_actionDebugger_triggered();

    // Disable features not available for the ColecoVision
    sms_ym2413_enabled = 0;
    actionEnableYM2413->setDisabled(true);
    actionDebugger->setDisabled(true);
    menuRegion->setDisabled(true);
    actionEnablePSG->setEnabled(true);
    actionEnableCheats->setDisabled(true);
    menuPSG_Channels->setEnabled(true);
    menuVideo_System->setEnabled(true);

    if(coleco_initialized) {
        if(sms_region & videoSystem) {
            if(coleco_reset()) {
                return -1;
            }
        }
        else {
            coleco_shutdown();
            coleco_init(videoSystem);
            updateFps();
        }
    }
    else {
        if(coleco_init(videoSystem)) {
            return -1;
        }
        updateFps();
    }

    biosfn = settings.value("ColecoVision BIOS ROM").toString();

    if(biosfn.isEmpty()) {
        QMessageBox::warning(QApplication::activeWindow(), "No BIOS "
                             "Configured", "You must configure a BIOS ROM\n"
                             "to play a ColecoVision ROM.");
        coleco_shutdown();
        return -1;
    }

    if(coleco_mem_load_bios(biosfn.toStdString().c_str())) {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
               "BIOS ROM file", "CrabEmu cannot open " + biosfn + "\n"
               "Please check that the file exists and is accessible.");
        coleco_shutdown();
        return -1;
    }

    if(coleco_mem_load_rom(filename.toStdString().c_str())) {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
               "ROM file", "CrabEmu cannot open " + filename + "\n"
               "Please check that the file exists and is accessible.");
        coleco_shutdown();
        return -1;
    }

    resizeEmu(256, smsvdp.lines);
    return 0;
}

int UICrabEmu::initializeSMS(QString filename, int console) {
    QString biosfn;
    QSettings settings;

    unpause();

    if(coleco_initialized) {
        saveSRAM();
        if(coleco_shutdown()) {
            return -1;
        }
    }
    else if(nes_initialized) {
        saveSRAM();
        if(nes_shutdown())
            return -1;
    }

    if(sms_initialized) {
        saveSRAM();

        if(sms_region & region && sms_region & videoSystem) {
            if(sms_reset()) {
                return -1;
            }
        }
        else {
            sms_shutdown();
            sms_init(videoSystem, region);
            updateFps();
        }
    }
    else {
        if(sms_init(videoSystem, region)) {
            return -1;
        }
        updateFps();
    }

    // Enable available features
    actionEnableYM2413->setEnabled(true);
    actionEnablePSG->setEnabled(true);
    actionDebugger->setEnabled(true);
    menuRegion->setEnabled(true);
    actionEnableCheats->setEnabled(true);
    menuPSG_Channels->setEnabled(true);
    menuVideo_System->setEnabled(true);
    sms_ym2413_enabled = actionEnableYM2413->isChecked();

    if(console == CONSOLE_SMS) {
        biosfn = settings.value("SMS BIOS ROM").toString();

        if(settings.value("SMS BIOS Enabled", false).toBool() &&
           biosfn.length()) {
            if(!sms_mem_load_bios(biosfn.toUtf8().data())) {
                sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                               SMS_MEMCTL_RAM) ^ 0xFF);
            }
            else {
                QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
                       "BIOS ROM file", "CrabEmu cannot open " + biosfn + "\n"
                       "Please check that the file exists and is accessible.");
            }
        }

        resizeEmu(256, smsvdp.lines);
    }
    else if(console == CONSOLE_GG) {
        biosfn = settings.value("Game Gear BIOS ROM").toString();

        if(settings.value("GG BIOS Enabled", false).toBool() &&
           biosfn.length()) {
            if(!sms_mem_load_bios(biosfn.toUtf8().data())) {
                sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                                  SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
            }
            else {
                QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
                       "BIOS ROM file", "CrabEmu cannot open " + biosfn + "\n"
                       "Please check that the file exists and is accessible.");
            }
        }

        resizeEmu(160, 144);
    }

    if(sms_mem_load_rom(filename.toUtf8().data(), console)) {
        QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
               "ROM file", "CrabEmu cannot open " + filename + "\n"
               "Please check that the file exists and is accessible.");
        sms_shutdown();
        return -1;
    }

    loadSRAM();
    return 0;
}

int UICrabEmu::initializeNES(QString filename) {
    QString biosfn;
    QSettings settings;

    unpause();

    if(coleco_initialized) {
        if(coleco_shutdown())
            return -1;
    }
    else if(sms_initialized) {
        saveSRAM();
        if(sms_shutdown())
            return -1;
    }

    /* Disable menu options that make no sense for the NES. */
    actionEnableYM2413->setDisabled(true);
    actionEnablePSG->setDisabled(true);
    actionEnableCheats->setDisabled(true);
    actionDebugger->setDisabled(true);
    menuRegion->setDisabled(true);
    menuPSG_Channels->setDisabled(true);
    menuVideo_System->setDisabled(true);

    if(nes_initialized) {
        saveSRAM();

        nes_shutdown();
        nes_init(videoSystem);
        updateFps();
    }
    else {
        if(nes_init(videoSystem))
            return -1;
        updateFps();
    }

    nes_mem_load_rom(filename.toUtf8().data());
    resizeEmu(256, 240);
    loadSRAM();
    return 0;
}

void UICrabEmu::loadRom(QString filename) {
    QStringList romPath;
    QStringList romExts;
    QFile file(filename);
    int console;

    /* Kill off the emulation thread... we'll restart it as needed. */
    thd->finish();
    thd->wait();

    QMutexLocker locker(&memoryLock);

    console = rom_detect_console(filename.toUtf8().data());

    if(console < 0) {
         QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
                "ROM file", "CrabEmu cannot open " + filename + "\n"
                "Unrecognized file type.");
         return;
    }

    if(console == CONSOLE_COLECOVISION) {
        if(initializeCV(filename))
            return;
    }
    else if(console == CONSOLE_NES) {
        if(initializeNES(filename))
            return;
    }
    else {
        if(initializeSMS(filename, console))
            return;
    }

    /* Get the rom name by splitting on the path separator...
        This may not be platform independent... */
    romPath = filename.split("/");
    romExts = filename.split(".");
    romName = romPath.last();
    romExt = romExts.last();
    openRom = filename;

    sound_unpause();
    thd->restart();
}

void UICrabEmu::on_actionOpenRom_triggered() {
    QString fn;

    fn = QFileDialog::getOpenFileName(QApplication::activeWindow(),
                    tr("Open Rom"), "",
                    tr("Roms (*.sc *.sg *.sms *.gg *.rom *.col *.nes *.gz *.bz2 *.zip)"), 0,
                    // this is obsolete & does nothing since Qt 4.5
                    QFileDialog::DontUseSheet);

    /* If a file was not selected, then don't try to open one. */
    if(fn.isEmpty())
        return;

    loadRom(fn);
}

void UICrabEmu::on_actionSaveSRAM_triggered() {
    saveSRAM();
}

void UICrabEmu::loadSRAM() {
    QSettings settings;

    if(sms_initialized) {
        if(settings.value("Save Directory Enabled").toBool())
            sms_read_cartram_from_file(
                (settings.value("Save Directory").toString() + "/" +
                romName + ".ces").toUtf8().data());
        else
            sms_read_cartram_from_file((openRom + ".ces").toUtf8().data());
    }
    else if(nes_initialized) {
        if(settings.value("Save Directory Enabled").toBool())
            nes_mem_read_sram(
            (settings.value("Save Directory").toString() + "/" +
            romName + ".ces").toUtf8().data());
        else
            nes_mem_read_sram((openRom + ".ces").toUtf8().data());
    }
}

void UICrabEmu::saveSRAM() {
    QSettings settings;

    if(sms_initialized) {
        if(settings.value("Save Directory Enabled").toBool())
            sms_write_cartram_to_file(
            (settings.value("Save Directory").toString() + "/" +
            romName + ".ces").toUtf8().data());
        else
            sms_write_cartram_to_file((openRom + ".ces").toUtf8().data());
    }
    else if(nes_initialized) {
        if(settings.value("Save Directory Enabled").toBool())
            nes_mem_write_sram(
            (settings.value("Save Directory").toString() + "/" +
            romName + ".ces").toUtf8().data());
        else
            nes_mem_write_sram((openRom + ".ces").toUtf8().data());
    }
}

void UICrabEmu::saveLoad(QString state, bool load) {
    QString filename;
    QSettings settings;
    int (*save_func)(const char *);
    int (*load_func)(const char *);

    if(!sms_initialized && !coleco_initialized && !nes_initialized)
        return;

    if(sms_initialized) {
        save_func = &sms_save_state;
        load_func = &sms_load_state;
    }
    else if(coleco_initialized) {
        save_func = &coleco_save_state;
        load_func = &coleco_load_state;
    }
    else if(nes_initialized) {
        save_func = &nes_save_state;
        load_func = &nes_load_state;
    }
    else {
        // huh?
        return;
    }

    if(settings.value("State Directory Enabled").toBool())
        filename = settings.value("State Directory").toString() + "/" +
            romName + "." + state + ".cestate";
    else
        filename = openRom + "." + state + ".cestate";

    QMutexLocker locker(&memoryLock);

    if(load) {
        if(load_func(filename.toUtf8().data())) {
            QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
                "state file", "CrabEmu cannot open " + filename + " for "
                "reading. Verify that you have the permissions to "
                "read to this directory and that the save state exists.");
        }
    }
    else {
        if(save_func(filename.toUtf8().data())) {
            QMessageBox::warning(QApplication::activeWindow(), "Cannot open "
            "state file", "CrabEmu cannot open " + filename +" for "
            "writing. Verify that you have the permissions to "
            "write to this directory and that the target "
            "volume is not full.");
        }
    }
}

void UICrabEmu::on_actionSaveSlot1_triggered() {
    saveLoad("1", false);
}

void UICrabEmu::on_actionSaveSlot2_triggered() {
    saveLoad("2", false);
}

void UICrabEmu::on_actionSaveSlot3_triggered() {
    saveLoad("3", false);
}

void UICrabEmu::on_actionSaveSlot4_triggered() {
    saveLoad("4", false);
}

void UICrabEmu::on_actionSaveSlot5_triggered() {
    saveLoad("5", false);
}

void UICrabEmu::on_actionSaveSlot6_triggered() {
    saveLoad("6", false);
}

void UICrabEmu::on_actionLoadSlot1_triggered() {
    saveLoad("1", true);
}

void UICrabEmu::on_actionLoadSlot2_triggered() {
    saveLoad("2", true);
}

void UICrabEmu::on_actionLoadSlot3_triggered() {
    saveLoad("3", true);
}

void UICrabEmu::on_actionLoadSlot4_triggered() {
    saveLoad("4", true);
}

void UICrabEmu::on_actionLoadSlot5_triggered() {
    saveLoad("5", true);
}

void UICrabEmu::on_actionLoadSlot6_triggered() {
    saveLoad("6", true);
}

void UICrabEmu::on_actionTakeScreenshot_triggered() {
    QMutexLocker locker(&memoryLock);

    if(!sms_initialized && !coleco_initialized && !nes_initialized) {
        return;
    }

    QImage *data;

    QImage img;
    QString fn = QFileDialog::getSaveFileName(QApplication::activeWindow(),
                tr("Save Screenshot As"), "",
                tr("PNG Images (*.png)"), 0,
                QFileDialog::DontUseSheet);

    if(fn.isEmpty()) {
        return;
    }

    if(!fn.endsWith(".png", Qt::CaseInsensitive)) {
        fn += ".png";
    }

    data = emuwidget->retainImage();

    if(sms_console == CONSOLE_GG) {
        img = data->copy(48, 24, 160, 144);
    }
    else if(sms_console == CONSOLE_NES) {
        img = data->copy(0, 0, 256, 240);
    }
    else {
        img = data->copy(0, 0, 256, smsvdp.lines);
    }

    img.save(fn, "PNG");

    emuwidget->releaseImage();
}

void UICrabEmu::on_actionExit_triggered() {
    if(debugger)
        debugWindow->close();

    close();
}

void UICrabEmu::on_actionEnablePSG_triggered() {
    if(!actionEnablePSG->isChecked()) {
        sms_psg_enabled = 0;
    }
    else {
        sms_psg_enabled = 1;
    }
}

void UICrabEmu::on_actionEnableYM2413_triggered() {
    if(!actionEnableYM2413->isChecked()) {
        sms_ym2413_enabled = 0;
    }
    else {
        sms_ym2413_enabled = 1;
    }
}

void UICrabEmu::on_actionTone0_triggered() {
    if(!actionTone0->isChecked()) {
        psg.enabled_channels &= ~ENABLE_TONE0;
    }
    else {
        psg.enabled_channels |= ENABLE_TONE0;
    }
}

void UICrabEmu::on_actionTone1_triggered() {
    if(!actionTone1->isChecked()) {
        psg.enabled_channels &= ~ENABLE_TONE1;
    }
    else {
        psg.enabled_channels |= ENABLE_TONE1;
    }
}

void UICrabEmu::on_actionTone2_triggered() {
    if(!actionTone2->isChecked()) {
        psg.enabled_channels &= ~ENABLE_TONE2;
    }
    else {
        psg.enabled_channels |= ENABLE_TONE2;
    }
}

void UICrabEmu::on_actionNoise_triggered() {
    if(!actionNoise->isChecked()) {
        psg.enabled_channels &= ~ENABLE_NOISE;
    }
    else {
        psg.enabled_channels |= ENABLE_NOISE;
    }
}

void UICrabEmu::on_actionExport_triggered() {
    if(!actionExport->isChecked()) {
        actionExport->setChecked(true);
        return;
    }

    actionDomestic->setChecked(false);

    region = SMS_REGION_EXPORT;

    QSettings settings;
    settings.setValue("SMS Domestic", false);
}

void UICrabEmu::on_actionDomestic_triggered() {
    if(!actionDomestic->isChecked()) {
        actionDomestic->setChecked(true);
        return;
    }

    actionExport->setChecked(false);

    region = SMS_REGION_DOMESTIC;

    QSettings settings;
    settings.setValue("SMS Domestic", true);
}

void UICrabEmu::on_actionNTSC_triggered() {
    if(!actionNTSC->isChecked()) {
        actionNTSC->setChecked(true);
        return;
    }

    actionPAL->setChecked(false);

    videoSystem = SMS_VIDEO_NTSC;

    QSettings settings;
    settings.setValue("SMS PAL", false);
}

void UICrabEmu::on_actionPAL_triggered() {
    if(!actionPAL->isChecked()) {
        actionPAL->setChecked(true);
        return;
    }

    actionNTSC->setChecked(false);

    videoSystem = SMS_VIDEO_PAL;

    QSettings settings;
    settings.setValue("SMS PAL", true);
}

void UICrabEmu::on_actionEnableCheats_triggered() {
    if(!sms_initialized) {
        actionEnableCheats->setChecked(false);
        return;
    }

    if(!actionEnableCheats->isChecked()) {
        sms_cheat_disable();
    }
    else {
        sms_cheat_enable();
    }
}

void UICrabEmu::on_actionNormal_triggered() {
    QMutexLocker l(&memoryLock);
    emuwidget->setRenderer(Render_Normal);
}

void UICrabEmu::on_actionScale2x_triggered() {
    QMutexLocker l(&memoryLock);
    emuwidget->setRenderer(Render_Scale2x);
}

void UICrabEmu::on_actionHQ2x_triggered() {
    QMutexLocker l(&memoryLock);
    emuwidget->setRenderer(Render_HQ2x);
}

void UICrabEmu::on_actionHQ4x_triggered() {
    QMutexLocker l(&memoryLock);
    emuwidget->setRenderer(Render_HQ4x);
}

#if 0
void UICrabEmu::on_action1X_triggered()
{
    videoMultiple = 1;

    if(!sms_initialized && !coleco_initialized && !nes_initialized) {
        return;
    }

    if (sms_console == CONSOLE_GG) {
        resizeEmu(160, 144);
    }
    else if(sms_console == CONSOLE_NES) {
        resizeEmu(256, 240);
    }
    else {
        resizeEmu(256, smsvdp.lines);
    }
}

void UICrabEmu::on_action2X_triggered()
{
    videoMultiple = 2;

    if(!sms_initialized && !coleco_initialized && !nes_initialized) {
        return;
    }

    if (sms_console == CONSOLE_GG) {
        resizeEmu(160, 144);
    }
    else if(sms_console == CONSOLE_NES) {
        resizeEmu(256, 240);
    }
    else {
        resizeEmu(256, smsvdp.lines);
    }
}
#endif

void UICrabEmu::on_actionPause_triggered() {
    if(!sms_initialized && !coleco_initialized && !nes_initialized) {
        actionPause->setChecked(false);
        return;
    }

    if(!paused) {
        memoryLock.lock();
        sound_pause();
        paused = true;
    }
    else {
        paused = false;
        sound_unpause();
        memoryLock.unlock();
    }

    actionPause->setChecked(paused);

    if(debugger) {
        connect(this, SIGNAL(pause(bool)), debugWindow, SLOT(on_main_pause(bool)));
        emit pause(paused);
    }
}

void UICrabEmu::on_actionReset_triggered() {
    if(!sms_initialized && !coleco_initialized && !nes_initialized)
        return;

    if(paused) {
        on_actionPause_triggered();
    }

    if(sms_console == CONSOLE_COLECOVISION) {
        if(sms_region & videoSystem) {
            coleco_soft_reset();
        }
        else {
            on_actionReloadRom_triggered();
        }
    }
    else if(sms_console == CONSOLE_NES) {
        nes_soft_reset();
    }
    else {
        if(sms_region & region && sms_region & videoSystem) {
            sms_soft_reset();
        }
        else {
            // We've changed the region or video system, so do a full reset
            on_actionReloadRom_triggered();
        }
    }
}

void UICrabEmu::on_actionShutdown_triggered() {
    if(!sms_initialized && !coleco_initialized && !nes_initialized)
        return;

    unpause();
    saveSRAM();
    thd->finish();
    thd->wait();

    if(sms_console == CONSOLE_COLECOVISION) {
        coleco_shutdown();
    }
    else if(sms_console == CONSOLE_NES) {
        nes_shutdown();
    }
    else {
        sms_shutdown();
    }
}

void UICrabEmu::on_actionReloadRom_triggered() {
    loadRom(openRom);
}

void UICrabEmu::on_actionPreferences_triggered() {
    UIPreferences prefs(this);

    if(!paused) {
        memoryLock.lock();
    }

    int result = prefs.exec();

    if(result == QDialog::Accepted) {
        prefs.updatePreferences();
    }

    if(!paused) {
        memoryLock.unlock();
    }
}

void UICrabEmu::on_actionDebugger_triggered() {
    if(coleco_initialized) {
        return;
    }

    if(debugger == false) {
        debugger = true;
        debugWindow = new UIDebugWindow(thread(), this);
        debugWindow->restoreSettings();
        debugWindow->show();
    }
    else {
        debugger = false;
        debugWindow->close();
        delete debugWindow;
        debugWindow = NULL;
    }
}

void UICrabEmu::on_actionAbout_triggered() {
    if(!paused) {
        memoryLock.lock();
    }

    UIAboutDialog(window()).exec();

    if(!paused) {
        memoryLock.unlock();
    }
}

void UICrabEmu::on_Debugger_destroyed() {
    debugger = false;
    debugWindow = NULL;
    actionDebugger->setChecked(false);
}

void UICrabEmu::keyPressEvent(QKeyEvent *e) {
    /* Ignore auto-repeat events. */
    if(e->isAutoRepeat())
        return;

    if(sms_console == CONSOLE_COLECOVISION) {
        colecoKeyPressEvent(e);
    }
    else if(sms_console == CONSOLE_NES) {
        nesKeyPressEvent(e);
    }
    else {
        smsKeyPressEvent(e);
    }
}

void UICrabEmu::smsKeyPressEvent(QKeyEvent* e) {
    int button = -1;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("SMS Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();

    /* Check the key in the event against each one in the 1st player's keyboard
       mapping. */
    if(e->key() == kb1.value("SMS Up").toInt()) {
        button = SMS_UP;
    }
    else if(e->key() == kb1.value("SMS Down").toInt()) {
        button = SMS_DOWN;
    }
    else if(e->key() == kb1.value("SMS Left").toInt()) {
        button = SMS_LEFT;
    }
    else if(e->key() == kb1.value("SMS Right").toInt()) {
        button = SMS_RIGHT;
    }
    else if(e->key() == kb1.value("SMS Button 1").toInt()) {
        button = SMS_BUTTON_1;
    }
    else if(e->key() == kb1.value("SMS Button 2").toInt()) {
        button = SMS_BUTTON_2;
    }
    else if(e->key() == kb1.value("Game Gear Start").toInt()) {
        button = GAMEGEAR_START;
    }

    if(button != -1) {
        sms_button_pressed(1, button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
       mapping. */
    if(e->key() == kb2.value("SMS Up").toInt()) {
        button = SMS_UP;
    }
    else if(e->key() == kb2.value("SMS Down").toInt()) {
        button = SMS_DOWN;
    }
    else if(e->key() == kb2.value("SMS Left").toInt()) {
        button = SMS_LEFT;
    }
    else if(e->key() == kb2.value("SMS Right").toInt()) {
        button = SMS_RIGHT;
    }
    else if(e->key() == kb2.value("SMS Button 1").toInt()) {
        button = SMS_BUTTON_1;
    }
    else if(e->key() == kb2.value("SMS Button 2").toInt()) {
        button = SMS_BUTTON_2;
    }

    if(button != -1) {
        sms_button_pressed(2, button);
    }
}

void UICrabEmu::colecoKeyPressEvent(QKeyEvent* e) {
    int button = -1;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("Coleco Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();

    /* Check the key in the event against each one in the 1st player's keyboard
     *  mapping. */
    if(e->key() == kb1.value("Coleco Up").toInt()) {
        button = COLECOVISION_UP;
    }
    else if(e->key() == kb1.value("Coleco Down").toInt()) {
        button = COLECOVISION_DOWN;
    }
    else if(e->key() == kb1.value("Coleco Left").toInt()) {
        button = COLECOVISION_LEFT;
    }
    else if(e->key() == kb1.value("Coleco Right").toInt()) {
        button = COLECOVISION_RIGHT;
    }
    else if(e->key() == kb1.value("Coleco L Action").toInt()) {
        button = COLECOVISION_L_ACTION;
    }
    else if(e->key() == kb1.value("Coleco R Action").toInt()) {
        button = COLECOVISION_R_ACTION;
    }
    else if(e->key() == kb1.value("Coleco 1").toInt()) {
        button = COLECOVISION_1;
    }
    else if(e->key() == kb1.value("Coleco 2").toInt()) {
        button = COLECOVISION_2;
    }
    else if(e->key() == kb1.value("Coleco 3").toInt()) {
        button = COLECOVISION_3;
    }
    else if(e->key() == kb1.value("Coleco 4").toInt()) {
        button = COLECOVISION_4;
    }
    else if(e->key() == kb1.value("Coleco 5").toInt()) {
        button = COLECOVISION_5;
    }
    else if(e->key() == kb1.value("Coleco 6").toInt()) {
        button = COLECOVISION_6;
    }
    else if(e->key() == kb1.value("Coleco 7").toInt()) {
        button = COLECOVISION_7;
    }
    else if(e->key() == kb1.value("Coleco 8").toInt()) {
        button = COLECOVISION_8;
    }
    else if(e->key() == kb1.value("Coleco 9").toInt()) {
        button = COLECOVISION_9;
    }
    else if(e->key() == kb1.value("Coleco *").toInt()) {
        button = COLECOVISION_STAR;
    }
    else if(e->key() == kb1.value("Coleco 0").toInt()) {
        button = COLECOVISION_0;
    }
    else if(e->key() == kb1.value("Coleco #").toInt()) {
        button = COLECOVISION_POUND;
    }

    if(button != -1) {
        coleco_button_pressed(1, button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
     *  mapping. */
    if(e->key() == kb2.value("Coleco Up").toInt()) {
        button = COLECOVISION_UP;
    }
    else if(e->key() == kb2.value("Coleco Down").toInt()) {
        button = COLECOVISION_DOWN;
    }
    else if(e->key() == kb2.value("Coleco Left").toInt()) {
        button = COLECOVISION_LEFT;
    }
    else if(e->key() == kb2.value("Coleco Right").toInt()) {
        button = COLECOVISION_RIGHT;
    }
    else if(e->key() == kb2.value("Coleco L Action").toInt()) {
        button = COLECOVISION_L_ACTION;
    }
    else if(e->key() == kb2.value("Coleco R Action").toInt()) {
        button = COLECOVISION_R_ACTION;
    }
    else if(e->key() == kb2.value("Coleco 1").toInt()) {
        button = COLECOVISION_1;
    }
    else if(e->key() == kb2.value("Coleco 2").toInt()) {
        button = COLECOVISION_2;
    }
    else if(e->key() == kb2.value("Coleco 3").toInt()) {
        button = COLECOVISION_3;
    }
    else if(e->key() == kb2.value("Coleco 4").toInt()) {
        button = COLECOVISION_4;
    }
    else if(e->key() == kb2.value("Coleco 5").toInt()) {
        button = COLECOVISION_5;
    }
    else if(e->key() == kb2.value("Coleco 6").toInt()) {
        button = COLECOVISION_6;
    }
    else if(e->key() == kb2.value("Coleco 7").toInt()) {
        button = COLECOVISION_7;
    }
    else if(e->key() == kb2.value("Coleco 8").toInt()) {
        button = COLECOVISION_8;
    }
    else if(e->key() == kb2.value("Coleco 9").toInt()) {
        button = COLECOVISION_9;
    }
    else if(e->key() == kb2.value("Coleco *").toInt()) {
        button = COLECOVISION_STAR;
    }
    else if(e->key() == kb2.value("Coleco 0").toInt()) {
        button = COLECOVISION_0;
    }
    else if(e->key() == kb2.value("Coleco #").toInt()) {
        button = COLECOVISION_POUND;
    }

    if(button != -1) {
        coleco_button_pressed(2, button);
    }
}

void UICrabEmu::nesKeyPressEvent(QKeyEvent* e) {
    int button = -1;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("NES Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();

    /* Check the key in the event against each one in the 1st player's keyboard
       mapping. */
    if(e->key() == kb1.value("Up").toInt())
        button = NES_UP;
    else if(e->key() == kb1.value("Down").toInt())
        button = NES_DOWN;
    else if(e->key() == kb1.value("Left").toInt())
        button = NES_LEFT;
    else if(e->key() == kb1.value("Right").toInt())
        button = NES_RIGHT;
    else if(e->key() == kb1.value("A").toInt())
        button = NES_A;
    else if(e->key() == kb1.value("B").toInt())
        button = NES_B;
    else if(e->key() == kb1.value("Start").toInt())
        button = NES_START;
    else if(e->key() == kb1.value("Select").toInt())
        button = NES_SELECT;

    if(button != -1) {
        nes_button_pressed(1, button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
       mapping. */
    if(e->key() == kb2.value("Up").toInt())
        button = NES_UP;
    else if(e->key() == kb2.value("Down").toInt())
        button = NES_DOWN;
    else if(e->key() == kb2.value("Left").toInt())
        button = NES_LEFT;
    else if(e->key() == kb2.value("Right").toInt())
        button = NES_RIGHT;
    else if(e->key() == kb2.value("A").toInt())
        button = NES_A;
    else if(e->key() == kb2.value("B").toInt())
        button = NES_B;
    else if(e->key() == kb2.value("Start").toInt())
        button = NES_START;
    else if(e->key() == kb2.value("Select").toInt())
        button = NES_SELECT;

    if(button != -1)
        nes_button_pressed(2, button);
}

void UICrabEmu::keyReleaseEvent(QKeyEvent *e) {
    /* Ignore auto-repeat events. */
    if(e->isAutoRepeat())
        return;

    if(sms_console == CONSOLE_COLECOVISION) {
        colecoKeyReleaseEvent(e);
    }
    else if(sms_console == CONSOLE_NES) {
        nesKeyReleaseEvent(e);
    }
    else {
        smsKeyReleaseEvent(e);
    }
}

void UICrabEmu::smsKeyReleaseEvent(QKeyEvent* e) {
    int button = -1;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("SMS Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();


    /* Check the key in the event against each one in the 1st player's keyboard
       mapping. */
    if(e->key() == kb1.value("SMS Up").toInt()) {
        button = SMS_UP;
    }
    else if(e->key() == kb1.value("SMS Down").toInt()) {
        button = SMS_DOWN;
    }
    else if(e->key() == kb1.value("SMS Left").toInt()) {
        button = SMS_LEFT;
    }
    else if(e->key() == kb1.value("SMS Right").toInt()) {
        button = SMS_RIGHT;
    }
    else if(e->key() == kb1.value("SMS Button 1").toInt()) {
        button = SMS_BUTTON_1;
    }
    else if(e->key() == kb1.value("SMS Button 2").toInt()) {
        button = SMS_BUTTON_2;
    }
    else if(e->key() == kb1.value("Game Gear Start").toInt()) {
        button = GAMEGEAR_START;
    }

    if(button != -1) {
        sms_button_released(1, button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
       mapping. */
    if(e->key() == kb2.value("SMS Up").toInt()) {
        button = SMS_UP;
    }
    else if(e->key() == kb2.value("SMS Down").toInt()) {
        button = SMS_DOWN;
    }
    else if(e->key() == kb2.value("SMS Left").toInt()) {
        button = SMS_LEFT;
    }
    else if(e->key() == kb2.value("SMS Right").toInt()) {
        button = SMS_RIGHT;
    }
    else if(e->key() == kb2.value("SMS Button 1").toInt()) {
        button = SMS_BUTTON_1;
    }
    else if(e->key() == kb2.value("SMS Button 2").toInt()) {
        button = SMS_BUTTON_2;
    }

    if(button != -1) {
        sms_button_released(2, button);
    }
}

void UICrabEmu::colecoKeyReleaseEvent(QKeyEvent* e) {
    int button = -1;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("Coleco Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();

    /* Check the key in the event against each one in the 1st player's keyboard
     *  mapping. */
    if(e->key() == kb1.value("Coleco Up").toInt()) {
        button = COLECOVISION_UP;
    }
    else if(e->key() == kb1.value("Coleco Down").toInt()) {
        button = COLECOVISION_DOWN;
    }
    else if(e->key() == kb1.value("Coleco Left").toInt()) {
        button = COLECOVISION_LEFT;
    }
    else if(e->key() == kb1.value("Coleco Right").toInt()) {
        button = COLECOVISION_RIGHT;
    }
    else if(e->key() == kb1.value("Coleco L Action").toInt()) {
        button = COLECOVISION_L_ACTION;
    }
    else if(e->key() == kb1.value("Coleco R Action").toInt()) {
        button = COLECOVISION_R_ACTION;
    }
    else if(e->key() == kb1.value("Coleco 1").toInt()) {
        button = COLECOVISION_1;
    }
    else if(e->key() == kb1.value("Coleco 2").toInt()) {
        button = COLECOVISION_2;
    }
    else if(e->key() == kb1.value("Coleco 3").toInt()) {
        button = COLECOVISION_3;
    }
    else if(e->key() == kb1.value("Coleco 4").toInt()) {
        button = COLECOVISION_4;
    }
    else if(e->key() == kb1.value("Coleco 5").toInt()) {
        button = COLECOVISION_5;
    }
    else if(e->key() == kb1.value("Coleco 6").toInt()) {
        button = COLECOVISION_6;
    }
    else if(e->key() == kb1.value("Coleco 7").toInt()) {
        button = COLECOVISION_7;
    }
    else if(e->key() == kb1.value("Coleco 8").toInt()) {
        button = COLECOVISION_8;
    }
    else if(e->key() == kb1.value("Coleco 9").toInt()) {
        button = COLECOVISION_9;
    }
    else if(e->key() == kb1.value("Coleco *").toInt()) {
        button = COLECOVISION_STAR;
    }
    else if(e->key() == kb1.value("Coleco 0").toInt()) {
        button = COLECOVISION_0;
    }
    else if(e->key() == kb1.value("Coleco #").toInt()) {
        button = COLECOVISION_POUND;
    }

    if(button != -1) {
        coleco_button_released(1, button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
     *  mapping. */
    if(e->key() == kb2.value("Coleco Up").toInt()) {
        button = COLECOVISION_UP;
    }
    else if(e->key() == kb2.value("Coleco Down").toInt()) {
        button = COLECOVISION_DOWN;
    }
    else if(e->key() == kb2.value("Coleco Left").toInt()) {
        button = COLECOVISION_LEFT;
    }
    else if(e->key() == kb2.value("Coleco Right").toInt()) {
        button = COLECOVISION_RIGHT;
    }
    else if(e->key() == kb2.value("Coleco L Action").toInt()) {
        button = COLECOVISION_L_ACTION;
    }
    else if(e->key() == kb2.value("Coleco R Action").toInt()) {
        button = COLECOVISION_R_ACTION;
    }
    else if(e->key() == kb2.value("Coleco 1").toInt()) {
        button = COLECOVISION_1;
    }
    else if(e->key() == kb2.value("Coleco 2").toInt()) {
        button = COLECOVISION_2;
    }
    else if(e->key() == kb2.value("Coleco 3").toInt()) {
        button = COLECOVISION_3;
    }
    else if(e->key() == kb2.value("Coleco 4").toInt()) {
        button = COLECOVISION_4;
    }
    else if(e->key() == kb2.value("Coleco 5").toInt()) {
        button = COLECOVISION_5;
    }
    else if(e->key() == kb2.value("Coleco 6").toInt()) {
        button = COLECOVISION_6;
    }
    else if(e->key() == kb2.value("Coleco 7").toInt()) {
        button = COLECOVISION_7;
    }
    else if(e->key() == kb2.value("Coleco 8").toInt()) {
        button = COLECOVISION_8;
    }
    else if(e->key() == kb2.value("Coleco 9").toInt()) {
        button = COLECOVISION_9;
    }
    else if(e->key() == kb2.value("Coleco *").toInt()) {
        button = COLECOVISION_STAR;
    }
    else if(e->key() == kb2.value("Coleco 0").toInt()) {
        button = COLECOVISION_0;
    }
    else if(e->key() == kb2.value("Coleco #").toInt()) {
        button = COLECOVISION_POUND;
    }

    if(button != -1) {
        coleco_button_released(2, button);
    }
}

void UICrabEmu::nesKeyReleaseEvent(QKeyEvent* e) {
    int button = -1;
    QSettings settings;
    QVariantMap buttonmap, kb1, kb2;

    /* Grab the button maps from the settings files. */
    buttonmap = settings.value("NES Buttonmap").toMap();
    kb1 = buttonmap.value("Qt Keyboard 1").toMap();
    kb2 = buttonmap.value("Qt Keyboard 2").toMap();

    /* Check the key in the event against each one in the 1st player's keyboard
       mapping. */
    if(e->key() == kb1.value("Up").toInt())
        button = NES_UP;
    else if(e->key() == kb1.value("Down").toInt())
        button = NES_DOWN;
    else if(e->key() == kb1.value("Left").toInt())
        button = NES_LEFT;
    else if(e->key() == kb1.value("Right").toInt())
        button = NES_RIGHT;
    else if(e->key() == kb1.value("A").toInt())
        button = NES_A;
    else if(e->key() == kb1.value("B").toInt())
        button = NES_B;
    else if(e->key() == kb1.value("Start").toInt())
        button = NES_START;
    else if(e->key() == kb1.value("Select").toInt())
        button = NES_SELECT;

    if(button != -1) {
        nes_button_released(1, button);
        return;
    }

    /* Check the key in the event against each one in the 2nd player's keyboard
       mapping. */
    if(e->key() == kb2.value("Up").toInt())
        button = NES_UP;
    else if(e->key() == kb2.value("Down").toInt())
        button = NES_DOWN;
    else if(e->key() == kb2.value("Left").toInt())
        button = NES_LEFT;
    else if(e->key() == kb2.value("Right").toInt())
        button = NES_RIGHT;
    else if(e->key() == kb2.value("A").toInt())
        button = NES_A;
    else if(e->key() == kb2.value("B").toInt())
        button = NES_B;
    else if(e->key() == kb2.value("Start").toInt())
        button = NES_START;
    else if(e->key() == kb2.value("Select").toInt())
        button = NES_SELECT;

    if(button != -1)
        nes_button_released(2, button);
}

void UICrabEmu::initializeSettings() {
    QSettings settings;
    QVariantMap sms_buttonmap, sms_kb1, sms_kb2;
    QVariantMap coleco_buttonmap, coleco_kb1, coleco_kb2;
    QVariantMap nes_buttonmap, nes_kb1, nes_kb2;

    settings.setValue("SMS BIOS ROM", "");
    settings.setValue("SMS Domestic", false);
    settings.setValue("SMS PAL", false);

    sms_kb1["SMS Up"] = Qt::Key_Up;
    sms_kb1["SMS Down"] = Qt::Key_Down;
    sms_kb1["SMS Left"] = Qt::Key_Left;
    sms_kb1["SMS Right"] = Qt::Key_Right;
    sms_kb1["SMS Button 1"] = Qt::Key_Semicolon;
    sms_kb1["SMS Button 2"] = Qt::Key_Apostrophe;
    sms_kb1["Game Gear Start"] = Qt::Key_P;

    sms_kb2["SMS Up"] = Qt::Key_W;
    sms_kb2["SMS Down"] = Qt::Key_S;
    sms_kb2["SMS Left"] = Qt::Key_A;
    sms_kb2["SMS Right"] = Qt::Key_D;
    sms_kb2["SMS Button 1"] = Qt::Key_V;
    sms_kb2["SMS Button 2"] = Qt::Key_B;
    sms_kb2["Game Gear Start"] = Qt::Key_R;

    sms_buttonmap["Qt Keyboard 1"] = sms_kb1;
    sms_buttonmap["Qt Keyboard 2"] = sms_kb2;
    settings.setValue("SMS Buttonmap", sms_buttonmap);

    coleco_kb1["Coleco Up"] = Qt::Key_W;
    coleco_kb1["Coleco Left"] = Qt::Key_A;
    coleco_kb1["Coleco Down"] = Qt::Key_S;
    coleco_kb1["Coleco Right"] = Qt::Key_D;
    coleco_kb1["Coleco L Action"] = Qt::Key_Q;
    coleco_kb1["Coleco R Action"] = Qt::Key_E;
    coleco_kb1["Coleco 1"] = Qt::Key_4;
    coleco_kb1["Coleco 2"] = Qt::Key_5;
    coleco_kb1["Coleco 3"] = Qt::Key_6;
    coleco_kb1["Coleco 4"] = Qt::Key_R;
    coleco_kb1["Coleco 5"] = Qt::Key_T;
    coleco_kb1["Coleco 6"] = Qt::Key_Y;
    coleco_kb1["Coleco 7"] = Qt::Key_F;
    coleco_kb1["Coleco 8"] = Qt::Key_G;
    coleco_kb1["Coleco 9"] = Qt::Key_H;
    coleco_kb1["Coleco *"] = Qt::Key_V;
    coleco_kb1["Coleco 0"] = Qt::Key_B;
    coleco_kb1["Coleco #"] = Qt::Key_N;

    coleco_kb2["Coleco Up"] = Qt::Key_Up;
    coleco_kb2["Coleco Left"] = Qt::Key_Left;
    coleco_kb2["Coleco Down"] = Qt::Key_Down;
    coleco_kb2["Coleco Right"] = Qt::Key_Right;
    coleco_kb2["Coleco L Action"] = Qt::Key_Minus;
    coleco_kb2["Coleco R Action"] = Qt::Key_Equal;
    coleco_kb2["Coleco 1"] = Qt::Key_8;
    coleco_kb2["Coleco 2"] = Qt::Key_9;
    coleco_kb2["Coleco 3"] = Qt::Key_0;
    coleco_kb2["Coleco 4"] = Qt::Key_I;
    coleco_kb2["Coleco 5"] = Qt::Key_O;
    coleco_kb2["Coleco 6"] = Qt::Key_P;
    coleco_kb2["Coleco 7"] = Qt::Key_K;
    coleco_kb2["Coleco 8"] = Qt::Key_L;
    coleco_kb2["Coleco 9"] = Qt::Key_Semicolon;
    coleco_kb2["Coleco *"] = Qt::Key_Comma;
    coleco_kb2["Coleco 0"] = Qt::Key_Period;
    coleco_kb2["Coleco #"] = Qt::Key_Slash;

    coleco_buttonmap["Qt Keyboard 1"] = coleco_kb1;
    coleco_buttonmap["Qt Keyboard 2"] = coleco_kb2;
    settings.setValue("Coleco Buttonmap", coleco_buttonmap);

    nes_kb1["Up"] = Qt::Key_Up;
    nes_kb1["Down"] = Qt::Key_Down;
    nes_kb1["Left"] = Qt::Key_Left;
    nes_kb1["Right"] = Qt::Key_Right;
    nes_kb1["A"] = Qt::Key_Apostrophe;
    nes_kb1["B"] = Qt::Key_Semicolon;
    nes_kb1["Start"] = Qt::Key_P;
    nes_kb1["Select"] = Qt::Key_O;

    nes_kb2["Up"] = Qt::Key_W;
    nes_kb2["Down"] = Qt::Key_S;
    nes_kb2["Left"] = Qt::Key_A;
    nes_kb2["Right"] = Qt::Key_D;
    nes_kb2["A"] = Qt::Key_B;
    nes_kb2["B"] = Qt::Key_V;
    nes_kb2["Start"] = Qt::Key_E;
    nes_kb2["Select"] = Qt::Key_Q;

    nes_buttonmap["Qt Keyboard 1"] = nes_kb1;
    nes_buttonmap["Qt Keyboard 2"] = nes_kb2;
    settings.setValue("NES Buttonmap", nes_buttonmap);
}

void UICrabEmu::resizeEmu(unsigned int w, unsigned int h)
{
    emuwidget->setMinimumSize(w * videoMultiple, h * videoMultiple);
    emuwidget->resize(0, 0);
    resize(0, 0);
}

void UICrabEmu::unpause(bool p)
{
    if(p && paused) {
        on_actionPause_triggered();
    }
    if(!p && !paused) {
        on_actionPause_triggered();
    }
}

void UICrabEmu::updateFps() {
    if(videoSystem & SMS_VIDEO_NTSC) {
        thd->setFps(60.0);
    }
    else {
        thd->setFps(50.0);
    }
}
