/*
    This file is part of CrabEmu

    Copyright (C) 2009,2011,2012 Harley Laue

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "EmuWidget.h"
#include "UIDebugWindow.h"
#include "UICrabEmu.h"
#include "CrabZ80.h"
#include "CrabZ80d.h"
#include "sms.h"
#include "smsz80.h"
#include "smsvdp.h"
#include "sn76489.h"
#include <cstdio>
#include <QSettings>

extern "C" int sms_initialized, coleco_initialized;
extern "C" CrabZ80_t *cpuz80;
extern "C" sn76489_t psg;
extern int emu_frame(int run);
extern void render_scene(void);

UIDebugWindow::UIDebugWindow(CrabEmuThread *thd, UICrabEmu* parent, bool paused): QMainWindow(parent) {
    this->paused = paused;
    this->parent = parent;

    setAttribute(Qt::WA_DeleteOnClose);

    connect(this, SIGNAL(togglePaused()), parent, SLOT(on_actionPause_triggered()));
    connect(this, SIGNAL(destroyed(QObject *)), parent, SLOT(on_Debugger_destroyed()));
    connect(this, SIGNAL(reset()), parent, SLOT(on_actionReset_triggered()));
    connect(this, SIGNAL(reloadRom()), parent, SLOT(on_actionReloadRom_triggered()));
    connect(thd, SIGNAL(frameRendered()), this, SLOT(update()));
    setupUi(this);

    // Disable unimplemented features
    ramDock->hide();
    debuggerLineEdit->hide();
    actionRunToCursor->setVisible(false);

    QLabel *q = new QLabel;
    vdpTilesDock->setWidget(q);

    resetSDSC();
}

void UIDebugWindow::restoreSettings() {
    QSettings settings;
    settings.beginGroup("DebugWindow");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
}

void UIDebugWindow::updateSDSC(const char console[25][80 * 2]) {
    for(int x = 0; x < 25; ++x) {
        memcpy(sdscData[x], console[x], 80 * 2);
    }
    sdscUpdated = true;
}

void UIDebugWindow::setSDSC() {
    QString c[25];

    if(sdscUpdated) {
        for(int x = 0; x < 25; ++x) {
            for(int y = 0; y < 80; ++y) {
                c[x].append(sdscData[x][(y << 1) + 1]);
            }
        }

        sdscConsole->clear();
        for(int x = 0; x < 25; ++x) {
            sdscConsole->append(c[x]);
        }
    }
    sdscUpdated = false;
}


void UIDebugWindow::resetSDSC() {
    int i, j;

    for(i = 0; i < 25; ++i) {
        for(j = 0; j < 80; ++j) {
            sdscData[i][j << 1] = 0;
            sdscData[i][(j << 1) + 1] = 0;
        }
    }

    sdscUpdated = true;
}

void UIDebugWindow::closeEvent(QCloseEvent* event) {
    QSettings settings;
    settings.beginGroup("DebugWindow");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
    QWidget::closeEvent(event);
}

void UIDebugWindow::setCpuInfo() {
    char tmp[16];

    sprintf(tmp, "%04x", cpuz80->pc.w);
    pc_label->setText(tmp);

    sprintf(tmp, "%02x", cpuz80->af.b.h);
    a_label->setText(tmp);

    // Two commented out are undocumented flags, so I'm leaving them out
#define GF(x) CRABZ80_GET_FLAG(cpuz80, CRABZ80_##x)
    sprintf(tmp, "%d%d%d%d%d%d%d%d", GF(SF), GF(ZF), GF(YF), GF(HF), GF(XF), GF(PF), GF(NF), GF(CF));
#undef GF

    flags_label->setText(tmp);

    sprintf(tmp, "%02x", cpuz80->ir.b.h);
    i_label->setText(tmp);

    sprintf(tmp, "%02x%02x", cpuz80->iff1, cpuz80->iff2);
    iff_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->bc.w);
    bc_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->de.w);
    de_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->hl.w);
    hl_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->afp.w);
    af_prime_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->bcp.w);
    bc_prime_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->dep.w);
    de_prime_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->hlp.w);
    hl_prime_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->ix.w);
    ix_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->iy.w);
    iy_label->setText(tmp);

    sprintf(tmp, "%04x", cpuz80->sp.w);
    sp_label->setText(tmp);

    setVdpInfo();
    setSndInfo();
    disassemble();
}

void UIDebugWindow::setVdpInfo()
{
    char tmp[32];

    // simple macro to save some typing for setting the Qt label
#define VDPREG(x) \
    sprintf(tmp, "%02x", smsvdp.regs[x]); \
    vdpReg##x->setText(tmp);
    VDPREG(0)
    VDPREG(1)
    VDPREG(2)
    VDPREG(3)
    VDPREG(4)
    VDPREG(5)
    VDPREG(6)
    VDPREG(7)
    VDPREG(8)
    VDPREG(9)
    VDPREG(10)
    VDPREG(11)
    VDPREG(12)
    VDPREG(13)
    VDPREG(14)
    VDPREG(15)
#undef VDPREG

    sprintf(tmp, "%01d", smsvdp.code);
    vdpCode->setText(tmp);

    sprintf(tmp, "%04x", smsvdp.addr);
    vdpAddress->setText(tmp);

    sprintf(tmp, "%02x", smsvdp.status);
    vdpStatus->setText(tmp);

    sprintf(tmp, "%02x", smsvdp.read_buf);
    vdpReadBuffer->setText(tmp);

    sprintf(tmp, "%02x", smsvdp.v_scroll);
    vdpvScrollLatch->setText(tmp);

    vdpFlag1->setChecked(smsvdp.flags & 0x01);
    vdpvScrollFlag->setChecked(smsvdp.flags & 0x02);
    vdpLineIntFlag->setChecked(smsvdp.flags & 0x04);
}

void UIDebugWindow::setVdpTiles() {
    int i, j, k, x = 0, y = 0, tmp;
#ifdef CRABEMU_32BIT_COLOR
    uint32 buf[256 * 256] = { 0 };
#else
    uint16 buf[256 * 256] = { 0 };
#endif

    for(i = 0; i < 512; ++i) {
        if(smsvdp.pattern[i].dirty)
            sms_vdp_update_cache(i);

        for(j = 0; j < 8; ++j) {
            for(k = 0; k < 8; ++k) {
                /* Draw with background palette first. */
                tmp = smsvdp.pattern[i].texture[0][j * 8 + k];
                buf[(y + j) * 256 + x + k] = smsvdp.pal[tmp];

                /* Then with the sprite palette. */
                tmp |= 0x10;
                buf[(y + j + 128) * 256 + x + k] = smsvdp.pal[tmp];

            }
        }

        x += 8;

        if(x == 256) {
            x = 0;
            y += 8;
        }
    }

#ifdef CRABEMU_32BIT_COLOR
    QImage img(256, 256, QImage::Format_RGB32);
#else
    QImage img(256, 256, QImage::Format_RGB555);
#endif
    uchar *bits = img.bits();

#ifdef CRABEMU_32BIT_COLOR
    memcpy(bits, buf, 256 * 256 * 4);
#else
    memcpy(bits, buf, 256 * 256 * 2);
#endif

    QLabel *q = dynamic_cast<QLabel *>(vdpTilesDock->widget());
    q->setPixmap(QPixmap::fromImage(img));
}

void UIDebugWindow::setSndInfo()
{
    char tmp[32];

    // simple macro to save some typing for setting the Qt label
#define SNDVOL(x) \
    sprintf(tmp, "%01x", psg.volume[x]); \
    sndVol##x->setText(tmp);
    SNDVOL(0)
    SNDVOL(1)
    SNDVOL(2)
    SNDVOL(3)
#undef SNDVOL

#define SNDTONE(x) \
    sprintf(tmp, "%03x", psg.tone[x]); \
    sndTone##x->setText(tmp);
    SNDTONE(0)
    SNDTONE(1)
    SNDTONE(2)
#undef SNDTONE

	sprintf(tmp, "%03x", psg.noise);
	sndNoise->setText(tmp);
}

void UIDebugWindow::disassemble() {
    char dis[64];

    CrabZ80_disassemble(dis, cpuz80, cpuz80->pc.w);
    disassemblyEdit->clear();
    disassemblyEdit->append(dis);
}


// When we press pause, we call our signal which is connected to the main
// window's pause slot to do the work
void UIDebugWindow::on_actionPause_triggered() {
    if(!sms_initialized && !coleco_initialized) {
        actionPause->setChecked(false);
        return;
    }

    paused = !paused;

    if(paused) {
        setCpuInfo();
        update();
    }

    emit togglePaused();
}

void UIDebugWindow::on_actionCPUStep_triggered() {
    // don't allow if we're not paused
    if((!sms_initialized && !coleco_initialized)|| !paused)
        return;
    sms_z80_run(1);
    setCpuInfo();
    update();
}

void UIDebugWindow::on_actionFrameStep_triggered() {
    if((!sms_initialized && !coleco_initialized) || !paused)
        return;

    sms_frame(1, 0);
    render_scene();
    parent->emuwidget->update();
    setCpuInfo();
    update();
}

void UIDebugWindow::on_actionRunToCursor_triggered() {
    fprintf(stderr, "Run to Cursor\n");
}

void UIDebugWindow::on_actionReloadRom_triggered() {
    emit reloadRom();
}

void UIDebugWindow::on_actionReset_triggered() {
    emit reset();
    if(paused)
        on_actionPause_triggered();
}

// When the main window presses pause we just toggle the button
void UIDebugWindow::on_main_pause(bool pause) {
    paused = pause;
    actionPause->setChecked(pause);

    if(paused) {
        setCpuInfo();
        update();
    }
}

void UIDebugWindow::updatePalette() {
    QPixmap p(16, 16);

#ifdef CRABEMU_32BIT_COLOR
#define PAL2QColor(x) QColor(x)
#else
#define PAL2QColor(x) QColor( \
    ((float)(x >> 10 & 0x1F) / 0x1F) * 255, \
    ((float)(x >> 5 & 0x1F) / 0x1F) * 255, \
    ((float)(x >> 0 & 0x1F) / 0x1F) * 255 \
)
#endif

#define SETPAL(x) \
    p.fill(PAL2QColor(smsvdp.pal[x]));  \
    pal##x->setPixmap(p);

    SETPAL(0);
    SETPAL(1);
    SETPAL(2);
    SETPAL(3);
    SETPAL(4);
    SETPAL(5);
    SETPAL(6);
    SETPAL(7);
    SETPAL(8);
    SETPAL(9);
    SETPAL(10);
    SETPAL(11);
    SETPAL(12);
    SETPAL(13);
    SETPAL(14);
    SETPAL(15);
    SETPAL(16);
    SETPAL(17);
    SETPAL(18);
    SETPAL(19);
    SETPAL(20);
    SETPAL(21);
    SETPAL(22);
    SETPAL(23);
    SETPAL(24);
    SETPAL(25);
    SETPAL(26);
    SETPAL(27);
    SETPAL(28);
    SETPAL(29);
    SETPAL(30);
    SETPAL(31);

#undef SETPAL
#undef PAL2QColor
}

void UIDebugWindow::update() {
    if(sms_initialized || coleco_initialized) {
        updatePalette();
        setSDSC();
        setSndInfo();
        setVdpInfo();
        setCpuInfo();
        setVdpTiles();
    }
}
