/*
    This file is part of CrabEmu.

    Copyright (C) 2011 Harley Laue

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <malloc.h>
#include <lua.h>
#include <lualib.h>
#include <lauxlib.h>
#include <SDL.h>

#include "config.h"
#include "sms.h"

lua_State *state = NULL;

/* The structure which defines the default values */
struct {
    struct {
        int value;
        char *comment;
        char *cfgName;
    } config_int[CONFIG_INT_SIZE];

    struct {
        char *value;
        char *comment;
        char *cfgName;
        int empty;
    } config_str[CONFIG_SIZE];
} defaultCfg, userCfg;

/* Create a new configuration file */
int config_new(const char *filename, lua_State *s) {
    FILE *file = fopen(filename, "w");
    int x;

    if(!file) {
        return -1;
    }

    fprintf(file, "--[[\n"
        "Changes to this file may be lost or overwritten.\n"
        "You can put you own changes in\n"
        "%s.local\n"
        "to override settings here.\n"
        "This file accepts valid Lua syntax.\n"
        "Only the following standard libraries are open:\n"
        "* Base\n"
        "* Math\n"
        "* String\n"
        "* Table\n"
        "]]--\n",
        filename);

    for(x = 0; x < CONFIG_INT_SIZE; ++x) {
        if(defaultCfg.config_int[x].comment)
            fprintf(file, "%s\n", defaultCfg.config_int[x].comment);
        fprintf(file, "%s = %d\n", defaultCfg.config_int[x].cfgName, defaultCfg.config_int[x].value);
    }

    for(x = CONFIG_INT_SIZE + 1; x < CONFIG_SIZE; ++x) {
        if(defaultCfg.config_str[x].comment)
            fprintf(file, "%s\n", defaultCfg.config_str[x].comment);
        fprintf(file, "%s = %s\n", defaultCfg.config_str[x].cfgName, defaultCfg.config_str[x].value);
    }

    fclose(file);
    return 0;
}

/* Setup the default values */
void config_defaults() {
    int ym2413_enabled = 1;

#if defined(DIDJ)
    ym2413_enabled = 0;
#endif

    /* default int values */
    #define cfg_int(e, c, v, com) \
    defaultCfg.config_int[e].cfgName = c; \
    defaultCfg.config_int[e].value = v; \
    if(com) defaultCfg.config_int[e].comment = com;

    cfg_int(WIDTH, "width", 256, NULL)
    cfg_int(HEIGHT, "height", 192, NULL)
    cfg_int(ACTUAL_SIZE, "actual_size", 1, "-- 1 for forcing the emulator to be the actual resolution\n-- 0 to allow stretching/scaling");
    cfg_int(PSG76489_ENABLED, "psg76489_enabled", 1, "-- 1 is enabled\n-- 0 is disabled");
    cfg_int(YM2413_ENABLED, "ym2413_enabled", ym2413_enabled, "-- 1 is enabled\n-- 0 is disabled");
    cfg_int(DELAY, "delay", 31, "-- 31 seems to be a good value for NTSC\n-- 37 seems to be a good value for PAL");
    cfg_int(REGION, "region", SMS_REGION_DOMESTIC, "-- 1 for domestic\n-- 2 for export\n-- all other values are invalid");
    cfg_int(VIDEOMODE, "video_mode", SMS_VIDEO_NTSC, "-- 16 for NTSC (60 FPS)\n-- 32 for PAL (50 FPS)\n-- all other values are invalid");

    /* Controller button setup */
    cfg_int(PAD1_UP, "pad1_up", SDLK_UP, "-- For input you can either use the numbers defined in SDL_keysym.h\n-- or it should work in most cases to use 'a' for the A key, 'd' for the D key, etc.");
    cfg_int(PAD1_DOWN, "pad1_down", SDLK_DOWN, NULL);
    cfg_int(PAD1_LEFT, "pad1_left", SDLK_LEFT, NULL);
    cfg_int(PAD1_RIGHT, "pad1_right", SDLK_RIGHT, NULL);
#ifdef DIDJ
    cfg_int(PAD1_1, "pad1_1", SDLK_a, NULL);
    cfg_int(PAD1_2, "pad1_2", SDLK_b, NULL);
    cfg_int(PAD2_UP, "pad2_up", SDLK_UNKNOWN, NULL);
    cfg_int(PAD2_DOWN, "pad2_down", SDLK_UNKNOWN, NULL);
    cfg_int(PAD2_LEFT, "pad2_left", SDLK_UNKNOWN, NULL);
    cfg_int(PAD2_RIGHT, "pad2_right", SDLK_UNKNOWN, NULL);
    cfg_int(PAD2_1, "pad2_1", SDLK_UNKNOWN, NULL);
    cfg_int(PAD2_2, "pad2_2", SDLK_UNKNOWN, NULL);
    cfg_int(SMS_PAUSE, "sms_pause", SDLK_p, NULL);
    cfg_int(GG_PAUSE, "gg_pause", SDLK_p, NULL);
    cfg_int(QUIT_BUTTON, "quit_button", SDLK_m, NULL);
    cfg_int(MUTE_BUTTON, "mute_button", SDLK_h, NULL);

    cfg_int(COLECO1_UP, "coleco1_up", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_DOWN, "coleco1_down", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_LEFT, "coleco1_left", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_RIGHT, "coleco1_right", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_LACTION, "coleco1_l_action", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_RACTION, "coleco1_r_action", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_1, "coleco1_1", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_2, "coleco1_2", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_3, "coleco1_3", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_4, "coleco1_4", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_5, "coleco1_5", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_6, "coleco1_6", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_7, "coleco1_7", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_8, "coleco1_8", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_9, "coleco1_9", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_STAR, "coleco1_star", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_0, "coleco1_0", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO1_POUND, "coleco1_pound", SDLK_UNKNOWN, NULL);

    cfg_int(COLECO2_UP, "coleco2_up", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_DOWN, "coleco2_down", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_LEFT, "coleco2_left", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_RIGHT, "coleco2_right", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_LACTION, "coleco2_l_action", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_RACTION, "coleco2_r_action", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_1, "coleco2_1", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_2, "coleco2_2", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_3, "coleco2_3", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_4, "coleco2_4", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_5, "coleco2_5", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_6, "coleco2_6", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_7, "coleco2_7", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_8, "coleco2_8", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_9, "coleco2_9", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_STAR, "coleco2_star", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_0, "coleco2_0", SDLK_UNKNOWN, NULL);
    cfg_int(COLECO2_POUND, "coleco2_pound", SDLK_UNKNOWN, NULL);
#else
    cfg_int(PAD1_1, "pad1_1", SDLK_a, NULL);
    cfg_int(PAD1_2, "pad1_2", SDLK_s, NULL);
    cfg_int(PAD2_UP, "pad2_up", SDLK_i, NULL);
    cfg_int(PAD2_DOWN, "pad2_down", SDLK_k, NULL);
    cfg_int(PAD2_LEFT, "pad2_left", SDLK_j, NULL);
    cfg_int(PAD2_RIGHT, "pad2_right", SDLK_l, NULL);
    cfg_int(PAD2_1, "pad2_1", SDLK_y, NULL);
    cfg_int(PAD2_2, "pad2_2", SDLK_u, NULL);
    cfg_int(SMS_PAUSE, "sms_pause", SDLK_SPACE, NULL);
    cfg_int(GG_PAUSE, "gg_pause", SDLK_SPACE, NULL);
    cfg_int(QUIT_BUTTON, "quit_button", SDLK_ESCAPE, NULL);
    cfg_int(MUTE_BUTTON, "mute_button", SDLK_m, NULL);

    cfg_int(COLECO1_UP, "coleco1_up", SDLK_w, NULL);
    cfg_int(COLECO1_DOWN, "coleco1_down", SDLK_s, NULL);
    cfg_int(COLECO1_LEFT, "coleco1_left", SDLK_a, NULL);
    cfg_int(COLECO1_RIGHT, "coleco1_right", SDLK_d, NULL);
    cfg_int(COLECO1_LACTION, "coleco1_l_action", SDLK_q, NULL);
    cfg_int(COLECO1_RACTION, "coleco1_r_action", SDLK_e, NULL);
    cfg_int(COLECO1_1, "coleco1_1", SDLK_4, NULL);
    cfg_int(COLECO1_2, "coleco1_2", SDLK_5, NULL);
    cfg_int(COLECO1_3, "coleco1_3", SDLK_6, NULL);
    cfg_int(COLECO1_4, "coleco1_4", SDLK_r, NULL);
    cfg_int(COLECO1_5, "coleco1_5", SDLK_t, NULL);
    cfg_int(COLECO1_6, "coleco1_6", SDLK_y, NULL);
    cfg_int(COLECO1_7, "coleco1_7", SDLK_g, NULL);
    cfg_int(COLECO1_8, "coleco1_8", SDLK_h, NULL);
    cfg_int(COLECO1_9, "coleco1_9", SDLK_j, NULL);
    cfg_int(COLECO1_STAR, "coleco1_star", SDLK_b, NULL);
    cfg_int(COLECO1_0, "coleco1_0", SDLK_n, NULL);
    cfg_int(COLECO1_POUND, "coleco1_pound", SDLK_m, NULL);

    cfg_int(COLECO2_UP, "coleco2_up", SDLK_UP, NULL);
    cfg_int(COLECO2_DOWN, "coleco2_down", SDLK_DOWN, NULL);
    cfg_int(COLECO2_LEFT, "coleco2_left", SDLK_LEFT, NULL);
    cfg_int(COLECO2_RIGHT, "coleco2_right", SDLK_RIGHT, NULL);
    cfg_int(COLECO2_LACTION, "coleco2_l_action", SDLK_MINUS, NULL);
    cfg_int(COLECO2_RACTION, "coleco2_r_action", SDLK_EQUALS, NULL);
    cfg_int(COLECO2_1, "coleco2_1", SDLK_8, NULL);
    cfg_int(COLECO2_2, "coleco2_2", SDLK_9, NULL);
    cfg_int(COLECO2_3, "coleco2_3", SDLK_0, NULL);
    cfg_int(COLECO2_4, "coleco2_4", SDLK_i, NULL);
    cfg_int(COLECO2_5, "coleco2_5", SDLK_o, NULL);
    cfg_int(COLECO2_6, "coleco2_6", SDLK_p, NULL);
    cfg_int(COLECO2_7, "coleco2_7", SDLK_k, NULL);
    cfg_int(COLECO2_8, "coleco2_8", SDLK_l, NULL);
    cfg_int(COLECO2_9, "coleco2_9", SDLK_SEMICOLON, NULL);
    cfg_int(COLECO2_STAR, "coleco2_star", SDLK_COMMA, NULL);
    cfg_int(COLECO2_0, "coleco2_0", SDLK_PERIOD, NULL);
    cfg_int(COLECO2_POUND, "coleco2_pound", SDLK_SLASH, NULL);
#endif

    #undef cfg_int

    /**
     * default string values
     * @param e the enum value it has from Config
     * @param c the name it's given in Lua
     * @param v the default value
     * @param com comment to save in the config
     * @param em if it's alright for the value to be an empty string
     */
    #define cfg_str(e, c, v, com, em) \
    defaultCfg.config_str[e].cfgName = c; \
    defaultCfg.config_str[e].value = v; \
    if(com) defaultCfg.config_str[e].comment = com; \
    if(em) defaultCfg.config_str[e].empty = em;

#ifdef DIDJ
    #define ROM_PATH "'/root/roms/'"
    #define SRAM_PATH "'/root/saves/'"
    #define STATE_PATH "'/root/states/'"
    #define FONT_PATH "'/root/.crabemu/FreeSerif.ttf'"
#else
    #define ROM_PATH "''"
    #define SRAM_PATH "''"
    #define STATE_PATH "''"
    #define FONT_PATH "'/usr/share/fonts/TTF/FreeSerif.ttf'"
#endif

    cfg_str(RENDERER, "renderer", "'sdl'", "-- Only 'sdl' is valid.", 1);
    cfg_str(SCALER, "scaler", "''", "-- Leave blank or undefined.", 1);
    cfg_str(SMS_BIOS_DIR, "sms_bios_dir", "''", "-- Leave blank or undefined to not load bios, otherwise the full path to the SMS Bios.", 1);
    cfg_str(GG_BIOS_DIR, "gg_bios_dir", "''", "-- Leave blank or undefined to not load bios, otherwise the full path to the GG Bios.", 1);
    cfg_str(COLECO_BIOS_DIR, "coleco_bios_dir", "''", "-- Leave blank or undefined to not load bios, otherwise the full path to the ColecoVision Bios (required for this system).", 1);
    cfg_str(SAVE_STATE_DIR, "save_state_dir", STATE_PATH, "-- Leave blank or undefined to save with the rom, otherwise the full directory path for the save states.", 1);
    cfg_str(SRAM_DIR, "sram_dir", SRAM_PATH, "-- Leave blank or undefined to save with the rom, otherwise the full directory path for the SRAM saves.", 1);
    cfg_str(ROM_DIR, "rom_dir", ROM_PATH, "-- the default directory to search for roms for the menu or the default rom", 1);
    cfg_str(DEFAULT_ROM, "default_rom", "rom_dir .. ''", "-- Path to the default rom file. Leave blank or undefined if desired.", 1);
    cfg_str(FONT_FILE, "font_file", FONT_PATH, "-- the font to load. The default is FreeSerif", 0);

    #undef cfg_str
}

/* fill the defaultCfg with the values from the loaded config */
void fill_config() {
    int x;
    assert(state != NULL);

    for(x = 0; x < CONFIG_INT_SIZE; ++x) {
        userCfg.config_int[x].value = config_get_int(x);
    }

    for(x = CONFIG_INT_SIZE + 1; x < CONFIG_SIZE; ++x) {
        userCfg.config_str[x].value = (char *)config_get_str(x);
    }
}

char *config_default_filename() {
    /* This is system specific, should probably ifdef this */
    const char *path = getenv("HOME");
    char *filename;
    filename = malloc(strlen(path) + 14);
    strncpy(filename, path, strlen(path)+1);
    strncat(filename, "/.crabemu.cfg", 13);

    return filename;
}

void config_open_local(char *filename) {
    char *error = NULL;
    int allocfilename = 0;
    if(state == NULL) {
        return;
    }
    if(filename == NULL) {
        filename = config_default_filename();
        allocfilename = 1;
    }

    size_t len = strlen(filename);
    char filenamelocal[len + 8];
    strncpy(filenamelocal, filename, strlen(filename)+1);
    strncat(filenamelocal, ".local", 6);

    switch(luaL_loadfile(state, filenamelocal)) {
        /* if we have an error, fall through */
        case LUA_ERRSYNTAX:
            error = "Configuration syntax error";
            goto cleanup_failure;
        case LUA_ERRMEM:
            /* only set the error if not already set */
            error = error ? error : "Error allocating memory for lua";
            goto cleanup_failure;
        case LUA_ERRFILE:
            // Not an error, the file just doesn't exist because it's optional
            return;
        default:
            break;
    }

    /* Now that the configuration is found and loaded we have to run it */
    switch(lua_pcall(state, 0, LUA_MULTRET, 0)) {
        /* Same as above */
        case LUA_ERRRUN:
            error = "Runtime error";
            goto cleanup_failure;
        case LUA_ERRMEM:
            error = error ? error : "Memory allocation error";
            goto cleanup_failure;
        case LUA_ERRERR:
            error = error ? error : "Error with error handler";

            fprintf(stderr, "%s: %s\n", error, lua_tostring(state, -1));
            goto cleanup_failure;
        default:
            break;
    }

    fill_config();

    if(allocfilename) {
        free(filename);
        filename = NULL;
    }

    return;
cleanup_failure:

    fprintf(stderr, "Error: %s\n", error);

    if(allocfilename) {
        free(filename);
        filename = NULL;
    }
}

int config_open(char* filename) {
    lua_State *s = luaL_newstate();
    char *error = NULL;
    FILE *file;
    int newfile = 0, rv = 0;

    // Load the base, math, string, and table libraries
    luaopen_base(s);
    luaopen_math(s);
    luaopen_string(s);
    luaopen_table(s);

    config_defaults();

    /* No passed in filename, use the default one */
    if(filename == NULL) {
        newfile = 1;
        filename = config_default_filename();
        printf("%s\n", filename);
    }

    file = fopen(filename, "r");

    /* configuration file does not exist, try creating one */
    if(!file) {
        printf("No configuration: creating a new one\n");

        if(config_new(filename, s) < 0) {
            fprintf(stderr, "Error creating new config: %s\n", filename);
            rv = -1;
            goto cleanup_failure;
        }
    }
    else {
        /* File exists and is readable, We're done with it. */
        fclose(file);
    }

    /* We have a valid filename by this point try loading it */
    switch(luaL_loadfile(s, filename)) {
        /* if we have an error, fall through */
        case LUA_ERRSYNTAX:
            error = "Configuration syntax error";
        case LUA_ERRMEM:
            /* only set the error if not already set */
            error = error ? error : "Error allocating memory for lua";
        case LUA_ERRFILE:
            /* This one should never actually happen */
            error = error ? error : "File error";

            fprintf(stderr, "%s: %s\n", error, lua_tostring(s, -1));
            rv = -2;
            goto cleanup_failure;
        default:
            break;
    }

    /* close any lua_State that may already be open */
    config_close();

    /* Now that the configuration is found and loaded we have to run it */
    switch(lua_pcall(s, 0, LUA_MULTRET, 0)) {
        /* Same as above */
        case LUA_ERRRUN:
            error = "Runtime error";
        case LUA_ERRMEM:
            error = error ? error : "Memory allocation error";
        case LUA_ERRERR:
            error = error ? error : "Error with error handler";

            fprintf(stderr, "%s: %s\n", error, lua_tostring(s, -1));
            rv = -3;
            goto cleanup_failure;
        default:
            break;
    }

    /* Ok, we've done well. Set our state to the locally created state */
    state = s;

    fill_config();

    if(newfile)
        free(filename);

    return rv;

cleanup_failure:
    lua_close(s);

    if(newfile)
        free(filename);

    return rv;
}

void config_close()
{
    if(state != NULL) {
        lua_close(state);
        state = NULL;
    }
}

const char *config_get_str(enum Config c) {
    assert(state != NULL && c < CONFIG_SIZE);
    char *value = NULL;

    lua_getglobal(state, defaultCfg.config_str[c].cfgName);
    /* First try the value as a string */
    if(lua_isstring(state, -1)) {
        value = (char *)lua_tostring(state, -1);
    }
    /* Ok, try it as undefined, nil, or blank if that's allowed for the type */
    else if(defaultCfg.config_str[c].empty && lua_isnoneornil(state, -1) && strncmp(defaultCfg.config_str[c].value, "''", 2) == 0) {
        value = "";
    }
    else {
        /* we failed to load the value, use the default */
        config_defaults();
        fprintf(stderr, "Configuration value '%s' unset. Using default value of %s\n", defaultCfg.config_str[c].cfgName, defaultCfg.config_str[c].value);
        value = defaultCfg.config_str[c].value;
    }

    lua_pop(state, 1);
    return value;
}

int config_get_int(enum Config c) {
    assert(state != NULL && c < CONFIG_SIZE);
    int value = 0;

    lua_getglobal(state, defaultCfg.config_int[c].cfgName);
    /* Try to get the value of the number */
    if(lua_isnumber(state, -1)) {
        value = lua_tonumber(state, -1);
    }
    // The "number" may actually be a char, so we'll try that
    else if(lua_isstring(state, -1)) {
        const char *str = lua_tostring(state, -1);
        value = str[0];
    }
    else {
        /* we failed to load the value, use the default */
        config_defaults();
        fprintf(stderr, "Configuration value '%s' unset. Using default value of %d\n", defaultCfg.config_int[c].cfgName, defaultCfg.config_int[c].value);
        value = defaultCfg.config_int[c].value;
    }

    lua_pop(state, 1);
    return value;
}
