/*
    This file is part of CrabEmu.

    Copyright (C) 2011,2012 Harley Laue

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include "SDL.h"
#include "SDL_endian.h"
#ifdef ENABLE_OPENGL
#include "SDL_opengl.h"
#endif

#include "CrabEmu.h"
#include "sms.h"
#include "colecovision.h"
#include "smsvdp.h"
#include "smsz80.h"
#include "sn76489.h"
#include "smsmem.h"
#include "smsmem-gg.h"
#include "colecomem.h"
#include "rom.h"
#include "sound.h"
#include "cheats.h"
#include "config.h"

extern int sms_console;
extern int sms_psg_enabled;
extern int sms_ym2413_enabled;
extern int sms_cheats_enabled;
extern int sms_initialized;
extern int coleco_initialized;
static const int PAL_FPS = 50;
static const int NTSC_FPS = 60;

int region = SMS_REGION_EXPORT;
int video_mode = SMS_VIDEO_NTSC;
SDL_Surface *SDL_Screen;
char *title = "CrabEmu";
char *loaded_filename = NULL;
Uint32 frames_rendered;
int delay = 30;
int target_fps;

void (*do_render_scene)(void) = NULL;
int runMenu(Uint32 flags);

#ifdef CRABEMU_32BIT_COLOR
    #define CRAB_BPP 32
    #define CRAB_COLOR Uint32
    #if defined(DIDJ)
        #define CRAB_SWAP(x) SDL_Swap32(x) >> 8
    #else
        #define CRAB_SWAP(x) x
    #endif
    #define CRAB_FB_SIZE 256*256*3
#else
    #if defined(DIDJ)
        /* Didj requires the BPP to be 16 */
        #define CRAB_BPP 16
        /* Doesn't really actually do a swap, it converts from RGB555 to RGB565 */
        #define CRAB_SWAP(x) (x&0xff00)<<1|(x&0x00ff)
    #else
        #define CRAB_BPP 15
        #define CRAB_SWAP(x) SDL_Swap16(x);
    #endif /* DIDJ */
    #define CRAB_COLOR Uint16
    #define CRAB_FB_SIZE 256*256*2
#endif

Uint32 emu_delay()
{
    static Uint32 nextTime = 0;
    Uint32 now = SDL_GetTicks();

    if(nextTime <= now) {
        nextTime = now + delay;
        return 0;
    }

    return(nextTime - now);
}

int emu_frame(int run) {
    frames_rendered++;
    if(sms_console == CONSOLE_COLECOVISION)
        return coleco_frame(run, 0);
    else
        return sms_frame(run, 0);
}

inline void render_didj() {
    int x;
    /* Hack to make it display correctly on the Didj */
    CRAB_COLOR *p = (CRAB_COLOR *)SDL_Screen->pixels;
    pixel_t *f = smsvdp.framebuffer;
    for(x = 0; x < SDL_Screen->h * SDL_Screen->w; ++x, ++p, ++f) {
        CRAB_COLOR t = CRAB_SWAP(*f);
        if(*p != t)
            *p = t;
    }
    /* Possible alternative implementation */
#if 0
    SDL_Rect r;

    if(sms_console == CONSOLE_GG) {
        r.x = 48;
        r.y = 24;
        r.w = 160;
        r.h = 144;
    }
    else {
        r.x = r.y = 0;
        r.w = 256;
        r.h = smsvdp.lines;
    }

    if(SDL_MUSTLOCK(SDL_Screen)) {
        SDL_UnlockSurface(SDL_Screen);
    }
    SDL_Surface *gg = SDL_CreateRGBSurfaceFrom(smsvdp.framebuffer, 256, smsvdp.lines, CRAB_BPP, 256*CRAB_BPP/8, 0x7C00, 0x3E0, 0x1F, 0);
    SDL_BlitSurface(gg, &r, SDL_Screen, NULL);
    SDL_FreeSurface(gg);
#endif
}

void sdl_render_scene(void) {
    SDL_Rect r;

    if(sms_console == CONSOLE_GG) {
        r.x = 48;
        r.y = 24;
        r.w = 160;
        r.h = 144;
    }
    else {
        r.x = r.y = 0;
        r.w = 256;
        r.h = smsvdp.lines;
    }

    if(SDL_MUSTLOCK(SDL_Screen)) {
        SDL_LockSurface(SDL_Screen);
    }

#if defined(DIDJ)
    render_didj();
#else
    if(sms_console == CONSOLE_GG) {
        if(SDL_MUSTLOCK(SDL_Screen)) {
            SDL_UnlockSurface(SDL_Screen);
        }
        SDL_Surface *gg = SDL_CreateRGBSurfaceFrom(smsvdp.framebuffer, 256, smsvdp.lines, CRAB_BPP, 256*CRAB_BPP/8, 0, 0, 0, 0);
        SDL_BlitSurface(gg, &r, SDL_Screen, NULL);
        SDL_FreeSurface(gg);
    }
    else {
        memcpy(SDL_Screen->pixels, smsvdp.framebuffer, CRAB_FB_SIZE);
    }
#endif

    if(SDL_MUSTLOCK(SDL_Screen)) {
        SDL_UnlockSurface(SDL_Screen);
    }

    SDL_Flip(SDL_Screen);
//     SDL_UpdateRect(SDL_Screen, 0, 0, SDL_Screen->w, SDL_Screen->h);
}

void emu_loop(void) {
    SDL_Event event;
    int run = 0, button = -1, x;
    /* we're a bit lazy with the size of system to make it easier to reference values */
    int player_input[CONFIG_INT_SIZE];
    SDLKey k;
    int frames = 0;
    time_t thetime = time(NULL);

    delay = config_get_int(DELAY);

    /* load the input values that map to emulated input */
    for(x = PAD1_UP; x < CONFIG_INT_SIZE; ++x) {
        player_input[x] = config_get_int(x);
    }

    /* disable key repeat */
    if(SDL_EnableKeyRepeat(0, SDL_DEFAULT_REPEAT_INTERVAL) == -1)
        return;

    /* just disable the cursor for now */
    SDL_ShowCursor(SDL_DISABLE);

    for(;;) {
        run = emu_frame(run);

        do_render_scene();

        while(SDL_PollEvent(&event)) {
            /* reset button */
            button = -1;
            switch(event.type) {
                case SDL_KEYUP:
                    k = event.key.keysym.sym;

                    // Emulator controlling buttons
                    if(k == player_input[QUIT_BUTTON])
                        return;
                    if(k == SDLK_BACKQUOTE) {
                        if(sms_console == CONSOLE_COLECOVISION)
                            coleco_soft_reset();
                        else
                            sms_soft_reset();
                    }
                    if(k == player_input[MUTE_BUTTON]) {
                        sms_psg_enabled = !sms_psg_enabled;
                        sms_ym2413_enabled = !sms_ym2413_enabled;
                    }

                    if(sms_console != CONSOLE_COLECOVISION) {
                        // Player 1 buttons
                        if(k == player_input[PAD1_UP])
                            button = SMS_UP;
                        if(k == player_input[PAD1_DOWN])
                            button = SMS_DOWN;
                        if(k == player_input[PAD1_LEFT])
                            button = SMS_LEFT;
                        if(k == player_input[PAD1_RIGHT])
                            button = SMS_RIGHT;
                        if(k == player_input[PAD1_1])
                            button = SMS_BUTTON_1;
                        if(k == player_input[PAD1_2])
                            button = SMS_BUTTON_2;
                        if(k == player_input[GG_PAUSE])
                            button = GG_START;

                        if(button != -1) {
                            sms_button_released(1, button);
                        }

                        // Reset button for player two
                        button = -1;
                        // Player 2 buttons
                        if(k == player_input[PAD2_UP])
                            button = SMS_UP;
                        if(k == player_input[PAD2_DOWN])
                            button = SMS_DOWN;
                        if(k == player_input[PAD2_LEFT])
                            button = SMS_LEFT;
                        if(k == player_input[PAD2_RIGHT])
                            button = SMS_RIGHT;
                        if(k == player_input[PAD2_1])
                            button = SMS_BUTTON_1;
                        if(k == player_input[PAD2_2])
                            button = SMS_BUTTON_2;

                        if(button != -1) {
                            sms_button_released(2, button);
                        }
                    }
                    else {
                        if(k == player_input[COLECO1_UP]) {
                            button = COLECOVISION_UP;
                        }
                        if(k == player_input[COLECO1_DOWN]) {
                            button = COLECOVISION_DOWN;
                        }
                        if(k == player_input[COLECO1_LEFT]) {
                            button = COLECOVISION_LEFT;
                        }
                        if(k == player_input[COLECO1_RIGHT]) {
                            button = COLECOVISION_RIGHT;
                        }
                        if(k == player_input[COLECO1_LACTION]) {
                            button = COLECOVISION_L_ACTION;
                        }
                        if(k == player_input[COLECO1_RACTION]) {
                            button = COLECOVISION_R_ACTION;
                        }
                        if(k == player_input[COLECO1_1]) {
                            button = COLECOVISION_1;
                        }
                        if(k == player_input[COLECO1_2]) {
                            button = COLECOVISION_2;
                        }
                        if(k == player_input[COLECO1_3]) {
                            button = COLECOVISION_3;
                        }
                        if(k == player_input[COLECO1_4]) {
                            button = COLECOVISION_4;
                        }
                        if(k == player_input[COLECO1_5]) {
                            button = COLECOVISION_5;
                        }
                        if(k == player_input[COLECO1_6]) {
                            button = COLECOVISION_6;
                        }
                        if(k == player_input[COLECO1_7]) {
                            button = COLECOVISION_7;
                        }
                        if(k == player_input[COLECO1_8]) {
                            button = COLECOVISION_8;
                        }
                        if(k == player_input[COLECO1_9]) {
                            button = COLECOVISION_9;
                        }
                        if(k == player_input[COLECO1_STAR]) {
                            button = COLECOVISION_STAR;
                        }
                        if(k == player_input[COLECO1_0]) {
                            button = COLECOVISION_0;
                        }
                        if(k == player_input[COLECO1_POUND]) {
                            button = COLECOVISION_POUND;
                        }

                        if(button != -1) {
                            coleco_button_released(1, button);
                        }

                        button = -1;
                        if(k == player_input[COLECO2_UP]) {
                            button = COLECOVISION_UP;
                        }
                        if(k == player_input[COLECO2_DOWN]) {
                            button = COLECOVISION_DOWN;
                        }
                        if(k == player_input[COLECO2_LEFT]) {
                            button = COLECOVISION_LEFT;
                        }
                        if(k == player_input[COLECO2_RIGHT]) {
                            button = COLECOVISION_RIGHT;
                        }
                        if(k == player_input[COLECO2_LACTION]) {
                            button = COLECOVISION_L_ACTION;
                        }
                        if(k == player_input[COLECO2_RACTION]) {
                            button = COLECOVISION_R_ACTION;
                        }
                        if(k == player_input[COLECO2_1]) {
                            button = COLECOVISION_1;
                        }
                        if(k == player_input[COLECO2_2]) {
                            button = COLECOVISION_2;
                        }
                        if(k == player_input[COLECO2_3]) {
                            button = COLECOVISION_3;
                        }
                        if(k == player_input[COLECO2_4]) {
                            button = COLECOVISION_4;
                        }
                        if(k == player_input[COLECO2_5]) {
                            button = COLECOVISION_5;
                        }
                        if(k == player_input[COLECO2_6]) {
                            button = COLECOVISION_6;
                        }
                        if(k == player_input[COLECO2_7]) {
                            button = COLECOVISION_7;
                        }
                        if(k == player_input[COLECO2_8]) {
                            button = COLECOVISION_8;
                        }
                        if(k == player_input[COLECO2_9]) {
                            button = COLECOVISION_9;
                        }
                        if(k == player_input[COLECO2_STAR]) {
                            button = COLECOVISION_STAR;
                        }
                        if(k == player_input[COLECO2_0]) {
                            button = COLECOVISION_0;
                        }
                        if(k == player_input[COLECO2_POUND]) {
                            button = COLECOVISION_POUND;
                        }

                        if(button != -1) {
                            coleco_button_released(2, button);
                        }
                    }

                    break;
                case SDL_KEYDOWN:
                    k = event.key.keysym.sym;

                    // Emulator Controlling buttons
                    if(k == player_input[QUIT_BUTTON])
                        return;
                    if(k == player_input[SMS_PAUSE]) {
                        if(sms_console == CONSOLE_SMS) {
                            sms_z80_nmi();
                        }
                    }

                    // SMS Player 1 buttons
                    if(sms_console != CONSOLE_COLECOVISION) {
                        if(k == player_input[PAD1_UP])
                            button = SMS_UP;
                        if(k == player_input[PAD1_DOWN])
                            button = SMS_DOWN;
                        if(k == player_input[PAD1_LEFT])
                            button = SMS_LEFT;
                        if(k == player_input[PAD1_RIGHT])
                            button = SMS_RIGHT;
                        if(k == player_input[PAD1_1])
                            button = SMS_BUTTON_1;
                        if(k == player_input[PAD1_2])
                            button = SMS_BUTTON_2;
                        if(k == player_input[GG_PAUSE]) {
                            if(sms_console == CONSOLE_GG) {
                                button = GG_START;
                            }
                        }

                        if(button != -1) {
                            sms_button_pressed(1, button);
                        }

                        // Reset button for player two
                        button = -1;
                        // Player 2 buttons
                        if(k == player_input[PAD2_UP])
                            button = SMS_UP;
                        if(k == player_input[PAD2_DOWN])
                            button = SMS_DOWN;
                        if(k == player_input[PAD2_LEFT])
                            button = SMS_LEFT;
                        if(k == player_input[PAD2_RIGHT])
                            button = SMS_RIGHT;
                        if(k == player_input[PAD2_1])
                            button = SMS_BUTTON_1;
                        if(k == player_input[PAD2_2])
                            button = SMS_BUTTON_2;

                        if(button != -1) {
                            sms_button_pressed(2, button);
                        }
                    }
                    else {
                        if(k == player_input[COLECO1_UP]) {
                            button = COLECOVISION_UP;
                        }
                        if(k == player_input[COLECO1_DOWN]) {
                            button = COLECOVISION_DOWN;
                        }
                        if(k == player_input[COLECO1_LEFT]) {
                            button = COLECOVISION_LEFT;
                        }
                        if(k == player_input[COLECO1_RIGHT]) {
                            button = COLECOVISION_RIGHT;
                        }
                        if(k == player_input[COLECO1_LACTION]) {
                            button = COLECOVISION_L_ACTION;
                        }
                        if(k == player_input[COLECO1_RACTION]) {
                            button = COLECOVISION_R_ACTION;
                        }
                        if(k == player_input[COLECO1_1]) {
                            button = COLECOVISION_1;
                        }
                        if(k == player_input[COLECO1_2]) {
                            button = COLECOVISION_2;
                        }
                        if(k == player_input[COLECO1_3]) {
                            button = COLECOVISION_3;
                        }
                        if(k == player_input[COLECO1_4]) {
                            button = COLECOVISION_4;
                        }
                        if(k == player_input[COLECO1_5]) {
                            button = COLECOVISION_5;
                        }
                        if(k == player_input[COLECO1_6]) {
                            button = COLECOVISION_6;
                        }
                        if(k == player_input[COLECO1_7]) {
                            button = COLECOVISION_7;
                        }
                        if(k == player_input[COLECO1_8]) {
                            button = COLECOVISION_8;
                        }
                        if(k == player_input[COLECO1_9]) {
                            button = COLECOVISION_9;
                        }
                        if(k == player_input[COLECO1_STAR]) {
                            button = COLECOVISION_STAR;
                        }
                        if(k == player_input[COLECO1_0]) {
                            button = COLECOVISION_0;
                        }
                        if(k == player_input[COLECO1_POUND]) {
                            button = COLECOVISION_POUND;
                        }

                        if(button != -1) {
                            coleco_button_pressed(1, button);
                        }

                        button = -1;
                        if(k == player_input[COLECO2_UP]) {
                            button = COLECOVISION_UP;
                        }
                        if(k == player_input[COLECO2_DOWN]) {
                            button = COLECOVISION_DOWN;
                        }
                        if(k == player_input[COLECO2_LEFT]) {
                            button = COLECOVISION_LEFT;
                        }
                        if(k == player_input[COLECO2_RIGHT]) {
                            button = COLECOVISION_RIGHT;
                        }
                        if(k == player_input[COLECO2_LACTION]) {
                            button = COLECOVISION_L_ACTION;
                        }
                        if(k == player_input[COLECO2_RACTION]) {
                            button = COLECOVISION_R_ACTION;
                        }
                        if(k == player_input[COLECO2_1]) {
                            button = COLECOVISION_1;
                        }
                        if(k == player_input[COLECO2_2]) {
                            button = COLECOVISION_2;
                        }
                        if(k == player_input[COLECO2_3]) {
                            button = COLECOVISION_3;
                        }
                        if(k == player_input[COLECO2_4]) {
                            button = COLECOVISION_4;
                        }
                        if(k == player_input[COLECO2_5]) {
                            button = COLECOVISION_5;
                        }
                        if(k == player_input[COLECO2_6]) {
                            button = COLECOVISION_6;
                        }
                        if(k == player_input[COLECO2_7]) {
                            button = COLECOVISION_7;
                        }
                        if(k == player_input[COLECO2_8]) {
                            button = COLECOVISION_8;
                        }
                        if(k == player_input[COLECO2_9]) {
                            button = COLECOVISION_9;
                        }
                        if(k == player_input[COLECO2_STAR]) {
                            button = COLECOVISION_STAR;
                        }
                        if(k == player_input[COLECO2_0]) {
                            button = COLECOVISION_0;
                        }
                        if(k == player_input[COLECO2_POUND]) {
                            button = COLECOVISION_POUND;
                        }

                        if(button != -1) {
                            coleco_button_pressed(2, button);
                        }
                    }
                    break;
                case SDL_QUIT:
                    return;
                    break;
                default:
                    break;
            }
        }

#if defined(DIDJ)
        /* SDL_Delay every frame on the Didj limits it to 50 fps */
        if(frames%10 == 0){
            usleep(emu_delay());
        }
#else
        SDL_Delay(emu_delay());
#endif

        frames++;

        if(thetime + 1 <= time(NULL))   {
            int fps = floor((double)frames/(time(NULL) - thetime));
            if(fps < target_fps && delay > 0) {
                delay--;
            }
            if(fps > target_fps) {
                delay++;
            }
            #ifdef TIMING_TEST
                printf("%d frames in %d seconds: %d FPS delay: %d\n", frames, time(NULL) - thetime,
                    fps, delay);
            #endif
            frames = 0;
            thetime = time(NULL);
        }
    }
}

void gui_set_aspect(float x, float y) {
}

void gui_set_title(const char *str) {
    SDL_WM_SetCaption(str, NULL);
    title = (char *) str;
}

int load_bios(void) {
    if(sms_console == CONSOLE_SMS && strlen(config_get_str(SMS_BIOS_DIR))) {
        const char *bios = config_get_str(SMS_BIOS_DIR);

        if(sms_mem_load_bios(bios) == 0) {
            sms_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                SMS_MEMCTL_RAM) ^ 0xFF);
        }
    }
    else if(sms_console == CONSOLE_GG && strlen(config_get_str(GG_BIOS_DIR))) {
        const char *bios = config_get_str(GG_BIOS_DIR);

        if(sms_mem_load_bios(bios) == 0) {
            sms_gg_port_write(0x3E, (SMS_MEMCTL_IO | SMS_MEMCTL_BIOS |
                SMS_MEMCTL_CART | SMS_MEMCTL_RAM) ^ 0xFF);
        }
    }
    else if(sms_console == CONSOLE_COLECOVISION) {
        const char *bios = config_get_str(COLECO_BIOS_DIR);

        printf("Loading ColecoVision Bios: %s\n", bios);

        if(coleco_mem_load_bios(bios) != 0) {
            /// TODO make this actually do something useful
            fprintf(stderr, "BIOS failed to load.\n");
            return 1;
        }
    }

    return 0;
}

void handleSRAM(const char *filename, int save) {
    FILE *file;
    char *saveName, *cfg_sram = (char *)config_get_str(SRAM_DIR);

    /* Take preferences into account */
    if(strlen(cfg_sram)) {
        char f[strlen(filename)], *prev = "", *t;

        for(strcpy(f, filename), t = strtok(f, "/");; t = strtok(NULL, "/")) {
            if(t == NULL) {
                break;
            }
            prev = t;
        }
        saveName = malloc(strlen(cfg_sram) + strlen(prev) + 5);
        strcpy(saveName, cfg_sram);
        strcat(saveName, prev);
    }
    else {
        saveName = malloc(strlen(filename) + 5);
        strcpy(saveName, filename);
    }

    strncat(saveName, ".ces", 4);

    if(save) {
        file = fopen(saveName, "wb");
        if(!file) {
            free(saveName);
            return;
        }
        sms_write_cartram_to_file(saveName);
    }
    else {
        file = fopen(saveName, "rb");
        if(!file) {
            free(saveName);
            return;
        }
        sms_read_cartram_from_file(saveName);
    }

    fclose(file);
    free(saveName);
}

int loadRom(const char *filename) {
    int console = rom_detect_console(filename);

    if(console < 0) {
        return 2;
    }

    if(console == CONSOLE_COLECOVISION) {
        coleco_init(video_mode);

        if(load_bios() == 0) {
            coleco_mem_load_rom(filename);
        }
        else {
            return 1;
        }
    }
    else {
        sms_init(video_mode, region);

        load_bios();
        sms_mem_load_rom(filename, console);
        loaded_filename = malloc(strlen(filename)*sizeof(char));
        strncpy(loaded_filename, filename, strlen(filename));
        handleSRAM(loaded_filename, 0);
    }

    return 0;
}

int runRom(const char *filename, Uint32 flags) {
    Uint32 start_time, end_time;

    printf("Loading: %s...\n", filename);
    if(loadRom(filename) != 0) {
        fprintf(stderr, "Error loading %s\n", filename);
        return 1;
    }

    if(sms_console == CONSOLE_GG) {
        SDL_Screen = SDL_SetVideoMode(160, 144, CRAB_BPP, flags);
    }
    else {
        SDL_Screen = SDL_SetVideoMode(256, smsvdp.lines, CRAB_BPP, flags);
    }

    if(SDL_Screen == NULL) {
        fprintf(stderr, "Cannot load screen: %s\n", SDL_GetError());
        return 1;
    }

    /* We manually call it again because the previous calls have no effect till
     SDL_SetVideoMode has been called */
     gui_set_title(title);

     frames_rendered = 0;
     start_time = SDL_GetTicks();
     emu_loop();
     end_time = SDL_GetTicks();

     if(sms_initialized) {
         handleSRAM(loaded_filename, 1);
         free(loaded_filename);
         loaded_filename = NULL;
         sms_shutdown();
     }
    if(coleco_initialized) {
        coleco_shutdown();
        free(loaded_filename);
        loaded_filename = NULL;
    }

     printf("Total rendered frames: %d\n", frames_rendered);
     printf("Total emulation time: %.4f s\n", (float)(end_time - start_time)/1000);
     printf("Average frames per second: %.4f f/s\n", (float)frames_rendered/((end_time - start_time)/1000));

     return 0;
}


void setOptions() {
    int new_value;
    sms_psg_enabled = config_get_int(PSG76489_ENABLED);
    sms_ym2413_enabled = config_get_int(YM2413_ENABLED);
    new_value = config_get_int(REGION);
    if(new_value == SMS_REGION_EXPORT || new_value == SMS_REGION_DOMESTIC) {
        region = new_value;
    }
    new_value = config_get_int(VIDEOMODE);
    if(new_value == SMS_VIDEO_NTSC || new_value == SMS_VIDEO_PAL) {
        video_mode = new_value;
    }

    if(video_mode == SMS_VIDEO_NTSC) {
        target_fps = NTSC_FPS;
    }
    else {
        target_fps = PAL_FPS;
    }
}

void showHelp(char *program) {
    printf("%s {--config|-c PATH/TO/crabemu.cfg} {--rom|-r PATH/TO/ROM.SMS} {--menu|-m} [PATH/TO/ROM.SMS]\n", program);
    printf("\t--config|-c [PATH/TO/crabemu.cfg]\tThis selects a different config than the default\n");
    printf("\t--rom|-r [PATH/TO/ROM.SMS]\tThis tell what rom to load\n");
    printf("\t--menu|-m\tThis loads up a graphical menu to select a rom\n");
    printf("\t--help|-h\tThis message\n");
    printf("\n-r is only required if there are more than one argument.\n");
}

int main(int argc, char *argv[]) {
    int cnt, use_menu = 0, rv = 0;
    char *filename = NULL, *configname = NULL;
    Uint32 flags = SDL_SWSURFACE;

    SDL_Init(SDL_INIT_VIDEO);
    atexit(SDL_Quit);

    if(argc == 2) {
        filename = argv[1];
    }

    /* read options */
    for(cnt = 0; cnt < argc;  ++cnt) {
        if(strncasecmp(argv[cnt], "--rom", 5) == 0 || strncasecmp(argv[cnt], "-r", 2) == 0) {
            if((cnt + 1) < argc) {
                cnt++;
                filename = argv[cnt];
            }
            else {
                fprintf(stderr, "%s requires an argument\n", argv[cnt]);
                exit(1);
            }
        }
        if(strncasecmp(argv[cnt], "--config", 8) == 0 || strncasecmp(argv[cnt], "-c", 2) == 0) {
            if((cnt + 1) < argc) {
                cnt++;
                configname = argv[cnt];
            }
            else {
                fprintf(stderr, "%s requires an argument\n", argv[cnt]);
                exit(1);
            }
        }
        if(strncasecmp(argv[cnt], "--menu", 6) == 0 || strncasecmp(argv[cnt], "-m", 2) == 0) {
            use_menu = 1;
        }
        if(strncasecmp(argv[cnt], "--help", 6) == 0 || strncasecmp(argv[cnt], "-h", 2) == 0) {
            showHelp(argv[0]);
            exit(0);
        }
    }

    // Load the main configuration file
    if(config_open(configname) < 0) {
        exit(1);
    }

    // Load the users local configuration
    config_open_local(configname);

    // Free the Lua state at exit
    atexit(config_close);

    if(argc == 1 || (filename == NULL && use_menu == 0)) {
        /* All is not lost, we may have a default rom in the config */
        if(strlen(config_get_str(DEFAULT_ROM))) {
            filename = (char *)config_get_str(DEFAULT_ROM);
            if(strlen(config_get_str(ROM_DIR))) {
                char *rom_dir = (char *)config_get_str(ROM_DIR);
                if(strncmp(filename, rom_dir, strlen(filename)) == 0) {
                    fprintf(stderr, "No rom selected\n");
                    exit(1);
                }
            }
        }
        else {
            /* Nope, go ahead and exit then */
            fprintf(stderr, "No rom selected\n");
            exit(1);
        }
    }

    if(do_render_scene == NULL) {
        if(strncasecmp("sdl", config_get_str(RENDERER), 3) == 0 || strlen(config_get_str(RENDERER)) == 0) {
#if defined(DIDJ)
            /* The Didj requires this to be a software surface */
            flags = SDL_SWSURFACE|SDL_ANYFORMAT|SDL_ASYNCBLIT;
//             flags = SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_ANYFORMAT|SDL_ASYNCBLIT;
#else
            flags = SDL_HWSURFACE|SDL_DOUBLEBUF|SDL_ANYFORMAT|SDL_ASYNCBLIT;
#endif

            do_render_scene = &sdl_render_scene;
        }
        else {
            fprintf(stderr, "Invalid renderer chosen\n");
            exit(1);
        }
    }

    setOptions();

    if(use_menu) {
        rv = runMenu(flags);
    }
    else {
        rv = runRom(filename, flags);
    }

    if(loaded_filename)
        free(loaded_filename);

    return rv;
}
