/*
    This file is part of CrabEmu.

    Copyright (C) 2011 Harley Laue

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SDL.h"
#include "SDL_ttf.h"
#include "config.h"
#include <dirent.h>
#include <malloc.h>
#include <unistd.h>

#if defined(__WIN32) || defined(__WIN64)
#define WINDOWS
#endif

#if !WINDOWS
#include <sys/stat.h>
#endif

/* From main.c */
extern SDL_Surface *SDL_Screen;
int runRom(const char *filename, Uint32 flags);
Uint32 emu_delay();

TTF_Font *font;
SDL_Color font_color = { 0xff, 0xff, 0xff },
    font_bgcolor = { 0, 0 ,0 },
    selected_color = { 0xab, 0xab, 0xab };
int dir_size = 0;
char **dir_lst = NULL;
SDL_Surface **dir_surface_lst = NULL;

enum DirAction {
    DIR_NO_ACTION,
    DIR_UP,
    DIR_DOWN,
    DIR_PAGEUP,
    DIR_PAGEDOWN
};

void free_dir_list() {
    int x;
    if(dir_lst != NULL) {
        for(x = 0; x < dir_size; x++) {
            free(dir_lst[x]);
        }
        free(dir_lst);
        dir_lst = NULL;
    }
    if(dir_surface_lst != NULL) {
        for(x = 0; x < dir_size; x++) {
            SDL_FreeSurface(dir_surface_lst[x]);
        }
        free(dir_surface_lst);
        dir_surface_lst = NULL;
    }
    dir_size = 0;
}

SDL_Surface *create_dname_surface(const char *name) {
    return TTF_RenderText_Shaded(font, name, font_color, font_bgcolor);
}

int is_dir(const char *name) {
#if !WINDOWS
    struct stat s;
    int d;

    /* try passed in first */
    d = stat(name, &s);
    if(d == 0 && S_ISDIR(s.st_mode)) {
        return 1;
    }
    else {
        /* Try a path relative to the current working dir */
        char *buf = getcwd(NULL, 0);
        char fullname[strlen(buf) + strlen(name) + 2];
        strncpy(fullname, buf, strlen(buf));
        /* ensure that the end of buf was null terminated */
        fullname[strlen(buf)] = '\0';
        strncat(fullname, "/", 1);
        strncat(fullname, name, strlen(name) + 1);

        free(buf);
        d = stat(fullname, &s);

        return d == 0 && S_ISDIR(s.st_mode);
    }
#else
    return 0;
#endif
}

int dir_list(const char *directory) {
    char *real_dir = (char *)directory;
    struct dirent **d_list;
    int d_num;

    /* Problem opening the directory specified, try one more time with the root dir */
    if(!is_dir(real_dir)) {
        printf("unable to open directory specified (%s)... trying /\n", real_dir);
        real_dir = "/";

        /* No use continuing now */
        if(!is_dir(real_dir)) {
            return 1;
        }
    }

    chdir(real_dir);
    d_num = scandir(".", &d_list, NULL, alphasort);

    if(d_num < 0) {
        return 1;
    }
    else {
        int x;
        dir_lst = (char **)malloc(sizeof(char *)*d_num);
        dir_surface_lst = (SDL_Surface **)malloc(sizeof(SDL_Surface *)*d_num);
        dir_size = d_num;
        for(x = 0; x < d_num; x++) {
            dir_lst[x] = malloc(strlen(d_list[x]->d_name)+1);
            strncpy(dir_lst[x], d_list[x]->d_name, strlen(d_list[x]->d_name)+1);
            if(strlen(d_list[x]->d_name) > 56) {
                char trunc[57];
                strncpy(trunc, dir_lst[x], 49);
                trunc[49] = '\0';
                strncat(trunc, "....sms", 7);
                trunc[56] = '\0';
                dir_surface_lst[x] = create_dname_surface(trunc);
            }
            else {
                dir_surface_lst[x] = create_dname_surface(dir_lst[x]);
            }
            free(d_list[x]);
        }
        free(d_list);
    }

    return 0;
}

void draw_dir_list(enum DirAction a, int selected) {
    int x, size = 10, offset = 0;
    SDL_Rect text_loc = { 25, 25, 0, 0 };

    if(dir_size <= size) {
        size = dir_size;
    }

    if(selected >= size) {
        offset = selected - size + 1;
        if(offset + 10 > dir_size) {
            offset = dir_size-10;
        }
    }

    SDL_FillRect(SDL_Screen, NULL, 0);
    for(x = 0; x < size; x++) {
        if((x + offset) == selected) {
            SDL_Rect r;
            r.x = 10;
            r.y = x * dir_surface_lst[x+offset]->h + 25 + 5 * x - 3;
            r.h = dir_surface_lst[x+offset]->h+5;
            r.w = SDL_Screen->w - 20;
            SDL_FillRect(SDL_Screen, &r, 0x55555555);
        }
        SDL_BlitSurface(dir_surface_lst[x+offset], NULL, SDL_Screen, &text_loc);
        text_loc.y += dir_surface_lst[x+offset]->h + 5;
    }
}

int runMenu(Uint32 flags) {
    SDL_Event event;
    SDLKey k;
    int x, run = 1, player[PAD2_2 + 1], sms_pause, gg_pause, quit_button, selected = 0;
    enum DirAction action = DIR_NO_ACTION;

    SDL_Screen = SDL_SetVideoMode(320, 240, 32, flags);

    if(TTF_Init() < 0) {
        fprintf(stderr, "Unable to start SDL_ttf: %s\n", TTF_GetError());
        return -1;
    }

    /* disable key repeat */
    if(SDL_EnableKeyRepeat(0, SDL_DEFAULT_REPEAT_INTERVAL) == -1)
        return -1;

    /* just disable the cursor for now */
    SDL_ShowCursor(SDL_DISABLE);

    font = TTF_OpenFont(config_get_str(FONT_FILE), 12);
    if(!font) {
        fprintf(stderr, "Error: %s\n", TTF_GetError());
        font = TTF_OpenFont("FreeSerif.ttf", 12);
        if(!font) {
            fprintf(stderr, "Error: %s\n", TTF_GetError());
            return -1;
        }
    }

    /* Load the buttons from the configuration */
    sms_pause = config_get_int(SMS_PAUSE);
    gg_pause = config_get_int(GG_PAUSE);
    quit_button = config_get_int(QUIT_BUTTON);

    for(x = PAD1_UP; x <= PAD2_2; ++x) {
        player[x] = config_get_int(x);
    }

    if(dir_list(config_get_str(ROM_DIR))) {
        fprintf(stderr, "Unable to open: %s\n", config_get_str(ROM_DIR));
        return -1;
    }

    draw_dir_list(action, selected);

    while(run) {
        while(SDL_PollEvent(&event)) {
            switch(event.type) {
                case SDL_KEYDOWN:
                    k = event.key.keysym.sym;

                    if(k == quit_button)
                        run = 0;
                    else if(k == player[PAD1_UP]) {
                        selected--;
                    }
                    else if(k == player[PAD1_DOWN]) {
                        selected++;
                    }
                    else if(k == player[PAD1_LEFT]) {
                        selected -= 10;
                    }
                    else if(k == player[PAD1_RIGHT]) {
                        selected += 10;
                    }
                    else if(k == player[PAD1_1] || k == player[PAD1_2]) {
                        if(is_dir(dir_lst[selected])) {
                            dir_list(dir_lst[selected]);
                            /** TODO save dir to config */
                            selected = 0;
                        }
                        else {
                            runRom(dir_lst[selected], flags);
                            /* Clear the event queue */
                            while (SDL_PollEvent(&event));
                            SDL_Screen = SDL_SetVideoMode(320, 240, 32, flags);
                            run = 1;
                            selected = 0;
                        }
                    }
                    else if(k == sms_pause) {}
                    else if(k == gg_pause) {}
                    else if(k == SDLK_BACKQUOTE) {}
                    else if(k == SDLK_h) {}
                    else {}

                    break;
                case SDL_QUIT:
                    run = 0;
                    break;
                default:
                    action = DIR_NO_ACTION;
                    break;
            }
        }

        /* Fix selected if it's too low or too high */
        if(selected < 0) {
            selected = 0;
        }

        if(selected >= dir_size) {
            selected = dir_size-1;
        }

        draw_dir_list(action, selected);
        SDL_Flip(SDL_Screen);

        SDL_Delay(emu_delay());
    }

    free_dir_list();
    TTF_Quit();

    return 0;
}
