/*
    This file is part of CrabEmu.

    Copyright (C) 2011 Harley Laue
    Copyright (C) 2013 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "SDL.h"
#include "SDL_thread.h"
#include "CrabEmu.h"
#include "sound.h"

#include "sms.h"

#define FRAMES_TO_BUFFER    4
#define FRAME_LOW_HALF      (FRAMES_TO_BUFFER / 2)
#define SAMPLE_RATE         44100

static const Uint32 NTSC_SAMPLES_PER_FRAME = 735;  /* SAMPLE_RATE / 60 */
static const Uint32 PAL_SAMPLES_PER_FRAME = 882;   /* SAMPLE_RATE / 50 */

static Uint8 *buffer = NULL;
static Uint32 buffer_size;
static Uint32 read_pos;
static Uint32 write_pos;
static Uint32 samples_per_frame;
static int num_channels;
static Uint32 samples_consumed;
static SDL_sem *sem;

void sound_update_buffer(int16 *buf, int length) {
    /* Only continue if we have a buffer to write to */
    if(!buffer)
        return;

#ifndef DIDJ
    /* It's safer to lock the audio, but the many lock/unlocks a frame cause the
     * Didj to slow down too much. The unfortunate side effect is that the audio
     * sync isn't perfect.
     */
    SDL_LockAudio();
#endif

    if(write_pos + length > buffer_size) {
        int l1, l2;

        l1 = buffer_size - write_pos;
        l2 = length - l1;

        memcpy(buffer + write_pos, buf, l1);
        memcpy(buffer, buf + l1, l2);

        write_pos = l2;
    }
    else {
        memcpy(buffer + write_pos, buf, length);
        write_pos += length;
    }

#ifndef DIDJ
    SDL_UnlockAudio();
#endif
}

static void sound_callback(void *unused, Uint8 *stream, int len) {
#ifdef DIDJ
    static Uint8 data[1024*32];
#elif !defined(_MSC_VER)
    Uint8 data[len];
#else
    Uint8 *data = (Uint8 *)_alloca(len);
#endif
    int8 volume = SDL_MIX_MAXVOLUME - 8;
    int len1, len2;

    /* Don't continue if we don't have a length or buffer */
    if(len <= 0 || !buffer)
        return;

    if(SDL_GetAudioStatus() == SDL_AUDIO_PAUSED ||
       (read_pos + len > write_pos && write_pos > read_pos)) {
        memset(data, 0, len);
    }
    else {
        if(read_pos + len > buffer_size) {
            len2 = buffer_size - read_pos;
            len1 = len - len2;

            memcpy(data, buffer + read_pos, len2);
            memcpy(data + len2, buffer, len1);
            read_pos = len1;
        }
        else {
            memcpy(data, buffer + read_pos, len);
            read_pos += len;

            if(read_pos == buffer_size)
                read_pos = 0;
        }

        samples_consumed += len / 2;
    }

    SDL_MixAudio(stream, data, len, volume);

    /* See if we need to increment our semaphore */
    while(samples_consumed > samples_per_frame) {
        SDL_SemPost(sem);
        samples_consumed -= samples_per_frame;
    }
}

int sound_init(int channels, int region) {
    SDL_AudioSpec audio_spec, actual;

    SDL_InitSubSystem(SDL_INIT_AUDIO);

    /* Create our semaphore */
    if(!(sem = SDL_CreateSemaphore(FRAMES_TO_BUFFER))) {
        fprintf(stderr, "Unable to create semaphore: %s\n", SDL_GetError());
        return -1;
    }

    memset(&audio_spec, 0, sizeof(audio_spec));
    audio_spec.freq = SAMPLE_RATE;
    audio_spec.format = AUDIO_S16SYS;
    audio_spec.channels = channels;
    audio_spec.samples = 512;
    audio_spec.callback = sound_callback;
    audio_spec.userdata = NULL;

    if(SDL_OpenAudio(&audio_spec, &actual) != 0) {
        fprintf(stderr, "Unable to open audio: %s\n", SDL_GetError());
        SDL_DestroySemaphore(sem);
        sem = NULL;
        return -1;
    }

    /* Figure out how big our internal buffer will be */
    if(region == SMS_VIDEO_NTSC)
        samples_per_frame = NTSC_SAMPLES_PER_FRAME * channels;
    else
        samples_per_frame = PAL_SAMPLES_PER_FRAME * channels;

    buffer_size = samples_per_frame * FRAMES_TO_BUFFER * 2;
    buffer = calloc(buffer_size, 1);
    read_pos = 0;
    write_pos = samples_per_frame * FRAME_LOW_HALF * 2;

    return 0;
}

void sound_shutdown(void) {
    if(SDL_WasInit(SDL_INIT_AUDIO) != 0) {
        SDL_PauseAudio(1);
        SDL_CloseAudio();
        SDL_QuitSubSystem(SDL_INIT_AUDIO);

        if(buffer) {
            free(buffer);
            buffer = NULL;
        }

        if(sem) {
            SDL_DestroySemaphore(sem);
            sem = NULL;
        }
    }
}

void sound_reset_buffer(void) {
    /* Again, don't do anything if we don't have a buffer */
    if(!buffer)
        return;

    SDL_LockAudio();

    memset(buffer, 0, buffer_size);
    read_pos = 0;
    write_pos = samples_per_frame * FRAME_LOW_HALF * 2;

    SDL_UnlockAudio();
}

void sound_pause(void) {
    SDL_PauseAudio(1);
}

void sound_unpause(void) {
    SDL_PauseAudio(0);
}

void sound_wait(void) {
    if(!sem)
        return;

    SDL_SemWait(sem);
}
